unit ExpPlot;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Graphs, GraphRange, Forms, {$IFDEF LCL_} LCLType, {$ENDIF} {$ENDIF}
     Graphics, Classes, SysUtils, DOM, MNode, Mixture, Units, Overlay;

type

  TExperimentalPlot = class(TPlot)
  private
    FHMin, FHMax: Double;
    FIMin, FIMax: Double;
  protected
    function GetHMin: Double; override;
    function GetHMax: Double; override;
    function GetIMin: Double; override;
    function GetIMax: Double; override;
  public
    Intensity, Frequency, Weights: TDoubleArray;
    function AdjustFrequency(Freq:Double):Double; override;
    function DeAdjustFrequency(Freq: Double):Double; override;
    function StartPoint: Integer;
    function EndPoint: Integer;
    constructor Create(AOwner: TComponent); override;
    function Empty: Boolean; override;
    procedure LoadFromFile(FileName: string); override;
    procedure WeightedLoadFromStream(Stream: TStream; FileName:string = ''; AskAboutWeights: Boolean = True);
    { procedure ReadDatStream(Stream: TStream); }
    procedure WriteDataToStream(Stream: TStream);
    procedure WriteDatStream(Stream: TStream);
    procedure Invert; override;
    function IntensityAt(Freq: Double; out Point: Integer):Double; { Interpolate Frequency }
    procedure Sort(Ascending: Boolean = True);
    function nDF: Integer; override;
    function FindPeak(AFmin, AFmax: Double; var Position: Double; PlotUnits: TResonanceUnits): Boolean; override;
    function FixupProperties:Exception; override;
    procedure ClearSelection; override;
    procedure SelectFrequencyRange(AFmin, AFmax: Double); override;
    {$IFNDEF CONSOLE}
    procedure Show(G: TGraphD; AImin, AImax: Double; PlotUnits: TResonanceUnits; var PlotRanges: TPlotRanges; ShowName: Boolean); override;
    {$ENDIF}
    procedure ReadProperties(Node: TDOMNode); override;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    function SimpleSaveToStream(Stream: TStream; WithLinkedMNodes: Boolean = False): Boolean; override;
    function WriteProperties(Parent: TDOMNode): TDomNode; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property FrequencyScale: Double index 2 read GetParameter write SetParameter;
    property Baseline:       Double index 3 read GetParameter write SetParameter;
    property Scale:          Double index 4 read GetParameter write SetParameter;

    property FixScale:   Boolean index 5 read GetFlag   write SetFlag;
    property FirstPoint: Integer index 6 read GetSetting write SetSetting;
    property LastPoint:  Integer index 7 read GetSetting write SetSetting;
    property IsLineList: Boolean index 8 read GetFlag   write SetFlag;
  end;

  TLoadType = (Only2, TwoOrThree, Only3);
  TArrayType = (tyI8, tyI16, tyI32, tySingle, tyDouble);

  TExperimentalPlotLoader = class
  private
    FExperimentalPlot: TExperimentalPlot;
    FStream: TStream;
    FExt: string;
    FReportErrors: Boolean;
  protected
    Plots: array of TExperimentalPlot;
    LineNumber: Integer;
    FrequencyFactor: Double;
    ChannelDest: TIntegerArray;
    procedure Read(var a; Count: Integer);
    procedure ReadArray(var a: TDoubleArray; ArrayType: TArrayType);
    procedure ReadAndCopyArray(ArrayIndex:Integer; ArrayType: TArrayType);
    procedure SetValue(Row, Col: Integer; Value: Double);
    procedure CheckAvailableChannels(N, nPoints: Integer; AskAboutWeights: Boolean);
    procedure AutoDetectFrequency;
  public
    constructor Create(AExperimentalPlot: TExperimentalPlot; AStream: TStream; AExt: string);
    function Load(AskAboutWeights: Boolean = True): Boolean; virtual; abstract;
    procedure FileErr(Msg: string);
    procedure IndicateUnits(AUnits:TResonanceUnits);
    property ExperimentalPlot: TExperimentalPlot read FExperimentalPlot;
    property Stream: TStream read FStream;
    property Ext: string read FExt write FExt;
    property ReportErrors: Boolean read FReportErrors;
  end;

  TExperimentalPlotLoaderClass = class of TExperimentalPlotLoader;

procedure RegisterPlotLoader(Loader: TExperimentalPlotLoaderClass; FileFilter: string; FileExt: string);

var
  PlotLoaderFilter: string;

implementation

uses {$IFNDEF CONSOLE} PickDest, ClipBrd, Huge,
       {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$ENDIF}
     MNodeColour, Math, MathSup, TextStrm, ShellSort, DatLoader;

var
  PlotLoaders: array of TExperimentalPlotLoaderClass;

procedure RegisterPlotLoader(Loader: TExperimentalPlotLoaderClass; FileFilter: string; FileExt: string);
  procedure SubRegister(Loader: TExperimentalPlotLoaderClass; FileFilter: string; FileExt: string);
  begin
    SetLength(PlotLoaders, Length(PlotLoaders)+1);
    PlotLoaders[Length(PlotLoaders)-1] := Loader;
    {$IFDEF LINUX}
    FileExt := FileExt + ';' + UpperCase(FileExt);
    {$ENDIF}
    PlotLoaderFilter := PlotLoaderFilter + FileFilter + ' (' + FileExt + ')|' + FileExt + '|';
  end;
begin
  { Force ordinary text files to be first on the list, so they will be tried last }
  if Length(PlotLoaders) = 0 then
    SubRegister(TDatFileLoader, 'Text Files', '*.dat;*.csv');
  SubRegister(Loader, FileFilter, FileExt);
end;

{ TExperimentalPlot }

procedure TExperimentalPlot.WeightedLoadFromStream(Stream: TStream; FileName: string; AskAboutWeights: Boolean);
{ If Ext is blank, try autodetecting file format }
var
  i: Integer;
  PlotLoader: TExperimentalPlotLoader;
  Ext: string;
begin
  SetLength(Peaks,0);
  Ext := AnsiLowerCase(ExtractFileExt(FileName));
  for i := High(PlotLoaders) downto 0 do begin
    PlotLoader := PlotLoaders[i].Create(Self, Stream, Ext);
    try
      if PlotLoader.Load(AskAboutWeights) then begin
        Sort;
        Exit;
      end;
    finally
      PlotLoader.Free;
    end;
  end;
  raise EReportException.Create('Failed to load ' + Ext + ' file');
end;

procedure TExperimentalPlot.LoadFromFile(FileName: string);
var
  Stream: TStream;
begin
  Stream := nil;
  try
{$IFNDEF CONSOLE}
    if FileName = '' then
      if ClipBoard.HasFormat(CF_TEXT) then begin
        Stream := THugeMemoryStream.Create;
        THugeMemoryStream(Stream).SetFromClipboard(CF_TEXT);
      end else
        raise Exception.Create('No Text on clipboard')
    else
{$ENDIF}
      Stream := TFileStream.Create(FileName, fmOpenRead);
    WeightedLoadFromStream(Stream, FileName, True);
  finally
    Stream.Free;
  end;
end;

function TExperimentalPlot.FixupProperties:Exception;
begin
  Result := inherited FixupProperties;
  if Assigned(Result) then Exit;
  LastPoint := EndPoint;
  if High(Frequency) < 0 then Exit; { Empty }
  Limits(Frequency, FHmin, FHMax, StartPoint, EndPoint);
  { TODO: sort by frequency }
  { Set defaults for vertical scaling }
  Limits(Intensity, FImin, FImax, StartPoint, EndPoint);
  if IsLineList then
    FIMin := Min(FIMin,0);
  if Scale = 0 then begin
    { Only set these properties if they appear never to have been set }
    Baseline := FImin;
    if FImax > FImin then
      Scale := 1/(FImax - FImin)
    else
      Scale := 1;
  end;
end;

procedure TExperimentalPlot.WriteDatStream(Stream: TStream);
var
  i: Integer;
  buffer: string;
begin
  for i := 0 to High(Frequency) do begin
    buffer := DtosM(Frequency[i],-1,-1) + #9 + DtosM(Intensity[i],-1,-1);
    if High(Weights) > 0 then
      buffer := buffer + #9 + DtosM(Weights[i],-1,-1);
    buffer := buffer + sLineBreak;
    Stream.Write(buffer[1], Length(buffer));
  end;
end;

{$IFDEF OLD}
procedure TExperimentalPlot.ReadDatStream(Stream: TStream);
var
  S: TTextStream;
  MaxCols, nCols, nRows, FrequencyChannel: Integer;
  Last1, Last2: Double;
  nInc1, nDec1, nInc2, nDec2: Integer;
  procedure CountIncrements(var Last: Double; var nInc, nDec: Integer);
  var
    Next: Double;
  begin
    if nRows = 0 then begin
      Last := S.RealToken;
      nInc := 0;
      nDec := 0;
    end else begin
      Next := S.RealToken;
      if Next > Last then
        Inc(nInc)
      else if Next < Last then
        Inc(nDec);
      Last := Next;
    end
  end;
begin
  S := TTextStream.Create(Stream);
  try
    S.ItemSeparators := [isTab, isComma];
    { Determine number of Columns and Rows of data }
    MaxCols := 0;
    nRows := 0;
    { Autodetect frequency channel by counting number of changes of each sign in each channel }
    while not S.AtEndOfFile do begin
      nCols := 0;
      while not S.AtEndOfLine do
        if S.GetFreeItem then begin
          if nCols = 0 then
            CountIncrements(Last1, nInc1, nDec1)
          else if nCols = 1 then
            CountIncrements(Last2, nInc2, nDec2);
          Inc(nCols);
        end;
      S.NextLine;
      if nCols > 0 then begin
        Inc(nRows);
        if nCols > MaxCols then
          MaxCols := nCols;
      end;
    end;
    if (nRows > 0) and (MaxCols >= 2) then begin
      if Abs(nInc1-nDec1) > Abs(nInc2-nDec2) then
        FrequencyChannel := 0
      else
        FrequencyChannel := 1;
      { Allocate space }
      SetLength(Frequency, nRows);
      SetLength(Intensity, nRows);
      if MaxCols > 2 then
        SetLength(Weights, nRows)
      else
        SetLength(Weights, 0);
      { And repeat read logic, this time setting arrays }
      S.Position := 0;
      nRows := 0;
      while not S.AtEndOfFile do begin
        nCols := 0;
        while not S.AtEndOfLine do
          if S.GetFreeItem then begin
            if nCols = 2 then
              Weights[nRows] := S.RealToken
            else if nCols = FrequencyChannel then
              Frequency[nRows] := S.RealToken
            else if nCols = 1-FrequencyChannel then
              Intensity[nRows] := S.RealToken;
            Inc(nCols);
          end;
        S.NextLine;
        if nCols > 0 then
          Inc(nRows);
      end;
    end else
      raise EReportException.Create('No Data Found');
    CheckProperties;
  finally
    S.Free;
  end;
end;
{$ENDIF}

{$IFNDEF CONSOLE}
procedure TExperimentalPlot.Show(G: TGraphD; AImin, AImax: Double; PlotUnits: TResonanceUnits; var PlotRanges: TPlotRanges; ShowName: Boolean);
var
  i, nPoints: Integer;
  Mi, Mx, Fi, Fx, s, x: Double;
begin
  G.MapArg(Fi, Fx, Mi, Mx);
  if Fi > Fx then begin
    s := Fi;
    Fi := Fx;
    Fx := s;
  end;
  nPoints := 0;
  if not FixScale then begin
    { Determine limits from points in range }
    for i := StartPoint to EndPoint do begin
      x := Convert(AdjustFrequency(Frequency[i]),Units,PlotUnits);
      if ( x >= Fi) and (x <= Fx) then begin
        if nPoints = 0 then begin
          Mi := Intensity[i];
          Mx := Mi;
        end else begin
          if Intensity[i] < Mi then Mi := Intensity[i];
          if Intensity[i] > Mx then Mx := Intensity[i];
        end;
        Inc(nPoints);
      end;
    end;
    if nPoints = 0 then Exit;
    if IsLineList then
      Mi := Min(Mi,0);
    if Mx <> Mi then
      s := (AImax-AImin)/(Mx-Mi)
    else
      s := 1; { Flat }
    Mi := Mi - AImin/s;
  end else begin
    s := Scale;
    Mi := Baseline;
  end;
  if StartPoint >= EndPoint then
    Exit;
  if Colour <> clNone then
    G.StyleP(Colour, G.LineStyle.Width, G.LineStyle.Pattern)
  else
    G.Style(1);
  if IsLineList then
    for i := StartPoint to EndPoint do begin
      x := Convert(AdjustFrequency(Frequency[i]),Units,PlotUnits);
      G.ClipPoint(x, AImin);
      G.ClipJoin(x, s*(Intensity[i]-Mi));
    end
  else begin
    G.ClipPoint(Convert(AdjustFrequency(Frequency[StartPoint]),Units,PlotUnits),
                s*(Intensity[StartPoint]-Mi));
    for i := StartPoint+1 to EndPoint do
      G.ClipJoin(Convert(AdjustFrequency(Frequency[i]),Units,PlotUnits),
                 s*(Intensity[i]-Mi));
  end;
  AddPlotRange(PlotRanges, G, AImin, AImax, AImin/s + Mi, AImax/s + Mi);
  inherited;
end;
{$ENDIF}

function TExperimentalPlot.WriteProperties(Parent: TDOMNode): TDomNode;
var
  CDATASection: TDOMCDATASection;
  Strings: TStringList;
  i: Integer;
  s: string;
begin
  if LastPoint >= High(Intensity) then
    LastPoint := -1;
  Strings := TStringList.Create;
  try
    Result := inherited WriteProperties(Parent);
    Strings.Capacity := Length(Intensity)+1;
    Strings.Add('');
    for i := 0 to High(Intensity) do begin
      s := FloatToStr(Frequency[i])+' '+FloatToStr(Intensity[i]);
      if i <= High(Weights) then
        s := s + ' '+FloatToStr(Weights[i]);
      Strings.Add(s);
    end;
    CDATASection := Parent.OwnerDocument.CreateCDATASection(Strings.Text);
    Result.AppendChild(CDATASection);
    WritePeaks(Parent, Result);
  finally
    Strings.Free;
  end;
end;

function TExperimentalPlot.SimpleSaveToStream(Stream: TStream; WithLinkedMNodes: Boolean): Boolean;
var
  i: Integer;
  s: string;
begin
  for i := 0 to High(Intensity) do begin
    s := FloatToStr(AdjustFrequency(Frequency[i]))+' '+FloatToStr(Intensity[i]);
    if i <= High(Weights) then
      s := s + ' '+FloatToStr(Weights[i]);
    s := s + sLineBreak;
    Stream.Write(s[1],Length(s))
  end;
  Result := True;
end;

procedure TExperimentalPlot.ReadProperties(Node: TDOMNode);
var
  Child: TDOMNode;
  Stream: TStringStream;
begin
  inherited ReadProperties(Node);
  Child := Node.FirstChild;
  while Assigned(Child) do begin
    if Child is TDOMCDATASection then begin
      Stream := TStringStream.Create(TDOMCDATASection(Child).Data);
      try
        WeightedLoadFromStream(Stream, '.dat', False);
        if LastPoint = -1 then
          LastPoint := High(Intensity);
        ReadPeaks(Child.NextSibling);
        Exit;
      finally
        Stream.Free;
      end;
    end;
    Child := Child.NextSibling;
  end;
end;

constructor TExperimentalPlot.Create(AOwner: TComponent);
begin
  SavedProperties := 'LastPoint=-1,FrequencyScale=1,' + SavedProperties;
  DimensionlessParameters := 'Baseline,Scale,FrequencyScale,' + DimensionlessParameters;
  inherited;
end;

procedure TExperimentalPlot.ScaleParameters(UFrom, UTo: TResonanceUnits);
var
  i: Integer;
  Sum: Double;
begin
  inherited;
  Sum := 0;
  for i := 0 to High(Frequency) do begin
    Sum := Sum + Frequency[i];
    Frequency[i] := Convert(Frequency[i], UFrom, UTo);
  end;
  if Length(Frequency) > 0 then
    FrequencyOffset := ConvertWidth(FrequencyOffset, Sum/Length(Frequency), UFrom, Uto);
  CheckProperties;
end;

function TExperimentalPlot.FindPeak(AFmin, AFmax: Double; var Position: Double; PlotUnits: TResonanceUnits): Boolean;
var
  i: Integer;
  First, Last: Integer; { Range of points within given frequency range }
  Mx, Mi: Double;
  Threshold, SumI, SumIX, v, f: Double;
begin
  AFmin := Convert(AFmin, PlotUnits, Units);
  AFmax := Convert(AFmax, PlotUnits, Units);
  InOrder(AFmin, AFmax);
  Threshold := 0.5;
  if Owner is TOverlays then
    Threshold := TOverlays(Owner).PeakThreshold;
  First := -1;
  Last := -1;
  Result := False;
  for i := StartPoint to EndPoint do begin
    f := AdjustFrequency(Frequency[i]);
    if (f >= AFmin) and (f <= AFmax) then begin
      Last := i;
      if First < 0 then begin
        First := i;
        Mx := Intensity[i];
        Mi := Intensity[i];
      end else begin
        if Intensity[i] < Mi then
          Mi := Intensity[i];
        if Intensity[i] > Mx then
          Mx := Intensity[i];
      end;
    end;
  end;
  if (First = Last) or (Mx = Mi) then begin
    { 1 or fewer points }
    if (First >= 0) and IsLineList then begin
      Position := AdjustFrequency(Frequency[First]);
      Result := True;
    end;
    Exit;
  end;
  { Now use centre of mass algorithm on the points above the threshold }
  Threshold := Mi + Threshold*(Mx-Mi);
  SumI := 0;
  SumIX := 0;
  for i := First to Last do begin
    f := AdjustFrequency(Frequency[i]);
    if (f >= AFmin) and (f <= AFmax) then begin
      if IsLineList then
        v := Abs(Intensity[i])
      else
        v := Intensity[i] - Threshold;
      if v > 0 then begin
        SumI := SumI + v;
        SumIX := SumIX + v * f;
      end;
    end;
  end;
  if SumI > 0.01*(Mx-Mi) then begin
    { Check region not too flat }
    Position := SumIX / SumI;
    Result := True;
  end;
end;

procedure TExperimentalPlot.Invert;
var
  i: Integer;
begin
  Limits(Intensity, FImin, FImax); { Just in case }
  for i := 0 to High(Intensity) do
    Intensity[i] := FImax+FImin-Intensity[i];
end;

function TExperimentalPlot.IntensityAt(Freq: Double; out Point: Integer):Double; { Interpolate Frequency }
var
  i: Integer;
begin
  if Frequency[0] > Frequency[High(Frequency)] then
    Sort;
  i := 1; { Find an interval i, i-1; if off end choose last interval }
  while (Freq > Frequency[i]) and (i < High(Frequency)) do
    Inc(i);
  Result := Frequency[i]-Frequency[i-1];
  if Result > 0 then
    Result := (Freq-Frequency[i-1])/Result
  else
    Result := Frequency[i]; { Zero interval }
  Result := Result * (Intensity[i]-Intensity[i-1]) + Intensity[i-1];
  Point := i;
end;

function TExperimentalPlot.EndPoint: Integer;
begin
  if (LastPoint < 0) or (LastPoint > High(Intensity)) then
    Result := High(Intensity)
  else
    Result := LastPoint;
end;

function TExperimentalPlot.StartPoint: Integer;
begin
  if FirstPoint < 0 then
    Result := 0
  else
    Result := FirstPoint;
end;

procedure TExperimentalPlot.SelectFrequencyRange(AFmin, AFmax: Double);
var
  i, First, Last: Integer;
begin
  First := High(Frequency);
  Last := 0;
  for i := 0 to High(Frequency) do
    if (Frequency[i] >= AFmin) and (Frequency[i] <= AFmax) then begin
      if First > i then First := i;
      if Last < i then Last := i;
    end;
  FirstPoint := First;
  LastPoint := Last;
end;

type
  TExperimentalPlotSorter = class(TShellSorter)
    E: TExperimentalPlot;
    Descending: Boolean;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

procedure TExperimentalPlot.Sort(Ascending: Boolean = True);
begin
  with TExperimentalPlotSorter.Create do begin
    E := Self;
    Descending := not Ascending;
    Sort(Length(Frequency));
    Free;
  end;
end;

procedure TExperimentalPlot.ClearSelection;
begin
  FirstPoint := 0;
  LastPoint := High(Intensity);
end;

procedure TExperimentalPlot.WriteDataToStream(Stream: TStream);
begin
  WriteDatStream(Stream);
end;

function TExperimentalPlot.nDF: Integer;
begin
  Result := Length(Frequency);
end;

function TExperimentalPlot.GetHMin: Double;
begin
  Result := AdjustFrequency(FHMin);
end;

function TExperimentalPlot.GetHMax: Double;
begin
  Result := AdjustFrequency(FHMax);;
end;

function TExperimentalPlot.GetIMin: Double;
begin
  Result := FIMin;
end;

function TExperimentalPlot.GetIMax: Double;
begin
  Result := FIMax;
end;

const
  ExperimentalPlotSettingNames: array[4..7] of string = (
'True to use Scale and Baseline when plotting, rather than autoscaling to fit.',
'First point to use in fitting and plotting.',
'Last point to use in fitting and plotting.',
'Set to treat data as a list of line positions, optionally with intensity.'
);

function TExperimentalPlot.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(ExperimentalPlotSettingNames))
     and (AIndex <= High(ExperimentalPlotSettingNames)) then
    Result := ExperimentalPlotSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 ExperimentalPlotParameterNames: array [2..4] of string = (
'Scaling factor for horizontal scale.',
'Baseline position.',
'Vertical scale factor used in plotting.'
);

function TExperimentalPlot.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(ExperimentalPlotParameterNames))
     and (AIndex <= High(ExperimentalPlotParameterNames)) then
    Result := ExperimentalPlotParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TExperimentalPlot.Empty: Boolean;
begin
  Result := High(Frequency) <= 0;
end;

function TExperimentalPlot.AdjustFrequency(Freq: Double): Double;
begin
  Result := Freq*FrequencyScale + FrequencyOffset;
end;

function TExperimentalPlot.DeAdjustFrequency(Freq: Double): Double;
begin
  Result := (Freq-FrequencyOffset)/FrequencyScale ;
end;

{ TExperimentalPlotSorter }

function TExperimentalPlotSorter.InOrder(I, J: Integer): Boolean;
begin
  if Descending then
    Result := E.Frequency[I] >= E.Frequency[J]
  else
    Result := E.Frequency[I] <= E.Frequency[J];
end;

procedure TExperimentalPlotSorter.Swap(I, J: Integer);
var
  T: Double;
begin
  with E do begin
    T := Frequency[I];
    Frequency[I] := Frequency[J];
    Frequency[J] := T;
    T := Intensity[I];
    Intensity[I] := Intensity[J];
    Intensity[J] := T;
    if Length(Weights) > 0 then begin
      T := Weights[I];
      Weights[I] := Weights[J];
      Weights[J] := T;
    end;
  end;
end;

{ TExperimentalPlotLoader }

procedure TExperimentalPlotLoader.AutoDetectFrequency;
{ Determine frequency channel by looking at the number of changes up and changes down in each channel }
var
  nIncf, nDecf, nInci, nDeci: Integer;
  t: TDoubleArray;
  procedure CountIncrements(const v: TDoubleArray; out nInc, nDec: Integer);
  var
    i: Integer;
  begin
    nInc := 0;
    nDec := 0;
    for i := 1 to High(v) do
      case Sign(v[i]-v[i-1]) of
        -1: Inc(nDec);
         1: Inc(nInc);
      end;
  end;
begin
  if Length(ExperimentalPlot.Weights) <> 0 then Exit;
  CountIncrements(ExperimentalPlot.Frequency, nIncf, nDecf);
  CountIncrements(ExperimentalPlot.Intensity, nInci, nDeci);
  if Abs(nInci-nDeci) > Abs(nIncf-nDecf) then begin
    t := ExperimentalPlot.Frequency;
    ExperimentalPlot.Frequency := ExperimentalPlot.Intensity;
    ExperimentalPlot.Intensity := t;
  end;
end;

procedure TExperimentalPlotLoader.CheckAvailableChannels(N, nPoints: Integer; AskAboutWeights: Boolean);
var
  nPlots, nWeights, i: Integer;
  Nam, Com: string;
begin
  if N < 1 then
    raise EReportException.CreateFmt('Only %d channels of data', [N]);
  nWeights := 0;
  nPlots := 1;
  if not AskAboutWeights then begin
    if N > 3 then
      raise EReportException.CreateFmt('Too many channels:%d', [N])
    else if N = 3 then begin
      for i := 0 to N-1 do
        if ChannelDest[i] < 0 then
          ChannelDest[i] := 1000; { Switch unused channel to weights }
      nWeights := 1;
    end;
  end else if N > 2 then begin
    {$IFNDEF CONSOLE}
    if High(ChannelDest) > 0 then begin
      nPlots := PickDestForm.Execute(ChannelDest, nWeights);
      if nPlots < 1 then
        Abort;
    end else
    {$ENDIF}
      raise EReportException.CreateFmt('Too many (%d) channels of data', [N]);
  end;
  SetLength(Plots, nPlots);
  Nam := ExperimentalPlot.Name;
  Com := ExperimentalPlot.Comment;
  for i := 0 to nPlots-1 do begin
    if i = 0 then
      Plots[0] := ExperimentalPlot
    else
      Plots[i] := TExperimentalPlot.Create(ExperimentalPlot.Owner);
    if nPlots > 1 then begin
      Plots[i].Name := Nam + ' ' + IntToStr(i+1);
      Plots[i].Comment := Com + ' Channel ' + IntToStr(i+1);
    end;
    SetLength(Plots[i].Frequency, nPoints);
    SetLength(Plots[i].Intensity, nPoints);
    if nWeights > 0 then
      SetLength(Plots[i].Weights, nPoints)
    else
      SetLength(Plots[i].Weights, 0)
  end;
  if N = 1 then
    for i := 0 to nPoints-1 do
      Plots[0].Intensity[i] := 1;
end;

constructor TExperimentalPlotLoader.Create(AExperimentalPlot: TExperimentalPlot; AStream: TStream; AExt: string);
begin
  FExperimentalPlot := AExperimentalPlot;
  FStream := AStream;
  FExt := AnsiLowerCase(AExt);
  FReportErrors := True;
  LineNumber := 0;
  FrequencyFactor := 1;
end;

procedure TExperimentalPlotLoader.FileErr(Msg: string);
begin
  if ReportErrors then begin
    if LineNumber <> 0 then
      Msg := Format('%s on line %d', [Msg, LineNumber]);
    {$IFNDEF CONSOLE}
    case Application.MessageBox(PChar(Msg + '. Report further errors?'), 'Error', MB_YESNOCANCEL) of
      IDCANCEL: Abort;
      IDNO: FReportErrors := False;
    end;
    {$ELSE}
    WriteLn('Error in file load:', Msg);
    {$ENDIF}
  end;
end;

procedure TExperimentalPlotLoader.IndicateUnits(AUnits: TResonanceUnits);
begin
  with ExperimentalPlot.Owner as TOverlays do
    if Units <> AUnits then
      if Count = 1 then
        Units := AUnits; { Only one overlay - change overlay units }
  ExperimentalPlot.Units := AUnits;
end;

procedure TExperimentalPlotLoader.Read(var a; Count: Integer);
begin
  if Stream.Read(a, Count) <> Count then
    raise Exception.Create('Unexpected end of file');
end;

procedure TExperimentalPlotLoader.ReadAndCopyArray(ArrayIndex: Integer; ArrayType: TArrayType);
var
  A: TDoubleArray;
  i: Integer;
begin
  if ArrayIndex = 0 then
    A := ExperimentalPlot.Frequency
  else if ArrayIndex < 1000 then
    A := Plots[ArrayIndex-1].Intensity
  else
    A := Plots[ArrayIndex-1000].Weights;
  ReadArray(A, ArrayType);
  if ArrayIndex = 0 then
    for i := 1 to High(Plots) do
      Move(A[0], Plots[i].Frequency[0], Length(A)*SizeOf(Double));
end;

procedure TExperimentalPlotLoader.ReadArray(var a: TDoubleArray; ArrayType: TArrayType);
var
  i: Integer;
  PSingle: ^Single;
  PLongint: ^Longint;
  PSmallint: ^Smallint;
  PShortint: ^Shortint;
begin
  case ArrayType of
    tyDouble: Read(a[0], Length(a)*Sizeof(Double)); { Easy case }
    tySingle: begin
               Read(a[0], Length(a)*Sizeof(Single)); { Read in place and re-pack }
               PSingle := @a[0];
               Inc(PSingle, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 a[i] := PSingle^;
                 Dec(PSingle);
               end;
             end;
    tyI32:   begin
               Read(a[0], Length(a)*Sizeof(Longint)); { Read in place and re-pack }
               PLongint := @a[0];
               Inc(PLongint, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 a[i] := PLongint^;
                 Dec(PLongint);
               end;
             end;
    tyI16:   begin
               Read(a[0], Length(a)*Sizeof(Smallint)); { Read in place and re-pack }
               PSmallint := @a[0];
               Inc(PSmallint, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 a[i] := PSmallint^;
                 Dec(PSmallint);
               end;
             end;
    tyI8:    begin
               Read(a[0], Length(a)*Sizeof(Shortint)); { Read in place and re-pack }
               PShortint := @a[0];
               Inc(PShortint, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 a[i] := PShortint^;
                 Dec(PShortint);
               end;
             end;
  end;
end;

procedure TExperimentalPlotLoader.SetValue(Row, Col: Integer; Value: Double);
var
  i: Integer;
begin
  Col := ChannelDest[Col];
  if Col < 0 then
    Exit;
  if Col = 0 then
    for i := 0 to High(Plots) do
      Plots[i].Frequency[Row] := Value
  else if Col < 1000 then
    Plots[Col-1].Intensity[Row] := Value
  else
    Plots[Col-1000].Weights[Row] := Value
end;

initialization
  RegisterClasses([TExperimentalPlot]);
end.
