unit JPLCat;

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

{ Read JPL line catalog files as defined in
  H.M. Pickett, R.L.Poynter, E.A. Cohen, M.L. Delitsky, J.C. Pearson and H.S.P. Muller
  J. Quant. Spectrosc. Radiat. Transfer 60, 883 (1998)
  See also http://spec.jpl.nasa.gov/
}

interface

uses Classes, Mixture, TextStrm, Obs;

type
  TJPLCatReader = class(TObservationsReader)
    function DoReadOneLine: Boolean; override;
  end;

function CheckJPLCatFile(Mixture: TMixture; S: TFreeFormatStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; Offset:Double): Double;

implementation

uses Graphics, MatrixVec, Math, MathSup, Units, SysUtils, QuantumNumber,
     HundA, OAsymT;

type
  TJPLCatLine = record
    FREQ, ERR, LGINT, ELO: Double;
    DR, GUP, TAG, Q, H, NQN: Integer;
    QNP, QN: array [0..5] of Integer;
  end;

function LoadLine(const Line: string; var JPLCatLine: TJPLCatLine):Boolean;
var
  p, i, NonBlankCount: Integer;
  function GetField(Width: Integer): string;
  begin
    Result := Copy(Line, p, Width);
    if Trim(Result) <> '' then
      Inc(NonBlankCount)
    else
      Result := '0'; { Black fields are read as zero to avoid spurious crashes }
    Inc(p, Width);
  end;
  function NextQno: Integer;
  var
    s: string[2];
  begin
    s := GetField(2);
    if s[1] in ['A'..'Z'] then
      Result := (Ord(s[1]) - Ord('A') + 10)*10 + StrToInt(s[2])
    else
      Result := StrToInt(s);
    { TODO: handle a..z as per CDMS }
  end;
begin
  p := 1;
  Result := False;
  with JPLCatLine do begin
    if not TryStrToFloat(GetField(13), FREQ) then
      Exit;
    ERR   := StrToFloat(GetField(8));
    LGINT := StrToFloat(GetField(8));
    DR    := StrToInt(GetField(2));
    ELO   := StrToFloat(GetField(10));
    GUP   := StrToInt(GetField(3));
    TAG   := StrToInt(GetField(7));
    Q     := StrToInt(GetField(2));
    H     := StrToInt(GetField(1));
    NQN   := StrToInt(GetField(1));
    NonBlankCount := 0;
    for i := 0 to 5 do
      QNP[i] := NextQno;
    for i := 0 to 5 do
      QN[i] := NextQno;
    Result := (NonBlankCount = 2*NQN) and (NQN > 0);
  end;
end;

procedure AdjustQuantumNumbers(var Line: TJPLCatLine);
var
  i, Mask: Integer;
begin
  { Double (and subtract 1 if necessary) the quantum numbers }
  with Line do begin
    Mask := H;
    for i := NQN-1 downto 0 do begin
      QNP[i] := QNP[i]*2 - (Mask and 1);
      QN[i] := QN[i]*2 - (Mask and 1);
      Mask := Mask shr 1;
    end;
  end;
end;

function ConvertQuantumNumbers(StateU: TState; var JU:TQuantumNumber; var SymU, IndexU: Integer;
                               StateL: TState; var JL:TQuantumNumber; var SymL, IndexL: Integer;
                               var Name: string;
                               const CatLine: TJPLCatLine):Boolean;
                               
  function MatchHyperfine(BasisState: TBasisState; const Q: array of Integer):Boolean;
  var
    i, ioff: Integer;
    FF : TQuantumNumber;
  begin
    Result := False;
    ioff := CatLine.NQN-1;
    FF := Q[ioff];
    for i := StateU.Manifold.Molecule.nNuclei-2 downto 0 do begin
      Dec(FF,BasisState.HyperfineState[i+1]);
      Dec(ioff);
      if Q[ioff] <> FF then
        Exit;
    end;
    Result := True;
  end;

  function GetLinearQnos(vflag: Integer; TestLambda: Integer = 0): Boolean;
    { Q=1 : Linear Sigma state }
    var
      NameU, NameL: string;
    function GetSym(var J: TQuantumNumber; const QS: array of Integer; State: TState; var ASym, Number: Integer; var AName: string):Boolean;
    var
      NMin: Integer;
      N: TQuantumNumber;
      isym, index, Sindex: Integer;
      EigenValues: TEigenvalues;
    begin
      with State as TLinear do begin
        N := QS[0];
        J := QS[CatLine.NQN-1];
        if Odd(N div 2) xor (Lambda < 0) then
          ASym := 1 { - parity }
        else
          ASym := 0;
        if CatLine.NQN = 2+vflag then begin
          NMin := J - S;
          if Odd((N-NMin) div 2) then
            Inc(NMin,2);
          while NMin < 0 do
            Inc(Nmin,4);
          Number := 1;
          while Nmin < N do begin
            Inc(Number,1);
            Inc(Nmin,4);
          end;
          Result := NMin = N;
          AName := ''; { TODO: may need to be present }
        end else begin
          { Hyperfine structure present }
          Sindex := 0;
          if vflag > 0 then
            Sindex := QS[1+TestLambda] div 2;
          for isym := 0 to State.Molecule.LastSym {=1} do begin
            Eigenvalues := State.Manifold.MakeBasis(J, isym);
            try
              for index := 1 to Eigenvalues.Size do
                with Eigenvalues.Basis[index] as TLinearBasisState do begin
                  // WriteLn(N, ' ', QS[0], ' ', J, ' ', QS[1+vflag]);
                  if (N = QS[0]) and (J = QS[1+vflag+TestLambda]) and (State = State.Manifold.State[Sindex])
                     and (not (TestLambda = 1) or ((QS[1]<0) = (isym=1)))
                     and MatchHyperfine(Eigenvalues.Basis[index],QS) then begin
                    Result := True;
                    ASym := isym;
                    Number := 0;
                    AName := StateLabel;
                    Exit;
                  end;
                end;
            finally
              Eigenvalues.Free;
            end;
          end;
          Result := False;
        end;
       end;
    end;
  begin
    with CatLine do begin
      Result := GetSym(JU, QNP, StateU, SymU, IndexU, NameU);
      Result := Result and GetSym(JL, QN, StateL, SymL, IndexL, NameL);
      if Result and ((IndexU = 0) or (IndexL = 0)) then
        Name := ':' + NameU + ' - ' + NameL + ':' + Name;
    end;
  end;

  function GetAsymmetricTopQnos(vflag: Integer): Boolean;
    var
      NameU, NameL: string;
    function GetSym(State: TState; const Q: array of Integer; var F: TQuantumNumber; var ASym, Number: Integer; var AName: string; Log: Boolean):Boolean;
    var
      Eigenvalues: TEigenvalues;
      isym, index, Sindex: Integer;
      JJ: TQuantumNumber;
    begin
      Result := False;
      F := Q[CatLine.NQN-1];
      if CatLine.NQN = 3+vflag then begin
        JJ := Q[0];
        F := JJ;
      end else
        JJ := Q[3+vflag];
      Sindex := 0;
      if vflag > 0 then
        Sindex := Q[3] div 2;
      if Sindex >= State.Manifold.nStates then begin
        if Log then WriteLn('Requested State ', Sindex, ' but only ', State.Manifold.nStates, ' available');
        Exit;
      end;
      for isym := 0 to State.Molecule.LastSym do begin
        if SmallestProjection(F) <> State.OddSpin then begin
          if Log then WriteLn('Odd/Even spin mismatch; F=', F);
          Exit;
        end;
        Eigenvalues := State.Manifold.MakeBasis(F, isym);
        try
          for index := 1 to Eigenvalues.Size do
            with Eigenvalues.Basis[index] as TAsymmetricTopBasisState do begin
              if Log then WriteLn('N=', N, ', Ka=', Ka, ', Kc=', Kc);
              if (N = Q[0]) and (Ka = Q[1]) and (Kc = Q[2])
                 and (J = JJ)
                 and (State = State.Manifold.State[Sindex]) then begin
                Result := True;
                ASym := isym;
                if (State.Manifold.nStates > 1) or (CatLine.NQN > 3+vflag) then
                  Number := 0
                else
                  Number := index;
                AName := StateLabel;
                Exit;
              end;
            end;
        finally
          Eigenvalues.Free;
        end;
      end;
    end;
  begin
    with CatLine do begin
      Result := GetSym(StateU, QNP, JU, SymU, IndexU, NameU, False);
      if not Result then
        GetSym(StateU, QNP, JU, SymU, IndexU, NameU, True)
      else begin
        Result := GetSym(StateL, QN, JL, SymL, IndexL, NameL, False);
        if not Result then
          GetSym(StateL, QN, JL, SymL, IndexL, NameL, True);
      end;
      if Result and ((IndexU = 0) or (IndexL = 0)) then
        Name := ':' + NameU + ' - ' + NameL + ':' + Name;
    end;
  end;

begin
  { This will need adjustment }
  case CatLine.Q of
    1: Result := not GetLinearQnos(0);
   12: Result := not GetLinearQnos(1);
    2: if StateL is TLinear then
         Result := not GetLinearQnos(0, 1)
       else
         raise NotYet;
   13: if StateL is TLinear then
         Result := not GetLinearQnos(1, 1)
       else
         raise NotYet;
    3: if StateL is TAsymmetricTop then
         Result := not GetAsymmetricTopQnos(0)
       else
         raise NotYet;
   14: Result := not GetAsymmetricTopQnos(1)
  else
    Result := True; { Not implemented }
  end;
end;

function CheckJPLCatFile(Mixture: TMixture; S: TFreeFormatStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; Offset:Double): Double;
var
  Buffer, Es: string;
  CatLine:  TJPLCatLine;
  Line: TLine;
  Jp, J: TQuantumNumber;
  symp, np: Integer; { Upper State }
  sym,  n : Integer; { Lower State }
  Failed: Boolean;
  StateL, StateU: TState;
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  D: TMatrix;
  cstr: Double;

  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn('Line ', S.Line, ':', Name);
      WriteLn(Buffer);
    end;
    Failed := True;
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string; AddToLargestError: Boolean = True);
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else if Abs(n1-n2) > tol then
      Fail(Format('%s out of tolerance: %g %g %g %d %d %d %d %d %d', [Name, n1, n2, n1-n2, Jp, symp, np, J, sym, n]));
    if AddToLargestError and (Abs(n1-n2) > Result) then
      Result := Abs(n1-n2);
  end;

begin
  Result := 0;
  Molecule := Mixture.Species[0].Isotopomer[0];
  StateL := Molecule.Manifold[0].State[0];
  StateU := StateL;
  if Molecule.nManifolds > 1 then
    StateU := Molecule.Manifold[1].State[0];
  TransitionMoments := Molecule.FindTransitionMoments(StateU.Manifold, StateL.Manifold);
  while not S.AtEndOfFile do begin
    Buffer := S.RestOfLine;
    if LoadLine(buffer, CatLine) then begin
      Inc(nTried);
      AdjustQuantumNumbers(CatLine);
      Line.CachedName := Buffer;
      Failed := ConvertQuantumNumbers(StateU, Jp, symp, np, StateL, J, sym, n, Line.CachedName, CatLine);
      if Failed then
        Fail('Quantum number assignment')
      else begin
        Line.Width := 0;
        Line.Width := 0;
        Line.EigenvaluesU := StateU.Manifold.Eigenvalues(Jp, symp, [meHrot, meWidth]);
        if np <= 0 then
          Line.IndexU := FindIndex(Line.EigenvaluesU, Line.CachedName, False, S.Line)
        else
          Line.IndexU := np;
        if (Line.IndexU > Line.EigenvaluesU.Size) or (Line.IndexU <= 0) then
          Fail(Format('Upper eigenvalue (%d) > %d', [np, Line.EigenvaluesU.Size]));

        Line.EigenvaluesL := StateL.Manifold.Eigenvalues(J, sym, [meHrot, meWidth]);
        if n <= 0 then
          Line.IndexL := FindIndex(Line.EigenvaluesL, Line.CachedName, True, S.Line)
        else
          Line.IndexL := n;
        if (Line.IndexL > Line.EigenvaluesL.Size) or (Line.IndexL <= 0) then
          Fail(Format('Lower eigenvalue (%d) > %d', [n, Line.EigenvaluesL.Size]))
        else
          CheckNo(CatLine.ELO, Line.EigenvaluesL.E[Line.IndexL]*ConversionFactor(Mixture.Units,cm1)+Offset, Tolerance, 'E"');

        if not failed then begin
          Line.Frequency := Line.EigenvaluesU.E[Line.IndexU]-Line.EigenvaluesL.E[Line.IndexL];
          if CatLine.TAG >= 0 then
            CheckNo(CatLine.FREQ, Line.Frequency*ConversionFactor(Mixture.Units,MHz), Tolerance, 'Frequency')
          else
            CheckNo(CatLine.FREQ, Line.Frequency*ConversionFactor(Mixture.Units,MHz), 3*CatLine.ERR, 'Frequency', False);
          D := TransitionMoments.TransitionMatrix(Line.EigenValuesL, Line.EigenValuesU, nil);
          try
            Line.Strength := D[Line.IndexL, Line.IndexU]*Line.EigenvaluesL.StatisticalWeight;
            cstr := Mixture.LineIntensity(Line, Mixture.Temperature, Mixture.Tvib);
            if cstr <> 0 then
              Es := '(Ratio = ' + FloatToStr(Power(10,CatLine.LGINT)/cstr) + ') '
            else
              Es := '';
            CheckNo(CatLine.LGINT, log10(cstr), Tolerance, Es+'Intensity');
          finally
            D.Free;
          end;
          // if Power(10,CatLine.LGINT)/cstr > 1.1 then
            // WriteLn(Line.EigenvaluesU.E[Line.IndexU], ' ', Line.EigenvaluesL.E[Line.IndexL]);
          // WriteLn(Line.EigenvaluesL.J, Line.EigenvaluesL.E[Line.IndexL], ' ', Power(10,CatLine.LGINT)/cstr);
          // WriteLn(Line.Strength*1e12:8:4, ' ', Mixture.LineIntensity(Line, Mixture.Temperature, Mixture.Tvib)*1e12:8:4);
        end;
      end;
    end;
    S.NextLine;
  end;
end;

{ TJPLCatReader }

function TJPLCatReader.DoReadOneLine: Boolean;
var
  CatLine: TJPLCatLine;
  Buffer: string;
  StateL, StateU: TState;
begin
  Result := False;
  Buffer := StringOfChar(' ',Max(f.Column-Length(f.Token)-1,0)) + f.Token + f.RestOfLine;
  StateL := DefaultManifoldL.State[0];
  StateU := DefaultManifoldU.State[0];
  if LoadLine(buffer, CatLine) then begin
    AdjustQuantumNumbers(CatLine);
    Obs.Frequency := CatLine.FREQ;
    Obs.Weight := Abs(CatLine.ERR);
    Obs.LineNumber := f.Line;
    Obs.FileName := FileName;
    Obs.Colour := Obs.ManifoldL.Colour;
    if Obs.Colour = clNone then
      Obs.Colour := Obs.ManifoldU.Colour;
    Obs.Name := Buffer;
    if ConvertQuantumNumbers(StateU, Obs.JU, Obs.SymU, Obs.IndexU, StateL, Obs.JL, Obs.SymL, Obs.IndexL, Obs.Name, CatLine) then
      ReportError('Bad Quantum numbers')
    else
      case Take of
        tkObs:  Result := CatLine.TAG < 0;
        tkCalc: Result := CatLine.TAG >= 0;
        tkAll:  Result := True;
      end;
  end;
end;

end.

