unit Mixture;
{ Generic Types required for overall description of the spectrum of a mixture of compounds }

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics, Contnrs, Classes, SysUtils, DOM, Vector, MatrixVec, IntVec, Life,
     MNode, Units, QuantumNumber, DRIntensity;

const
  { Various special index values }
  TemperatureIndex = 3;
  GaussianIndex = 4;
  LorentzianIndex = 5;
  FoffsetIndex = 6;
  ShowSumIndex = 9;
  ShowPartsIndex = 10;
  ShowFortratIndex = 11;
  LSP = 3; { First unused parameter index in TState }
  LSI = 3; { First unused setting index in TState }
  LMI = 5; { First unused setting index in TMolecule }
  LPI = 2; { First unused setting index in TPerturbation }

type

  TEigenvalues = class;
  TBasisState = class;

  TLine = record
    Frequency: Double;
    Strength:  Single;
    Width:     Single;
    { May need to use TManifold here instead if memory usage too high }
    EigenvaluesU, EigenvaluesL: TEigenvalues;
    IndexU, IndexL: Integer;  { Index into Eigenvalues }
    Colour:    TColor;
    CachedName:string; { Not necessarily set; if blank call MakeTxName }
  end;

  TMixtureMNode = class(TMNode)
    function BoltzmannConstant: Double; { k in units of cm-1/K or MHz/K }
    function Population(Eigenvalues: TEigenvalues; i: Integer; Temperature: Double; Tvib: Double): Double;
    function LineIntensity(const Line: TLine; Temperature: Double; Tvib: Double): Double;
    function Precision: Integer;
    function DoubleQuantumNumbers: Boolean;
    function QuantumNumberToStr(QuantumNumber: TQuantumNumber): string;
  end;

  TStreamableMNode = class(TMixtureMNode)
  private
    procedure SetIntensityUnits(const Index: Integer; const AValue: TIntensityUnits);
    function GetIntensityUnits(const Index: Integer): TIntensityUnits;
  protected
    procedure SetUnits(const Index: Integer; const AValue: TResonanceUnits);
    function GetUnits(const Index: Integer): TResonanceUnits;
  public
    procedure LoadFromFile(AFileName: string); virtual;
    procedure LoadFromStream(Stream: TStream; AFileName: string = '<Stream>');
    procedure MergeFromStream(Stream: TStream; AFileName: string = '<Stream>');
    procedure SaveToFile(AFileName: string); virtual;
    class function OwnerType: TComponentClass; override;
    function UnitsName: string;
    function UnitsTypeName: string;
  published
    property IntensityUnits: TIntensityUnits index 2 read GetIntensityUnits write SetIntensityUnits;
  end;

  TBasisOrder = (boDefault, boDiagonal, boStateDiagonal); { How to order the basis states }

  TMixture = class;     { Note logical hierarchy }
   TSpecies = class;
    TMolecule = class;
     TManifold = class;
      TState = class;

  TLineList = class;

  TSpectrum = class
  private
    FColor: TColor;
    FUnits: TResonanceUnits;
    FHeadingPrinted: Boolean;
    procedure SetHeadingPrinted(const Value: Boolean);
  protected
    FnLines: Integer;
  public
    FMin: Double;
    FMax: Double;
    Temperature: Double;
    Tvib: Double;
    ITotal: Double; { TODO: Check/automate usage - this should depend on temperature }
    constructor Create(AUnits: TResonanceUnits);
    procedure Clear; virtual;
    procedure AddLine(Frequency, Strength, Width: Double; Colour: TColor; AUnits: TResonanceUnits); overload; virtual;
    procedure AddLine(const Line: TLine; Mixture: TMixture); overload; virtual; abstract;
    procedure AddIn(LineList: TLineList; Mixture: TMixture);
    procedure ShowFooter(var F: TextFile; Separator: Char);
    procedure ShowLine(var Line: TLine; var F: TextFile; Separator: Char; Threshold: Double; LinFormat: Boolean);
    property nLines: Integer read FnLines;
    property HeadingPrinted: Boolean read FHeadingPrinted write SetHeadingPrinted; { If heading for transtion list has been printed; clear to indicate scan though line list }
    property Colour: TColor read FColor write FColor;
    property Units: TResonanceUnits read FUnits write FUnits;
  end;

  TLineList = class(TSpectrum)
  private
    FLines: array of TLine;
    function GetLine(AIndex: Integer): TLine;
  public
    Selected: Boolean;
    procedure AddLine(const Line: TLine; Mixture: TMixture); override;
    procedure Clear; override;
    function LargestIntensity: Double;
    {$IFNDEF CONSOLE}
    procedure FindLines(Centre, Range: Double; AUnits: TResonanceUnits);
    {$ENDIF}
    procedure ShowLine(LineNumber: Integer; var F: TextFile; Separator: Char; Threshold:Double; LinFormat: Boolean); overload;
    property Line[AIndex: Integer]: TLine read GetLine;
  end;

  TTransitionMoment = class(TMNodePair)
  private
    FMagnetic: Boolean;
  public
    function BraOwner: TMNode; override;
    function KetOwner: TMNode; override;
    class function BraKetClass: TMNodeClass; override;
    class function OwnerType: TComponentClass; override;
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create(AOwner: TComponent; State1Name, State2Name: string); overload;
    function Allowed(Sym1, Sym2: Integer): Boolean;
    function MatrixElement(ABra, AKet: TBasisState): Double; virtual;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    property Magnetic: Boolean read FMagnetic;
    function nChildTypes: Integer; override;
  published
    property Rank:      Integer index 1 read GetSetting write SetSetting;
    property Strength:  Double index 1 read GetParameter write SetParameter;
    property C:         Double index 2 read GetParameter write SetParameter;
    property D:         Double index 3 read GetParameter write SetParameter;
  end;

  TSphericalTransitionMoment = class(TTransitionMoment)
  public
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Component: Integer index 2 read GetSetting write SetSetting;
  end;

  { TMagneticTransitionMoment }

  TMagneticTransitionMoment = class(TSphericalTransitionMoment)
  public
    function PairName: string; override;
    constructor Create(AOwner: TComponent); override;
  end;

  { TOrbitalTransitionMoment }

  TOrbitalTransitionMoment = class(TMagneticTransitionMoment)
    function PairName: string; override;
  end;

  { TSpinTransitionMoment }

  TSpinTransitionMoment = class(TMagneticTransitionMoment)
  public
    function PairName: string; override;
    constructor Create(AOwner: TComponent); override;
  end;

  { TRotationalTransitionMoment }

  TRotationalTransitionMoment = class(TMagneticTransitionMoment)
  public
    function PairName: string; override;
  end;
  
  TabcAxis = (aAxis=0, bAxis, cAxis);
  TxyzAxis = (zAxis=0, yAxis, xAxis);

  TCartesianTransitionMoment = class(TTransitionMoment)
  public
    function PairName: string; override;
  private
    function GetAxis(const Index: Integer): TabcAxis;
    procedure SetAxis(const Index: Integer; const AValue: TabcAxis);
  public
    constructor Create(AOwner: TComponent); override;
    function FixupProperties:Exception; override;
    function GetSettingsNames(AIndex: Integer; var LowestValue:Integer): PPChar; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Axis: TabcAxis index 2 read GetAxis write SetAxis;
  end;

  TTransitionMoments = class(TMNodePair)
  private
    FnRanks: Integer;
    function GetTransitionMoment(AIndex: Integer): TTransitionMoment;
  protected
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
  public
    class function BraKetClass: TMNodeClass; override;
    class function OwnerType: TComponentClass; override;
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create(AOwner: TComponent; Manifold1Name, Manifold2Name: string); overload;
    function FixupProperties:Exception; override;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure AddToSpectrum(Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; TransitionMoment: TTransitionMoment; PrintLevel: Integer; const Excitation: TExcitation);
    procedure AddMatrixToSpectrum(D: TMatrix; Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; PrintLevel: Integer; Excitation: TExcitation);
    function OriginalTransitionMatrix(Ev1, Ev2: TEigenvalues; Rank: TQuantumNumber; ATransitionMoment: TTransitionMoment): TMatrix;
    function TransitionMatrix(Ev1, Ev2: TEigenvalues; TransitionMomentSel: TTransitionMoment; PrintLevel: Integer = -10): TMatrix;
    procedure TransformTransitionMatrix(D: TMatrix; Ev1, Ev2: TEigenvalues);
    function Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean; virtual;
    function PairName: string; override;
    function MaxRank: TQuantumNumber;
    function Mixture: TMixture;
    function Molecule: TMolecule;
    function nChildTypes: Integer; override;
    { TODO: Add checks in loader for duplicate or erroneous TransitionMoment(s) }
    property nRanks: Integer read FnRanks;
    property TransitionMoment[AIndex: Integer]: TTransitionMoment read GetTransitionMoment;
  published
    property Colour: TColor index 1 read GetColour write SetColour;
  end;

  TNucleus = class(TMNode)
  private
    FCreating: Boolean;
    procedure SetSpin(const Index, Value: Integer);
    function GetSpin(const Index: Integer): Integer;
  protected
    function GetName: TComponentName; override;
    procedure SetName(const Value: TComponentName); override;
  public
    constructor Create(AOwner: TComponent); override;
    class function OwnerType: TComponentClass; override;
    function Molecule: TMolecule;
    function HelpOnSetting(AIndex: Integer): string; override;
    function nChildTypes: Integer; override;
  published
    property Spin: Integer index 1 read GetSpin write SetSpin;
  end;

  TNucleusClass = class of TNucleus;

  THyperfineBasis = array of TQuantumNumber;

  TMolecule = class(TMixtureMNode) { Isotopomer }
  private
    FLongestManifoldName: string;
    FnManifolds: Integer;
    FnTransitionMoments: Integer;
    FOddSpin, FMaxRank, FMaxNuclearSpin: TQuantumNumber;
    FLowestOrigin: Double;
    FLastPartitionFunction, FLastTemperature, FLastTvib: Double;
    function GetManifold(AIndex: Integer): TManifold;
    function GetTransitionMomentsFunction(AIndex, BIndex: Integer): TTransitionMoments;
    function GetMixture: TMixture;
    procedure SetnNuclei(const Index, Value: Integer);
  protected
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
    procedure SetModified(const Value: Boolean); override;
  public
    Nuclei: array of TNuclearInfo;
    HyperfineBasisStates: array of THyperfineBasis;
    constructor Create(AOwner: TComponent); override;
    class function OwnerType: TComponentClass; override;
    class function ManifoldType: TMNodeClass; virtual; abstract;
    procedure ReadProperties(Node: TDOMNode); override;
    procedure ClearCache; override;
    function FixupProperties:Exception; override;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function GetTransitionMoments(AIndex: Integer): TTransitionMoments;
    property OddSpin: TQuantumNumber read FOddSpin;
    function Species: TSpecies;
    function AddToSpectrum(Spectrum: TSpectrum; PrintLevel: Integer;  const Excitation: TExcitation;
                            J1Sel: Integer = -1; Sym1Sel: Integer = -1; J2Sel: Integer = -1; Sym2Sel: Integer = -1;
                            Manifold1Sel: TManifold = nil; Manifold2Sel: TManifold = nil;
                            DeltaJ: TQuantumNumber = -10000;
                            TransitionMomentsSel: TTransitionMoments = nil; TransitionMomentSel: TTransitionMoment = nil;
                            State1Sel: TState = nil; State2Sel: TState = nil; IncludePerturbations: Boolean = True):Boolean;
    function CalculatePartitionFunction(Temperature, Tvib: Double; Print: Boolean = false; ShowMolecule: Boolean = false; ShowHeader: Boolean = false):Double;
    function PartitionFunction(Temperature, Tvib: Double):Double; { As above, but with cached result }
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; virtual; abstract;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; virtual; abstract;
    function JDeAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; virtual;
    function LastSym: Integer; virtual; abstract;
    function SymName(Sym: Integer): string; virtual;
    function JAdjustedSymName(J: TQuantumNumber; Sym: Integer): string; virtual;
    property LongestManifoldName: string read FLongestManifoldName; { For formatting only }
    function ManifoldWithName(AName: string): TManifold;
    function ManifoldStartingWithName(AName: string): TManifold;
    property LowestOrigin: Double read FLowestOrigin;
    property nManifolds: Integer read FnManifolds;
    property Manifold[AIndex: Integer]: TManifold read GetManifold;
    property MaxRank: TQuantumNumber read FMaxRank;
    property Mixture: TMixture read GetMixture;
    function FindTransitionMoments(ManifoldA, ManifoldB: TManifold):TTransitionMoments;
    property TransitionMoments[AIndex, BIndex:Integer]: TTransitionMoments read GetTransitionMomentsFunction;
    property nTransitionMoments: Integer read FnTransitionMoments;
    property MaxNuclearSpin: TQuantumNumber read FMaxNuclearSpin;
    function nChildTypes: Integer; override;
  published
    property Abundance: Double index 1 read GetParameter write SetParameter;
    property nNuclei:   Integer index 1 read GetSetting write SetnNuclei;
    property Jmin:      Integer index 2 read GetSetting write SetSetting;
    property Jmax:      Integer index 3 read GetSetting write SetSetting;
    property Colour:    TColor  index 4 read GetColour write SetColour;
  end;

  TMoleculeClass = class of TMolecule;
  
  { TBasisState }

  TBasisState = class(TComponent)
  public
    State: TState;
    StateIndex: Integer; { For help with eigenvalue searches }
    HyperfineState: THyperfineBasis;
    J: TQuantumNumber;
    Phase: Integer;
    function F: TQuantumNumber;
    function Sym: Integer;
    function TrackIndex:Integer; virtual; { A number >= 0 for tracking related states, mainly in energy level plots. }
    function BasisLabel: string; virtual; { Basis State Label }
    function BasisLabelKey: string; virtual; { Explanation of Basis State Label }
    function StateLabel: string; virtual; { State name that goes with the given basis state
                                            (assuming match between basis state and final state is OK) }
    function StateLabelKey: string; virtual; { Explanation of State Name }
    function HyperfineStateLabel: string;
    function HyperfineStateLabelKey: string;
    function TxName(Upper, Lower: TBasisState): string; virtual;    { Make transition label such as R(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; virtual; { String explaining transition label }
  end;

  { Type of matrix elements required }
  TMatElemType = (meHrot {Normal Hamiltonian}, meWidth {Operator for lifetime});
  TMatElemTypes = set of TMatElemType;
  TDArray = array of Double;

  TEigenvalues = class(TComponent)
  private
    FE: TColumnVector;
    FWidths: TColumnVector;
    FC: TMatrix;
    FDominantBasisIndex: TIntColumnVector; { Dominant basis state for each eigenvalue }
    // function GetBasisState(AIndex: Integer): TBasisState;
    function GetDominantBasisState(AIndex: Integer): TBasisState; {$IFDEF INLINE} inline; {$ENDIF}
    procedure SetC(const Value: TMatrix);
    procedure SetE(const Value: TColumnVector);
    procedure SetWidths(const Value: TColumnVector);
  public
    Manifold: TManifold;
    J: TQuantumNumber;
    Sym: Integer;
    StatisticalWeight: Integer;
    LastStateIndex: Integer;
    SelectedState: TState;  { Eigenvalues calculated with this State only in basis }
    NoPerturbations: Boolean; { Eigenvalues calculated excluding perturbations }
    Basis: array of TBasisState;
    function Size: Integer;
    function MatrixElements(MatElemType: TMatElemType; IncludePerturbations: Boolean = True): TMatrix;
    procedure DoEigenSearch;
    procedure DoLimitedEigenSearch;
    function DiagonalMatrixElements(MatElemType: TMatElemType; Base: Integer = 0; IncludePerturbations: Boolean = False): TDArray;
    procedure Diagonalize(EigenvectorsWanted: Boolean);
    procedure Fixup;
    procedure SortBasisOn(V: TDarray; ByState: Boolean; Base: Integer = 0);
    procedure SortFrom(Start: Integer);
    procedure Multiply(M: TMatrix);
    procedure MakeHamiltonian(BasisOrder: TBasisOrder; IncludePerturbations: Boolean = True);
    procedure ShowBasisMatrix(M: TMatrix; const Title: string);
    procedure ShowTransitionMatrix(Ev: TEigenValues; M: TMatrix; const Title: string; ShowBasis: Boolean = True);
    procedure ShowEigenvalues;
    function Transform(M: TMatrix): TColumnVector;
    destructor Destroy; override;
    // property Basis[AIndex: Integer]: TBasisState read GetBasisState;
    property DominantBasis[AIndex: Integer]: TBasisState read GetDominantBasisState;
    property E: TColumnVector read FE write SetE;
    property C: TMatrix read FC write SetC;
    property Widths: TColumnVector read FWidths write SetWidths;
  end;

  { TPerturbation }

  TPerturbation = class(TMNodePair)
  public
    NuclearRank: TQuantumNumber; { Rank of nuclear spin operator }
    class function BraKetClass: TMNodeClass; override;
    class function OwnerType: TComponentClass; override;
    function Connects(State1, State2: TState): Boolean;
    function MatrixElement(State1, State2: TBasisState; MatElemType: TMatElemType): Double;
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function nChildTypes: Integer; override;
  published
    property Value:  Double index 1 read GetParameter write SetParameter;
    { Index, starting from 1, of nuclear spin involved in perturbation
      0 (default) for those not involving a nuclear spin }
    property Nucleus: Integer index LPI-1 read GetSetting write SetSetting;
  end;

  TMatrixElementFunction = function (A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer):Double of object;
  TSymCheck = (MustEqual, MayCancel, NoCheck, FullSum);

  { TState }

  TState = class(TMNode)
  private
    FMolecule: TMolecule;
  protected
    FFullRotationalSymmetry: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    function OddSpin: TQuantumNumber; virtual; abstract;
    function AddBasis(Eigenvalues: TEigenvalues): Integer; { Returns Statistical Weight }
    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer; virtual; abstract;
    function AddNucleus: TNucleus; virtual;
    class function NucleusType: TNucleusClass; virtual; abstract;
    procedure AdjustNuclei;
    function FixupProperties:Exception; override;
    function nChildTypes: Integer; override;
    function HasWidth: Boolean; virtual;

    { Functions working on symmetrized basis functions }
    function MatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Double; { virtual; }
    function SymmetrizedRotationalMatrixElement(A, B: TBasisState; MNode: TMnode; Flag: Integer; MatrixElementFunction: TMatrixElementFunction; SymCheck: TSymCheck = MustEqual): Double; { virtual; }
    function SymmetrizedHyperfineMatrixElement(A, B: TBasisState; F, Rank: TQuantumNumber; Nucleus: TNucleus;
                                    MatrixElementFunction: TMatrixElementFunction; MNode: TMNode): Double; { virtual; }
    { Note that transition moment matrix elements can be implemented in TState or TransitionMoment descendents }
    function TransitionMatrixElement(TransitionMoment: TTransitionMoment; A, B: TBasisState): Double; { virtual; }
    function PerturbationMatrixElement(Perturbation: TPerturbation; A, B: TBasisState; MatElemType: TMatElemType): Double; { virtual; }

    { Functions working on unsymmetrized basis functions }
    // function RotSpinMatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Double; virtual; abstract;
    function RotationalMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Double; virtual; abstract;
    function RotationalTransitionMatrixElement(A, B: TBasisState; SignA, SignB: Integer; TransitionMoment: TMNode; F: TQuantumNumber):Double; virtual; abstract;
    function RotationalPerturbationMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Double; virtual; abstract;
    function HyperfineMatrixElement(A, B: TBasisState; Signp, Sign: Integer; MNode: TMnode; Flag: Integer): Double; virtual; abstract;

    procedure ReadProperties(Node: TDOMNode); override;
    procedure ConvertToPerturbations; virtual;
    function Manifold: TManifold;
    function FindColour: TColor;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function StateOrigin: Double; virtual; { Default is just Origin, but ideally choose lowest J=0 energy }
                                           { Used to stop Boltzman factors blowing up at low temperatures }
    function Nucleus(Index:Integer): TNucleus;
    property Molecule: TMolecule read FMolecule;
    property FullRotationalSymmetry: Boolean read FFullRotationalSymmetry; { If point group symmetry used is not a subgroup of the rotational Hamiltonian }
  published
    property Origin: Double index 1 read GetParameter write SetParameter;
    property Width:  Double index 2 read GetParameter write SetParameter;
    property Colour: TColor index 1 read GetColour write SetColour;
    property S: TQuantumNumber index 2 read GetSetting write SetSetting; { Spin }
  end;

  { TManifold }

  TManifold = class(TMixtureMNode)
  private
    FEigenValueCache: TComponentList;
    FHasWidth: Boolean;
    FnPerturbations: Integer;
    FnStates: Integer;
    FLowestStateOrigin: double;
    FMaxSpin: TQuantumNumber;
    function GetState(AIndex: Integer): TState;
    function GetPerturbation(AIndex: Integer): TPerturbation;
  protected
    procedure SetModified(const Value: Boolean); override;
  public
    property HasWidth: Boolean read FHasWidth;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function OddSpin: TQuantumNumber;
    procedure ClearCache; override;
    function Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
                         PrintLevel: Integer = -10; SelState: TState = nil; IncludePerturbations: Boolean = True): TEigenvalues;
    function FindJmin(Default: Integer):Integer;
    function FindJmax(Default: Integer):Integer;
    function FirstSym(J: TQuantumNumber):Integer; virtual; abstract;
    function FixupProperties:Exception; override;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function NextSym(J: TQuantumNumber; Sym: Integer):Integer; virtual; abstract;
    function nSym: Integer; virtual; abstract; { Number of different symmetries }
    function MakeBasis(J: TQuantumNumber; Sym: Integer; AState: TState = nil): TEigenvalues;
    function Mixture: TMixture;
    function Molecule: TMolecule;
    // function PartitionFunction(Temperature: Double):Double;
    procedure ConvertToPerturbations;
    function StateWithName(AName: string): TState;
    property nStates: Integer read FnStates;
    property State[AIndex: Integer]: TState read GetState;
    property LowestStateOrigin: double read FLowestStateOrigin;
    function nChildTypes: Integer; override;
    property nPerturbations: Integer read FnPerturbations;
    property Perturbation[AIndex: Integer]: TPerturbation read GetPerturbation; default;
    property MaxSpin: TQuantumNumber read FMaxSpin;
  published
    property Jmin:             Integer index 1 read GetSetting write SetSetting;
    property Jmax:             Integer index 2 read GetSetting write SetSetting;
    property Initial:          Boolean index 3 read GetFlag write SetFlag;
    property Colour:           TColor  index 4 read GetColour write SetColour;
    property EigenSearch:      Boolean index 5 read GetFlag write SetFlag;
    property LimitSearch:      Boolean index 6 read GetFlag write SetFlag;
    { True to show states as eigenvalue number, symmetry rather than any more meaningful label }
    { TODO: property ShowStateNumbers: Boolean index 7 read GetFlag write SetFlag; }
  end;

  TSpecies = class(TMNode)
  private
    function GetIsotopomer(AIndex: Integer): TMolecule;
  protected
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
  public
    constructor Create(AOwner: TComponent); override;
    class function OwnerType: TComponentClass; override;
    class procedure AddMoleculeType(MoleculeClass: TMNodeClass);
    function AddToSpectrum(Spectrum: TSpectrum): Boolean;
    function FindMolecule(AName: string): TMolecule;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    property Isotopomer[AIndex: Integer]: TMolecule read GetIsotopomer;
    function nChildTypes: Integer; override;
  published
    property Concentration: Double index 1 read GetParameter write SetParameter;
    property Jmin:      Integer index 1 read GetSetting write SetSetting;
    property Jmax:      Integer index 2 read GetSetting write SetSetting;
    property Colour:    TColor  index 3 read GetColour write SetColour;
  end;

  TParameter = record
    MNode: TMNode;
    Index: Integer;
    Value: Double;
    Increment: Double;
    Name: string;
  end;

  TFitType = (ftLine, ftContour);

  TParameters = array of TParameter;

  TSavedParameters = record
    FitType: TFitType;
    Parameters: TParameters;
  end;

  { TMixture }

  TMixture = class(TStreamableMNode)
  private
    FnMolecules: Integer;
    FSavedParameters: array of TSavedParameters;
    FUndoPosition: Integer;
    FLineList: TLineList; { Index within the above }
    FLongestMoleculeName: string; { For formatting }
    FLongestManifoldName: string; { For formatting }
    function GetSpecies(AIndex: Integer): TSpecies;
    procedure SetLifeModel(const Index: Integer; const AValue: TLifeModel);
    function GetLifeModel(const Index: Integer): TLifeModel;
    procedure SetBasisOrder(const Index: Integer; const AValue: TBasisOrder);
    function GetBasisOrder(const Index: Integer): TBasisOrder;
    procedure SetLineList(const Value: TLineList);
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
    procedure SetModified(const Value: Boolean); override;
    procedure LoadCurrentSavedParameters;
  protected
    procedure SetEUnits(const Index: Integer; const AValue: TEnergyUnits);
    function GetEUnits(const Index: Integer): TEnergyUnits;
  public
    FittingParameters: TParameters;
    HasPlotted: Boolean;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddNodeParameters(MNode: TMNode);
    procedure FillParameters; { Set up parameters array }
    procedure PushParameters(FitType: TFitType);
    procedure DiscardFit;
    procedure UndoFit;
    procedure ClearUndo;
    function CanUndoFit: Boolean;
    procedure RedoFit;
    function CanRedoFit: Boolean;
    procedure ChangePlotUnits(NewPlotUnits: TResonanceUnits);
    procedure CycleUnits; override;
    function FixupProperties:Exception; override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function MakeLineList: TLineList;
    procedure PreCalculatePartitionFunctions;
    procedure ConvertToPerturbations;
    property LineList: TLineList read FLineList write SetLineList;
    property Species[AIndex: Integer]: TSpecies read GetSpecies; default;
    function FindMolecule(AName: string): TMolecule;
    property nMolecules: Integer read FnMolecules;
    property LongestMoleculeName: string read FLongestMoleculeName; { For formatting }
    property LongestManifoldName: string read FLongestManifoldName; { For formatting }
  published
    property Units:         TEnergyUnits index 1 read GetEUnits write SetEUnits;
    property PrintLevel:    Integer index 3 read GetSetting write SetSetting;
    property nDF:           Integer index 4 read GetSetting write SetSetting;
    property WidthMult:     Integer index 5 read GetSetting write SetSetting;     { Multiple of linewidth to extend convolution over }
    property Precision:     Integer index 6 read GetSetting write SetSetting;     { Precision (of printout) }
    property LifeModel:     TLifeModel index 7 read GetLifeModel write SetLifeModel;
    property BasisOrder:    TBasisOrder index 8 read GetBasisOrder write SetBasisOrder;
    property ShowSum:       Boolean index ShowSumIndex read GetFlag write SetFlag;
    property ShowParts:     Boolean index ShowPartsIndex read GetFlag write SetFlag;
    property ShowFortrat:   Boolean index ShowFortratIndex read GetFlag write SetFlag;
    property UseUpper:      Boolean index ShowFortratIndex+1 read GetFlag write SetFlag;
    property JAdjustSym:    Boolean index 13 read GetFlag write SetFlag;
    property ShowSymmetry:  Boolean index 14 read GetFlag write SetFlag; { ... in fortrat plots }
    property ShowDeltaJ:    Boolean index 15 read GetFlag write SetFlag; { ... in fortrat plots }
    property ScaleMarkSize: Boolean index 16 read GetFlag write SetFlag; { ... in fortrat plots }
    property DoubleQnos:    Boolean index 17 read GetFlag write SetFlag; { How to display quantum numbers }
    property FitCycles:     Integer index 18 read GetSetting write SetSetting;
    property AutoReplot:    Boolean index 19 read GetFlag write SetFlag; { Turn off if replot is rather slow }
    property PlotUnits:     TResonanceUnits index 20 read GetUnits write SetUnits;

    property Fmin:        Double index  1 read GetParameter write SetParameter;
    property Fmax:        Double index  2 read GetParameter write SetParameter;
    property Temperature: Double index TemperatureIndex read GetParameter write SetParameter;
    property Gaussian:    Double index GaussianIndex read GetParameter write SetParameter;
    property Lorentzian:  Double index LorentzianIndex read GetParameter write SetParameter;
    property Foffset:     Double index FoffsetIndex read GetParameter write SetParameter;
    property SMargin:     Double index  7 read GetParameter write SetParameter;
    { Relative Size of peak to ignore when convoluting for contour fitting }
    property OThreshold:  Double index  8 read GetParameter write SetParameter;
    property RefWidth:    Double index  9 read GetParameter write SetParameter;
    property Tvib:        Double index 10 read GetParameter write SetParameter;
    property MinI:        Double index 11 read GetParameter write SetParameter; { as a fraction of maximum intensity; For Fortrat plots; TODO: Use elsewhere }
    { TODO: Some of these should probably be settable at the Molecule or Manifold level }
  end;

function NotYet:Exception;

procedure ConvertLine(var Line: TLine; UFrom, UTo: TResonanceUnits);
function MakeTxName(var Line: TLine): string;
function AxisName(AAxis: TabcAxis): string;

implementation

uses {$IFNDEF CONSOLE} LineFrm, Main, {$ENDIF}
     {$IFDEF EXPLICITTEST} ExplicitTest, {$ENDIF}
     Progress, nj, Math, MathSup, Eig, XMLWrite, XMLRead, Fitter, Profile, Huge, ShellSort;

function NotYet:Exception;
begin
  Result := Exception.Create('Not Yet Implemented');
end;

{ Profiling Variables }
var
 AddMatrixToSpectrumTime: string = 'AddMatrixToSpectrum';
 EigenvaluesTime: string = 'Make and Diagonalize Hamiltonian';
 MakeHamiltonianTime: string = 'Make Hamiltonian Matrix';
 PerturbationMatrixElementTime: string = 'Perturbation Matrix Elements';
 EigenSearchTime: string = 'Identifying eigenvalues';
 EigenvaluesSearchTime: string = 'EigenvaluesSearch';
 MakeLineListTime: string = 'MakeLineList';
 AddInLinesTime: string = 'AddInLines';
 TransitionMatrixTime: string = 'Make and Transform Transition Matrix';
 MakeTransitionMatrixTime: string = 'Make Transition Matrix';
 MatrixMultiplyTime: string = 'MatrixMultiply';
 PostMultiplyTime: string = 'PostMultiply';
 TxNameTime: string = 'Transition Name';
 t1: string = 't1';

const
  ColourHelpString = 'Colour - set to "None" to take value from elsewhere.';

{ TLine }

function MakeTxName(var Line: TLine): string;
begin
  TStart(TxNameTime);
  with Line do
    if (Length(CachedName) = 0) and Assigned(EigenvaluesU) and Assigned(EigenvaluesL) then begin
      Result := EigenvaluesU.Basis[1].TxName(EigenvaluesU.DominantBasis[IndexU], EigenvaluesL.DominantBasis[IndexL]);
      if Length(Result) > 0 then
        Result := Result + ' : ';
      if EigenvaluesL.Manifold.Mixture.nMolecules > 1 then
        Result := Result + EigenvaluesL.Manifold.Molecule.Name + ' ';
      Result := Result + EigenvaluesU.DominantBasis[IndexU].StateLabel
                + ' - ' + EigenvaluesL.DominantBasis[IndexL].StateLabel;
      CachedName := Result;
    end else
      Result := CachedName;
  TEnd(TxNameTime);
end;

procedure ConvertLine(var Line: TLine; UFrom, UTo: TResonanceUnits);
begin
  if UFrom <> UTo then begin
    Line.Width := ConvertWidth(Line.Width, Line.Frequency, UFrom, UTo);
    Line.Frequency := Convert(Line.Frequency, UFrom, UTo);
  end;
end;

{ TManifold }

function TManifold.GetState(AIndex: Integer): TState;
var
  i: Integer;
begin
  { TODO: Speed up? }
  for i := 0 to ComponentCount-1 do
    if Items[i] is TState then begin
      Dec(AIndex);
      if AIndex < 0 then begin
        Result := TState(Items[i]);
        Exit;
      end;
    end;
  raise Exception.Create('TManifold.GetState');
end;

function TManifold.GetPerturbation(AIndex: Integer): TPerturbation;
var
  i: Integer;
begin
  { TODO: Speed up? }
  for i := 0 to ComponentCount-1 do
    if Items[i] is TPerturbation then begin
      Dec(AIndex);
      if AIndex < 0 then begin
        Result := TPerturbation(Items[i]);
        Exit;
      end;
    end;
  raise Exception.Create('TManifold.GetPerturbation');
end;

function TManifold.StateWithName(AName: string): TState;
var
  i: Integer;
begin
  Result := nil;
  { NB: can't use FindComponent as Name property has been overridden }
  for i := 0 to ComponentCount-1 do
    if (Items[i] as TState).Name = AName then begin
      Result := TState(Items[i]);
      Exit;
    end;
end;

constructor TManifold.Create(AOwner: TComponent);
begin
  FEigenValueCache := nil;
  SavedProperties := 'Colour=None,Jmin=-1,Jmax=-1,EigenSearch=True,' + SavedProperties;
  inherited Create(AOwner);
end;

destructor TManifold.Destroy;
begin
  ClearCache;
  inherited Destroy;
end;

function TManifold.OddSpin: TQuantumNumber;
begin
  Result := State[0].OddSpin;
end;

procedure TManifold.ClearCache;
begin
  inherited;
  {$IFNDEF CONSOLE}
  if Assigned(LineForm) then
    LineForm.ClearTLines;
  {$ENDIF}
  FreeAndNil(FEigenValueCache);
end;

function TManifold.MakeBasis(J: TQuantumNumber; Sym: Integer; AState: TState): TEigenvalues;
var
  i, wt: Integer;
begin
  Result := TEigenvalues.Create(nil);
  try
    if (Sym < 0) or (Sym >= nSym) then
      raise Exception.CreateFmt('%s: Invalid symmetry number %d', [Name, Sym]);
    Result.J := J;
    Result.Sym := Sym;
    Result.StatisticalWeight := 0;
    Result.Manifold := Self;
    Result.SelectedState := AState;
    Result.LastStateIndex := 0;
    if Assigned(AState) then begin
      { If asked, just use specified state }
      if AState.Manifold <> Self then
        raise Exception.Create('State does not belong to manifold');
      Result.StatisticalWeight := AState.AddBasis(Result);
    end else begin
      { Make up Basis }
      for i := 0 to nStates-1 do begin
        wt := State[i].AddBasis(Result);
        if i = 0 then
          Result.StatisticalWeight := wt
        else if wt <> Result.StatisticalWeight then
          raise Exception.CreateFmt('Inconsistent Statistical Weights for %s - %d, %s=%d', [Name, Result.StatisticalWeight, State[i].Name, wt]);
      end;
    end;
    Result.Fixup;
  except
    Result.Free;
    raise;
  end;
end;

function TManifold.Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes; PrintLevel: Integer; SelState: TState; IncludePerturbations: Boolean): TEigenvalues;
var
  i: Integer;
  M: TMatrix;
  InCache: Boolean;
begin
  if PrintLevel = -10 then
    PrintLevel := Mixture.PrintLevel;
  { Turn off selection if it makes no difference }
  if nStates = 1 then
    SelState := nil;
  if nPerturbations = 0 then
    IncludePerturbations := True;
  TStart(EigenvaluesSearchTime);
  { Have we already Calculated it ? }
  if not Assigned(FEigenValueCache) then
    FEigenValueCache := TComponentList.Create;
  { TODO: Optimize Search }
  InCache := False;
  for i := 0 to FEigenValueCache.Count - 1 do begin
    Result := TEigenValues(FEigenValueCache[i]);
    if Assigned(Result) and (Result.J = J) and (Result.Sym = Sym)
       and (Result.SelectedState = SelState) and (Result.NoPerturbations <> IncludePerturbations) then begin
      if (Result.Size > 0) and not Assigned(Result.C) and (MatElemTypes <> []) then begin
        { In cache, but not with eigenvectors that are now needed }
        InCache := True;
        Break;
      end else begin
        TEnd(EigenvaluesSearchTime);
        Exit;
      end;
    end;
  end;
  TEnd(EigenvaluesSearchTime);
  TStart(EigenvaluesTime);
  if not InCache then begin
    Result := MakeBasis(J, Sym, SelState);
  end;
  try
    { Make up Matrix }
    if Result.Size > 0 then begin
      TStart(MakeHamiltonianTime);
      Result.MakeHamiltonian(Mixture.BasisOrder, IncludePerturbations);
      TEnd(MakeHamiltonianTime);
      if PrintLevel > 2 then begin
        WriteLn;
        Result.ShowBasisMatrix(Result.C, 'Hamiltonian Matrix for ');
      end;
      { TODO: Consider this more carefully - when are eigenvalues really wanted? See also cache usage above }
      Result.Diagonalize((MatElemTypes <> []) or EigenSearch or (PrintLevel > 1));
      {$IFDEF EXPLICITTEST}
      DoExplicitTest(Result);
      {$ENDIF}
      if EigenSearch then begin
        TStart(EigenSearchTime);
        if LimitSearch then
          Result.DoLimitedEigenSearch
        else
          Result.DoEigenSearch;
        TEnd(EigenSearchTime);
      end;
      if (MatElemTypes = []) and (PrintLevel <= 1) then
        Result.C := nil
      else begin
        if HasWidth and (meWidth in MatElemTypes) then begin
          M := Result.MatrixElements(meWidth);
          try
            if PrintLevel > 2 then
              Result.ShowBasisMatrix(M, 'Width Matrix for ');
            Result.Widths := Result.Transform(M);
          finally
            M.Free;
          end;
        end;
        if PrintLevel > 1 then
          Result.ShowEigenvalues;
      end;
    end;
  except
    Result.Free;
    raise;
  end;
  if not InCache then
    FEigenValueCache.Add(Result); { TODO: Locking? }
  TEnd(EigenvaluesTime);
end;

function TManifold.FindJmin(Default: Integer): Integer;
begin
  Result := Jmin;
  if Result < 0 then
    Result := Molecule.Jmin;
  if Result < 0 then
    Result := Molecule.Species.Jmin;
  if Result < 0 then
    Result := Default;
  if Odd(OddSpin+Result) then
    Dec(Result);
  if Result < OddSpin then
    Result := OddSpin
end;

function TManifold.FindJmax(Default: Integer): Integer;
begin
  Result := Jmax;
  if Result < 0 then
    Result := Molecule.Jmax;
  if Result < 0 then
    Result := Molecule.Species.Jmax;
  if Result < 0 then
    Result := Default;
end;

function TManifold.Molecule: TMolecule;
begin
  Result := Owner as TMolecule;
end;

procedure TManifold.ConvertToPerturbations;
var
  i: Integer;
begin
  for i := 0 to nStates-1 do
    State[i].ConvertToPerturbations;
end;

{
function TManifold.PartitionFunction(Temperature: Double): Double;
var
  J: TQuantumNumber;
  Sym, i: Integer;
  Ev: TEigenvalues;
  Sum: Double;
begin
  Result := 0;
  J := OddSpin;
  JMax := FindJmax(100);
  repeat
    Sum := 0;
    for Sym := 0 to nSym-1 do begin
      Ev := Eigenvalues(J, Sym, [], 0);
      for i := 1 to Ev.Size do
        Sum := Sum + 1; // TODO : Boltzman
    end;
    Result := Result + Sum;
    Inc(J,2);
  until J > JMax;
end;
}

function TManifold.Mixture: TMixture;
begin
  Result := Owner.Owner.Owner as TMixture;
end;

procedure TManifold.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache;
end;

function TManifold.FixupProperties:Exception;
var
  i: Integer;
begin
  Result := inherited FixupProperties;
  FnStates := 0;
  FnPerturbations := 0;
  FHasWidth := False;
  FMaxSpin := 0;
  for i := 0 to ComponentCount-1 do
    if Components[i] is TState then
      with TState(Components[i]) do begin
        if HasWidth then
          FHasWidth := True;
        if FnStates = 0 then
          FLowestStateOrigin := StateOrigin
        else
          FLowestStateOrigin := Min(FLowestStateOrigin, StateOrigin);
        Inc(FnStates);
        FMaxSpin := Max(FMaxSpin, S);
      end
    else if Components[i] is TPerturbation then
      Inc(FnPerturbations);
end;

function TManifold.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 4;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

const
  ManifoldSettingNames: array[1..6] of string = (
'Minimum J in calculation - set negative to take from molecule or manifold.',
'Maximum J in calculation - set negative to take from molecule or manifold.',
'True to include population in this state when calculating spectra.',
'Colour for spectra - set to "None" to get colour from elsewhere.',
'Identify state by looking for largest coefficient in eigenvector.',
'Set to assume the energy ordering within an individual state does not change, but the eigenvectors are used to select the state.'
);

function TManifold.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(ManifoldSettingNames))
     and (AIndex <= High(ManifoldSettingNames)) then
    Result := ManifoldSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TManifold.nChildTypes: Integer;
begin
  Result := 2;
end;

{ TMolecule }

constructor TMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := 'Colour=None,Abundance=1,Jmin=-1,Jmax=-1,' + SavedProperties;
  DimensionlessParameters := 'Abundance,' + DimensionlessParameters;
  FLastPartitionFunction := -99;
  inherited Create(AOwner);
end;

function TMolecule.GetManifold(AIndex: Integer): TManifold;
var
  i, n: Integer;
begin
  { TODO: Speed up search? }
  n := 0;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TManifold then begin
      if n = AIndex then begin
        Result := TManifold(Items[i]);
        Exit;
      end;
      Inc(n);
    end;
  raise Exception.CreateFmt('No such Mainfold Index in %s: %d', [Name, AIndex]);
end;

class function TMolecule.OwnerType: TComponentClass;
begin
  Result := TSpecies;
end;

procedure TMolecule.ReadProperties(Node: TDOMNode);
begin
  inherited ReadProperties(Node);
  CheckProperties;
end;

function TMolecule.FixupProperties:Exception;
var
  n, i, j, nLower: Integer;
  Carry, New: TQuantumNumber;
  OddSpinDetermined: Boolean;
begin
  Result := nil;
  FnManifolds := 0;
  FnTransitionMoments := 0;
  FMaxRank := 0;
  { Compute nManifolds and check for duplicate names }
  for i := 0 to ComponentCount-1 do begin
    if Items[i] is TManifold then
      Inc(FnManifolds)
    else if Items[i] is TTransitionMoments then begin
      Inc(FnTransitionMoments);
      FMaxRank := Max(MaxRank, TTransitionMoments(Items[i]).MaxRank);
    end;
    for j := i+1 to ComponentCount-1 do
      if (Items[j] is TManifold) and (Items[i].Name = Items[j].Name) then
        Result := EReportException.Create('Duplicate Manifold Names:'+ Items[i].Name)
  end;
  if not Assigned(Result) then
    Result := inherited FixupProperties;
  if nManifolds = 0 then Exit;
  nLower := 0;
  FLowestOrigin := Manifold[0].LowestStateOrigin; { Not necessarily valid, but avoid awkward numbers }
  FLongestManifoldName := '';
  OddSpinDetermined := False;
  for i := 0 to nManifolds-1 do begin
    if Manifold[i].nStates > 0 then
      if not OddSpinDetermined then begin
        OddSpinDetermined := True;
        FOddSpin := Manifold[i].OddSpin;
      end else if (Result = nil) and (FOddSpin <> Manifold[i].OddSpin) then
        Result := EReportException.Create('Odd and even numbers of electrons in ' + Name);
    if Manifold[i].Initial and ((nLower=0) or (Manifold[i].LowestStateOrigin < FLowestOrigin)) then
      FLowestOrigin := Manifold[i].LowestStateOrigin;
    if Manifold[i].Initial then
      Inc(nLower);
    if Length(Manifold[i].Name) > Length(LongestManifoldName) then
      FLongestManifoldName := Manifold[i].Name;
  end;
  { Make up HyperfineBasisStates }
  { First Determine total number of spin combinations }
  n := 1;
  FMaxNuclearSpin := 0;
  for i := 0 to nNuclei-1 do begin
    n := n * (Nuclei[i].Spin+1);
    Inc(FMaxNuclearSpin,Nuclei[i].Spin);
  end;
  SetLength(HyperfineBasisStates,n,nNuclei);
  { Now make up different combinations }
  for i := 0 to nNuclei-1 do
    HyperfineBasisStates[0,i] := -Nuclei[i].Spin;
  for i := 1 to n-1 do begin
    Carry := 2;
    for j := nNuclei-1 downto 0 do begin
      New := HyperfineBasisStates[i-1,j] + Carry;
      if New > Nuclei[j].Spin then
        New := -Nuclei[j].Spin
      else
        Carry := 0;
      HyperfineBasisStates[i,j] := New
    end;
  end
end;

function TMolecule.CalculatePartitionFunction(Temperature, Tvib: Double; Print: Boolean = False;
                                              ShowMolecule: Boolean = false; ShowHeader: Boolean = false): Double;
var
  Q: Double;
  i, l: Integer;
  J, JUp: TQuantumNumber;
  Sym: Integer;
  Ev: TEigenvalues;
  nInitialManifolds: Integer;
  LastSums: array [0..8] of Double; { Record last 9 partial sums }
  LastIndex: Integer;
  Converged: Boolean;
begin
  Result := 0;
  nInitialManifolds := 0;
  LastIndex := 0;
  Converged := False;
  if Print then begin
    for i := 0 to nManifolds-1 do
      if Manifold[i].Initial then
        Inc(nInitialManifolds);
    if (nInitialManifolds > 0) and ShowHeader then begin
      Write('Partition function sum ');
      if not ShowMolecule then
        Write('for ', Name);
      WriteLn(' at Temperature = ', Temperature:6:3, ' K, Tvib = ', Tvib:6:3, ' K');
      if ShowMolecule then
        Write(Copy('Molecule',1,Length(Mixture.LongestMoleculeName)):Length(Mixture.LongestMoleculeName), ' ');
      if nInitialManifolds > 1 then
        Write('M':Length(Mixture.LongestManifoldName), ' ');
      WriteLn('   J        Contribution      Total');
    end;
  end;
  for i := 0 to nManifolds-1 do
    if Manifold[i].Initial then begin
      J := Manifold[i].FindJmin(OddSpin);
      JUp := Manifold[i].FindJmax(Species.Jmax);
      PushProgress('Partition Function for ' + Manifold[i].Name, JUp { div 2 - don't know, so be pessimistic });
      while not Converged or (J <= JUp) do begin
        Sym := Manifold[i].FirstSym(J);
        Q := 0;
        while Sym >= 0 do begin
          Ev := Manifold[i].Eigenvalues(J, Sym, [meHrot,meWidth]);
          for l := 1 to Ev.Size do
            Q := Q + Ev.StatisticalWeight * (J+1) * Population(Ev, l, Temperature, Tvib);
          Sym := Manifold[i].NextSym(J, Sym);
        end;
        Result := Result + Q;
        if Result <> 0 then begin
          Converged := (LastIndex > High(LastSums))
                       and not ((Result - LastSums[LastIndex mod Length(LastSums)])/ Result > 1e-4);
          LastSums[LastIndex mod Length(LastSums)] := Result;
          Inc(LastIndex);
        end;
        if (Print) then begin
          if ShowMolecule then
            Write(Manifold[i].Molecule.Name:Length(Mixture.LongestMoleculeName), ' ');
          if nInitialManifolds > 1 then
            Write(Manifold[i].Name:Length(Mixture.LongestManifoldName), ' ');
          WriteLn(J:4, Result:20:6, Q:20:6);
        end;
        SetProgress(J div 2);
        Inc(J, 2);
      end;
      PopProgress;
    end;
end;

function TMolecule.PartitionFunction(Temperature, Tvib: Double):Double;
{ As above, but with cached result }
begin
  if (FLastPartitionFunction >= 0) and (Temperature = FLastTemperature) and (Tvib = FLastTvib) then
    Result := FLastPartitionFunction
  else begin
    FLastTemperature := Temperature;
    FLastTvib := Tvib;
    FLastPartitionFunction := CalculatePartitionFunction(Temperature, Tvib);
    Result := FLastPartitionFunction;
  end;
end;


function TMolecule.AddToSpectrum(
  Spectrum: TSpectrum; PrintLevel: Integer;
  const Excitation: TExcitation;
  { Note that all the following parameters have a default, which implies all possible values }
  J1Sel, Sym1Sel, J2Sel, Sym2Sel: Integer;
  Manifold1Sel, Manifold2Sel: TManifold;
  DeltaJ: TQuantumNumber;
  TransitionMomentsSel: TTransitionMoments;
  TransitionMomentSel: TTransitionMoment;
  State1Sel, State2Sel: TState;
  IncludePerturbations: Boolean):Boolean;
var
  i, j, Sym1, Sym2: Integer;
  J1, J1Low, J1Up: TQuantumNumber;
  J2, J2Low, J2Up, J2Upp: TQuantumNumber;
  MaxDj: TQuantumNumber;
  Manifold1, Manifold2: TManifold;
  Connection: TTransitionMoments;
  Ev1, Ev2: TEigenvalues;
  nInitial, nConnections: Integer;
  JAdjustSym: Boolean;
begin
  if Abundance = 0 then Exit;
  nInitial := 0;
  nConnections := 0;
  JAdjustSym := GetMixture.JAdjustSym;
  for i := 0 to nManifolds-1 do
    if (Manifold[i] = Manifold1Sel)
       or (not Assigned(Manifold1Sel) and Manifold[i].Initial ) then begin
      { For each manifold marked as initial }
      Manifold1 := Manifold[i];
      Inc(nInitial);
      for j := 0 to nManifolds-1 do
        if (Manifold[j] = Manifold2Sel)
           or (not Assigned(Manifold2Sel) and ((j >= i) or not Manifold[j].Initial)) then begin
          { For every other manifold, except those we have already started from }
          Manifold2 := Manifold[j];
          Connection := TransitionMoments[i, j];
          if Assigned(Connection)
            and ((Connection = TransitionMomentsSel) or not Assigned(TransitionMomentsSel)) then begin
            Inc(nConnections);
            if Assigned(TransitionMomentSel) then
              MaxDj := TransitionMomentSel.Rank
            else
              MaxDj := Connection.MaxRank;
            if J1Sel >= 0 then begin
              J1Low := J1Sel;
              J1Up := J1Sel;
              if J2Sel >= 0 then begin
                J2Low := J2Sel;
                J2Up := J2Sel;
              end else if Abs(DeltaJ) < 100 then begin
                J2Low := J1Low + DeltaJ;
                J2Up := J2Low;
                if J2Low < 0 then Exit;
              end else begin
                J2Low := SmallestProjection(J1Sel - MaxDj);
                J2Up := J1Sel + MaxDj;
              end;
            end else if J2Sel >= 0 then begin
              J2Low := J2Sel;
              J2Up := J2Sel;
              if Abs(DeltaJ) < 100 then begin
                J1Low := J2Low - DeltaJ;
                J1Up := J1Low;
                if J1Low < 0 then Exit;
              end else begin
                J1Low := SmallestProjection(J2Sel - MaxDj);
                J1Up := J2Sel + MaxDj;
              end;
            end else begin
              J1Low := Manifold1.FindJmin(OddSpin);
              J1Up := Manifold1.FindJmax(Manifold1.Molecule.Species.Jmax);
              J2Low := Manifold1.FindJmin(J1Low - MaxDj);
              J2Up := Manifold1.FindJmax(J1Up + MaxDj);
            end;
            J1 := J1Low;
            PushProgress(Manifold2.Name + ' - ' + Manifold1.Name, J1Up div 2);
            while J1 <= J1Up do begin
              if Sym1Sel >= 0 then begin
                Sym1 := Sym1Sel;
                if JAdjustSym then
                  Sym1 := JAdjustedSym(J1, Sym1);
              end else
                Sym1 := Manifold1.FirstSym(J1);
              while Sym1 >= 0 do begin
                Ev1 := Manifold1.Eigenvalues(J1, Sym1, [meHrot, meWidth], PrintLevel, State1Sel, IncludePerturbations);
                if Ev1.Size > 0 then begin
                  if Abs(DeltaJ) < 100 then begin
                    J2 := J1 + DeltaJ;
                    J2Upp := J2;
                  end else begin
                    J2 := J1-MaxDj;
                    J2Upp := J2Up;
                  end;
                  if J2 < J2Low then J2 := J2Low;
                  while (J2 <= J2Upp) and (J2 <= J1+MaxDj) do begin
                    if Sym2Sel >= 0 then begin
                      Sym2 := Sym2Sel;
                      if JAdjustSym then
                        Sym2 := JAdjustedSym(J2, Sym2);
                    end else
                      Sym2 := Manifold2.FirstSym(J2);
                    while sym2 >= 0 do begin
                      if Connection.Allowed(J1, Sym1, J2, Sym2) then begin
                        Ev1 := Manifold1.Eigenvalues(J1, Sym1, [meHrot, meWidth], PrintLevel, State1Sel, IncludePerturbations); { Just in case it was flushed from the cache }
                        Ev2 := Manifold2.Eigenvalues(J2, Sym2, [meHrot, meWidth], PrintLevel, State2Sel, IncludePerturbations);
                        if Ev2.Size > 0 then
                          Connection.AddToSpectrum(Spectrum, Ev1, Ev2, TransitionMomentSel, PrintLevel, Excitation);
                      end;
                      if Sym2Sel >= 0 then
                        Break
                      else
                        Sym2 := Manifold2.NextSym(J2, Sym2);
                    end;
                    Inc(J2, 2);
                  end;
                end;
                if Sym1Sel >= 0 then
                  Break
                else
                  Sym1 := Manifold1.NextSym(J1, Sym1);
              end;
              SetProgress(J1 div 2);
              Inc(J1, 2);
            end;
            PopProgress;
          end;
        end;
    end;
  if (nManifolds > 0) and (nInitial = 0) then
      raise EReportException.Create('No Initial States set for '+ Name);
  Result := nConnections > 0;
end;

function TMolecule.ManifoldWithName(AName: string): TManifold;
var
  i: Integer;
begin
  Result := nil;
  { NB: can't use FindComponent as Name property has been overridden }
  for i := 0 to ComponentCount-1 do
    if (Items[i] is TManifold) and (TManifold(Items[i]).Name = AName) then begin
      Result := TManifold(Items[i]);
      Exit;
    end;
end;

function TMolecule.ManifoldStartingWithName(AName: string): TManifold;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to ComponentCount-1 do
    if (Items[i] is TManifold) and (Pos(AName,TManifold(Items[i]).Name) = 1) then begin
      Result := TManifold(Items[i]);
      Exit;
    end;
end;

function TMolecule.FindTransitionMoments(ManifoldA, ManifoldB: TManifold): TTransitionMoments;
var
  i: Integer;
  tms: TTransitionMoments;
begin
  Result := nil;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TTransitionMoments then begin
      tms := TTransitionMoments(Items[i]);
      if ((tms.Bra = ManifoldA) and (tms.Ket = ManifoldB))
         or ((tms.Bra = ManifoldB) and (tms.Ket = ManifoldA)) then begin
          if Assigned(Result) then
            raise EReportException.Create('Duplicate transition moments <' + ManifoldA.Name + '|' + Result.PairName + '|' + ManifoldB.Name + '> in '+Name);
          Result := TTransitionMoments(Items[i]);
      end;
    end;
end;

function TMolecule.GetTransitionMomentsFunction(AIndex, BIndex: Integer): TTransitionMoments;
begin
  { TODO: Speed up search? }
  Result := FindTransitionMoments(Manifold[AIndex], Manifold[BIndex]);
end;

function TMolecule.Species: TSpecies;
begin
  Result := Owner as TSpecies;
end;

function TMolecule.GetMixture: TMixture;
begin
  Result := Owner.Owner as TMixture;
end;

function TMolecule.GetTransitionMoments(AIndex: Integer): TTransitionMoments;
var
  i: Integer;
begin
  Result := nil;
  if nTransitionMoments < 1 then Exit;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TTransitionMoments then begin
      Dec(AIndex);
      if AIndex < 0 then begin
        Result := TTransitionMoments(Items[i]);
        Exit;
      end;
    end;
end;

procedure TMolecule.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache;
end;

function TMolecule.SymName(Sym: Integer): string;
begin
  Result := IntToStr(Sym);
end;

function TMolecule.JAdjustedSymName(J: TQuantumNumber; Sym: Integer): string;
begin
  Result := SymName(Sym); { Default to normal name }
end;

function TMolecule.JDeAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
var
  i: Integer;
begin
  { Reverse effects of JAdjustedSym }
  { This can be overridden at the individual molecule level to make it more efficient }
  if JAdjustedSym(J, Sym) = Sym then
    Result := Sym { No adjustment - easy }
  else begin
    for i := 0 to LastSym do
      if (i <> Sym) and (JAdjustedSym(J, i) = Sym) then begin
        Result := i;
        Exit;
      end;
    raise Exception.CreateFmt('JDeAdjustedSym fails for %s: %d %d', [Name, J, Sym]);
  end;
end;

procedure TMolecule.SetnNuclei(const Index, Value: Integer);
var
  i, j: Integer;
begin
  inherited SetSetting(Index, Value);
  SetLength(Nuclei,Value);
  { Make sure each state has the right number of nuclei }
  for i := 0 to ComponentCount-1 do
    if Items[i] is TManifold then
      with TManifold(Items[i]) do
        for j := 0 to ComponentCount-1 do
          if Items[j] is TState then
            TState(Items[j]).AdjustNuclei;
end;

function TMolecule.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 4;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Number of nuclei to calculate hyperfine structure for.'
  else if AIndex = 2 then
    Result := 'Minimum J to use in calculation - set to -1 (default) to use value from species.'
  else if AIndex = 3 then
    Result := 'Maximum J to use in calculation - set to -1 (default) to use value from species.'
  else if AIndex = 4 then
    Result := ColourHelpString
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TMolecule.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Abundance of this isotope; default is 1'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TMolecule.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  case AIndex of
    0: Result := ManifoldType;
    1: Result := TTransitionMoments
  else
    Result := nil;
  end;
end;

function TMolecule.nChildTypes: Integer;
begin
  Result := 2;
end;

procedure TMolecule.ClearCache;
begin
  inherited;
  FLastPartitionFunction := -1; { TODO: May be able to shift partition fuction calculation to TManifold(s) }
end;

{ TSpecies }

function TSpecies.GetIsotopomer(AIndex: Integer): TMolecule;
begin
  Result := TMolecule(Items[AIndex]);
end;

constructor TSpecies.Create(AOwner: TComponent);
begin
  SavedProperties := 'Colour=None,Concentration=1,Jmin=-1,Jmax=50,' + SavedProperties;
  DimensionlessParameters := 'Concentration,' + DimensionlessParameters;
  inherited Create(AOwner);
end;

class function TSpecies.OwnerType: TComponentClass;
begin
  Result := TMixture;
end;

function TSpecies.AddToSpectrum(Spectrum: TSpectrum):Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Count-1 do
    { Note that we want to call AddToSpectrum for every species }
    Result := Isotopomer[i].AddToSpectrum(Spectrum, (Owner as TMixture).PrintLevel, RandomExcitation)
              or Result ;
end;

function TSpecies.FindMolecule(AName: string): TMolecule;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Count-1 do
    if CompareText(Isotopomer[i].Name, AName) = 0 then
      if Assigned(Result) then
        raise EReportException.CreateFmt('Duplicate Isotopomer name in %s: %s', [AName, Result.Name])
      else
        Result := Isotopomer[i];
end;

function TSpecies.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 3;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TSpecies.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Minimum J to use in calculation - set to -1 to use lowest J.'
  else if AIndex = 2 then
    Result := 'Maximum J to use in calculation.'
  else if AIndex = 3 then
    Result := ColourHelpString
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSpecies.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative concentration of species'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

var
  MoleculeTypes: TList = nil; { array of TMNodeClass would be nicer, but can't guarantee initialisation order }

function TSpecies.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  Result := TMNodeClass(MoleculeTypes[AIndex]);
end;

function TSpecies.nChildTypes: Integer;
begin
  Result := MoleculeTypes.Count;
end;

class procedure TSpecies.AddMoleculeType(MoleculeClass: TMNodeClass);
begin
  if not Assigned(MoleculeTypes) then
    MoleculeTypes := TList.Create;
  MoleculeTypes.Add(MoleculeClass);
end;

{ TMixture }

function TMixture.GetSpecies(AIndex: Integer): TSpecies;
begin
  Result := TSpecies(Items[AIndex]);
end;

constructor TMixture.Create(AOwner: TComponent);
begin
  Name := 'Mixture';
  SavedProperties := 'OThreshold=1e-6,Precision=4,WidthMult=6,SMargin=-1,Temperature=300,FMin=0,FMax=50000,nDF=2000,ShowSum=True,JAdjustSym=True,Tvib=-1,PrintLevel=-3,' + SavedProperties;
  SavedProperties := 'AutoReplot=True,ScaleMarkSize=True,ShowSymmetry=True,DoubleQnos=True,FitCycles=1,' + SavedProperties;
  DimensionlessParameters := 'OThreshold,SMargin,Temperature,Imin,Tvib,' + DimensionlessParameters;
  { The next five are not dimensionless, but are only affected by PlotUnits }
  DimensionlessParameters := 'Foffset,Fmin,Fmax,Gaussian,Lorentzian,' + DimensionlessParameters;
  inherited;
  ClearUndo;
  HasPlotted := False;
end;

destructor TMixture.Destroy;
begin
  FreeAndNil(FLineList);
  inherited Destroy;
end;

function TMixture.MakeLineList: TLineList;
var
  i: Integer;
  TransitionMomentsSet: Boolean;
begin
  TStart(MakeLineListTime);
  CheckProperties;
  Result := TLineList.Create(Units);
  Result.Temperature := Temperature;
  Result.Tvib := Tvib;
  Result.HeadingPrinted := False;
  TransitionMomentsSet := False;
  try
    for i := 0 to Count-1 do
      TransitionMomentsSet := Species[i].AddToSpectrum(Result)
                              or TransitionMomentsSet;
  except
    Result.Free;
    raise;
  end;
  if PrintLevel >= 0 then
    Result.ShowFooter(Output,' ');
  LineList := Result;
  TEnd(MakeLineListTime);
end;

procedure TMixture.PreCalculatePartitionFunctions;
var
  i, j: Integer;
begin
  for i := 0 to Count-1 do
    for j := 0 to Species[i].Count-1 do
      Species[i].Isotopomer[j].PartitionFunction(Temperature, Tvib);
end;

procedure TMixture.ConvertToPerturbations;
var
  i, j, k: Integer;
begin
  for i := 0 to Count-1 do
    with Species[i] do
      for j := 0 to Count-1 do
        with Isotopomer[j] do
          for k := 0 to nManifolds-1 do
            Manifold[i].ConvertToPerturbations;
  FixupProperties;
end;

function TMixture.FixupProperties:Exception;
var
  i, j: Integer;
  LManName: string;
begin
  Result := inherited FixupProperties;
  FnMolecules := 0;
  FLongestMoleculeName := '';
  LManName := '';
  for i := 0 to Count - 1 do
    with Species[i] do begin
      Inc(FnMolecules, Count);
      for j := 0 to Count-1 do
        with Isotopomer[j] do begin
          if Length(LongestMoleculeName) < Length(Name) then
            FLongestMoleculeName := Name;
          if Length(LongestManifoldName) > Length(LManName) then
            LManName := LongestManifoldName;
        end;
    end;
  FLongestManifoldName := LManName;
end;

procedure TMixture.SetModified(const Value: Boolean);
begin
  inherited;
  LineList := nil;
end;

procedure TMixture.SetParameter(AIndex: Integer; const Value: Double);
begin
  inherited ModifyParameter(AIndex, Value, False);
end;

procedure TMixture.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  inherited ModifySetting(AIndex, AValue, False);
end;

function TMixture.FindMolecule(AName: string): TMolecule;
var
  Molecule: TMolecule;
  i: Integer;
begin
  Result := nil;
  for i := 0 to Count-1 do begin
    Molecule := Species[i].FindMolecule(AName);
    if Assigned(Result) and Assigned(Molecule) then
      raise EReportException.Create('Duplicate molecule name: ' + Result.Name);
    Result := Molecule;
  end;
end;

const
  MixtureSettingNames: array[1..20] of string = (
'Units for calculation - valid values are cm1, MHz, Kelvin and eV. The key change this makes is in calculating Boltzmann factors from temperature.',
'Units for intensity. If not Arbitrary or Normalized, transition moments should be in Debye.',
'Controls the amount of printout in the Log window. Negative for no output, 0 for a line list, larger numbers for printouts of Hamiltonian and transition matrices.',
'Number of points (between Fmin and Fmax) to calculate the spectrum at. Note that if a peak width is less than 3*(Fmax-Fmin)./nDF, i.e. only a few points wide, then it is shown as a stick, rather than a peak.',
'Multiple of linewidth to extend convolution over.',
'Precision of printout (decimal places after the point).',
'Model to use for state dependent lifetimes and linewidths. Valid values are lmNone, lmWidth, lmProduct, lmProductWidth, lmParent, lmParentWidth, lmGate, lmGateWidth.',
'How to order the basis states making up a basis. Valid values are: boDefault, boDiagonal and boStateDiagonal.',
'Plot overall sum of individual spectra.',
'Plot individual spectra making up overall spectrum. The individual spectra are grouped by colour, so you will need to set colours to see something different to the sum. Colours can be set at the transition moment, state, molecule or species level.',
'Plot a Fortrat diagram, i.e. J against frequency, in the main window.',
'Set to use upper rather than lower state J and symmetry in the Fortrat diagram.', 
'If set, energy level plots and Fortrat diagrams will take account of the alternation in symmetry with J that is often found. For example in linear molecules an energy level will typically alternate between + and - parity with J (hence the e/f notation).',
'Show symmetry in fortrat plots.',
'Show change in J in fortrat plots.',
'Scale Mark Size with intensity in fortrat plots.',
'If set all quantum numbers are displayed as twice their actual values.',
'Number of least squares fit cycles to do on each press of a "fit" button.',
'If false, automatic recalculation of the simulation is disabled in some circumstances. Use if the program is running slowly.',
'Units for the horizontal scale for spectrum plots'
);

function TMixture.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(MixtureSettingNames))
     and (AIndex <= High(MixtureSettingNames)) then
    Result := MixtureSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 MixtureParameterNames: array [1..11] of string = (
'Left edge of plot range in main window.',
'Right edge of plot range in main window.',
'Rotational temeperature (Kelvin).',
'Gaussian contribution to linewidth (full width half maximum). If both this and a Lorentzian width (below) is set the result is a convolution of the two, a Voigt profile.',
'Lorentzian contribution to linewidth (full width half maximum). This and the Gaussian width can also be set from the main window.',
'Frequency offset to simulation.',
'Relative Size of peak to ignore when convoluting for contour fitting.',
'Ignore peaks smaller than this fraction of the maximum peak intensity.',
'Reference width in linewidth (predissociation) calculations.',
'Vibrational Temperature (Kelvin); set to -1 (default) to use rotational temperature for all Boltzmann factors.',
'Set the scaling of mark sizes in Fortrat plots.'
);

function TMixture.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(MixtureParameterNames))
     and (AIndex <= High(MixtureParameterNames)) then
    Result := MixtureParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TMixture.AddNodeParameters(MNode: TMNode);
var
  i, n: Integer;
  s: string;
begin
  for i := 1 to MNode.nParameters do
    if MNode.Floating[i] then begin
      n := Length(FittingParameters);
      SetLength(FittingParameters, n+1);
      FittingParameters[n].MNode := MNode;
      FittingParameters[n].Index := i;
      FittingParameters[n].Value := MNode.GetParameter(i);
      FittingParameters[n].Increment := MNode.Increment[i];
      s := MNode.UniqueName;
      if s <> '' then
        s := s + ' ';
      FittingParameters[n].Name := s + MNode.ParameterNames[i];
    end;
  for i := 0 to MNode.Count-1 do
    AddNodeParameters(MNode.Items[i]);
end;

procedure TMixture.FillParameters;
begin
  SetLength(FittingParameters, 0);
  AddNodeParameters(Self);
end;

procedure TMixture.SetLifeModel(const Index: Integer; const AValue: TLifeModel);
begin
  SetSetting(Index, Ord(AValue));
end;

function TMixture.GetLifeModel(const Index: Integer): TLifeModel;
begin
  Result := TLifeModel(GetSetting(Index));
end;

procedure TMixture.SetBasisOrder(const Index: Integer; const AValue: TBasisOrder);
begin
  SetSetting(Index, Ord(AValue));
end;

function TMixture.GetBasisOrder(const Index: Integer): TBasisOrder;
begin
  Result := TBasisOrder(GetSetting(Index));
end;

function TMixture.CanRedoFit: Boolean;
begin
  Result := FUndoPosition < Length(FSavedParameters);
end;

function TMixture.CanUndoFit: Boolean;
begin
  Result := FUndoPosition > 0;
end;

procedure TMixture.PushParameters(FitType: TFitType);
begin
  if (Length(FSavedParameters) > 0) and (FSavedParameters[0].FitType <> FitType) then
    ClearUndo;
  SetLength(FSavedParameters, FUndoPosition+1);
  FSavedParameters[FUndoPosition].FitType := FitType;
  FSavedParameters[FUndoPosition].Parameters := FittingParameters; { TODO: do we need to force copy? }
  Inc(FUndoPosition);
end;

procedure TMixture.RedoFit;
begin
  if CanRedoFit then begin
    LoadCurrentSavedParameters;
    Inc(FUndoPosition);
  end;
end;

procedure TMixture.UndoFit;
begin
  if CanUndoFit then begin
    Dec(FUndoPosition);
    LoadCurrentSavedParameters;
  end;
end;

procedure TMixture.ClearUndo;
begin
  FUndoPosition := 0;
  SetLength(FSavedParameters, 0);
end;

procedure TMixture.DiscardFit;
begin
  Dec(FUndoPosition);
  if FUndoPosition < 0 then
    FUndoPosition := 0;
  SetLength(FSavedParameters, FUndoPosition);
end;

procedure TMixture.LoadCurrentSavedParameters;
var
  i: Integer;
begin
  with FSavedParameters[FUndoPosition] do
    for i := 0 to High(Parameters) do
      with Parameters[i] do
        MNode.SetParameter(Index, Value);
  {$IFNDEF CONSOLE}
  ResetProgress;
  with MainForm do
    if FSavedParameters[FUndoPosition].FitType = ftContour then
      FitContour(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowContour, nil)
    else if not Split.Visible and Mixture.HasPlotted then
      DoPlot;
  {$ENDIF}
end;

procedure TMixture.SetLineList(const Value: TLineList);
begin
  FLineList.Free;
  FLineList := Value;
end;

function TMixture.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  Result := TSpecies;
end;

function TState.AddBasis(Eigenvalues: TEigenvalues): Integer;
var
  F, J: TQuantumNumber;
  StatWt, i, nuc: Integer;
begin
  Result := -1;
  for i := 0 to High(Molecule.HyperfineBasisStates) do begin
    F := Eigenvalues.J;
    for nuc := Molecule.nNuclei-1 downto 0 do begin
      J := F - Molecule.HyperfineBasisStates[i,nuc];
      if (J < 0) or not IsTriangle(J,Molecule.Nuclei[nuc].Spin,F) then begin
        F := -1;
        Break;
      end;
      F := J;
    end;
    if F >= 0 then begin
      StatWt := AddOneJ(Eigenvalues,F,Molecule.HyperfineBasisStates[i]);
      if Result < 0 then
        Result := StatWt
      else if Result <> StatWt then
        raise Exception.Create('AddBasis: Inconsistent Statistical Weights in '+Name);
    end;
  end;
end;

function TState.AddNucleus: TNucleus;
begin
  Result := NucleusType.Create(Self);
end;

procedure TState.AdjustNuclei; { Add or remove Nuclei as required }
begin
  while ComponentCount < Molecule.nNuclei do
    AddNucleus;
  while ComponentCount > Molecule.nNuclei do
    Components[ComponentCount-1].Free;
end;

constructor TState.Create(AOwner: TComponent);
begin
  if Assigned(AOwner) and Assigned(AOwner.Owner) then
    FMolecule := AOwner.Owner as TMolecule;
  SavedProperties := 'Colour=None,' + SavedProperties;
  inherited;
  FFullRotationalSymmetry := True;
end;

function TState.FindColour: TColor;
begin
  Result := Colour;
  if Result = clNone then
    Result := Manifold.Colour;
  if Result = clNone then
    Result := Molecule.Colour;
  if Result = clNone then
    Result := Molecule.Species.Colour;
end;

function TState.FixupProperties: Exception;
begin
  AdjustNuclei;
  Result := inherited FixupProperties;
end;

function TState.nChildTypes: Integer;
begin
  Result := 0;
end;

function TState.HasWidth: Boolean;
begin
  Result := Width <> 0;
end;

function TState.SymmetrizedRotationalMatrixElement(A, B: TBasisState; MNode: TMnode; Flag: Integer;
  MatrixElementFunction: TMatrixElementFunction; SymCheck: TSymCheck): Double;
var
  r: Double;
  procedure FullCheck(Id: Integer);
  begin
    { Slower test here, but allow for slight rounding errors }
    if not SameValue(r, Result)
       and not ((SymCheck = MayCancel) and SameValue(r, -Result)) then begin
        Writeln('Symmetry mismatch ', Id, ':', r, ' ', Result, ' ', A.BasisLabel, ' ', B.BasisLabel);
        raise Exception.Create('Symmetry mismatch ' + IntToStr(Id));
    end;
  end;
begin
  Result := MatrixElementFunction(A, B, 1, 1, MNode, Flag);
  if A.Phase = 0 then begin
    if B.Phase <> 0 then
      Result := (Result + B.Phase*MatrixElementFunction(A, B, 1, -1, MNode, Flag))/Sqrt(2.0);
  end else if B.Phase = 0 then
    Result := (Result + A.Phase*MatrixElementFunction(A, B, -1, 1, MNode, Flag))/Sqrt(2.0)
  else begin
    if not FullRotationalSymmetry and (SymCheck = MustEqual) then
      SymCheck := MayCancel;
    {$IFOPT R-}
    if SymCheck = MustEqual then
      SymCheck := NoCheck;
    {$ENDIF}
    // SymCheck := FullSum;
    if SymCheck <> NoCheck then begin
      r := A.Phase*B.Phase*MatrixElementFunction(A, B, -1, -1, MNode, Flag);
      if (SymCheck <> FullSum) and (r <> Result) and not ((SymCheck = MayCancel) and (r =-Result)) then begin
        // MatrixElementFunction(A, B, 1, 1, MNode, Flag);
        // MatrixElementFunction(A, B, -1, -1, MNode, Flag);
        FullCheck(1);
      end;
    end;
    Result := Result + B.Phase*MatrixElementFunction(A, B, 1, -1, MNode, Flag);
    if SymCheck <> NoCheck then begin
      r := r + A.Phase*MatrixElementFunction(A, B, -1, 1, MNode, Flag);
      if (SymCheck <> FullSum) and (r <> Result) and not ((SymCheck = MayCancel) and (r =-Result)) then
        FullCheck(2);
    end;
    if SymCheck = MayCancel then
      Result := (Result + r)/2;
  end;
end;

function TState.SymmetrizedHyperfineMatrixElement(A, B: TBasisState;
  F, Rank: TQuantumNumber; Nucleus: TNucleus;
  MatrixElementFunction: TMatrixElementFunction;
  MNode: TMNode): Double;
begin
  Result := 0;
  if (Nucleus.Spin < Rank div 2) or (Abs(A.J-B.J) > Rank) then
    Exit;
  Result := SymmetrizedRotationalMatrixElement(A, B, MNode, F, MatrixElementFunction);
end;

procedure TState.ConvertToPerturbations;
begin
  { By default, no parameters converted }
end;

function TState.Manifold: TManifold;
begin
  Result := Owner as TManifold;
end;

function TState.MatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Double;
{ Matrix element of normal Hamiltonian }
var
  FFp, FF, FF1p, FF1: TQuantumNumber; { Partial F quantum numbers }
  nuc: Integer;
  Rank1Sum, Rank2Sum: Double;
begin
  Result := 0;
  if A.State <> B.State then
    Exit;
  { Handle multiple nuclear spin case here, by decomposing into simple calls
    for a single nucleus }
  if (MatElemType = meHrot) { TODO: Other matrix element types }
     and (Molecule.nNuclei > 0) and (Abs(A.J-B.J) <= 4) then begin
    { Work from last nucleus }
    FFp := A.F;
    FF := B.F;
    Rank1Sum := 0;
    Rank2Sum := 0;
    for nuc := Molecule.nNuclei-1 downto 0 do begin
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      if Rank1Sum <> 0 then
        Rank1Sum := Rank1Sum * Minus(FF1p+Molecule.Nuclei[nuc].spin+FF+2)*Sqrt((FFp+1)*(FF+1))
                             * sixj(FF1p, FFp, Molecule.Nuclei[nuc].spin, FF, FF1, 2);
      if Rank2Sum <> 0 then
        Rank2Sum := Rank2Sum * Minus(FF1p+Molecule.Nuclei[nuc].spin+FF+4)*Sqrt((FFp+1)*(FF+1))
                             * sixj(FF1p, FFp, Molecule.Nuclei[nuc].spin, FF, FF1, 4);
      if FF = FFp then begin
        { This nucleus contributes }
        Rank1Sum := Rank1Sum + Minus(FF1+Molecule.Nuclei[nuc].spin+FF)*sixj(FF,Molecule.Nuclei[nuc].spin,FF1p,2,FF1,Molecule.Nuclei[nuc].spin)
                               *SymmetrizedHyperfineMatrixElement(A, B, FF*10+2, 2, A.State.Nucleus(nuc), {$IFDEF FPC}@{$ENDIF}HyperfineMatrixElement, A.State.Nucleus(nuc));
        if Molecule.Nuclei[nuc].Spin > 1 then
          Rank2Sum := Rank2Sum + Minus(FF1+Molecule.Nuclei[nuc].spin+FF)*sixj(FF,Molecule.Nuclei[nuc].spin,FF1p,4,FF1,Molecule.Nuclei[nuc].spin)
                                 *SymmetrizedHyperfineMatrixElement(A, B, FF*10+4, 4, A.State.Nucleus(nuc), {$IFDEF FPC}@{$ENDIF}HyperfineMatrixElement, A.State.Nucleus(nuc));
      end;
      FFp := FF1p;
      FF := FF1;
    end;
    Result := Rank1Sum + Rank2Sum;
  end;
  if A.HyperfineState = B.HyperfineState then
    Result := Result + SymmetrizedRotationalMatrixElement(A, B, nil, Ord(MatElemType), {$IFDEF FPC}@{$ENDIF}RotationalMatrixElement);
    // Result := Result + RotSpinMatrixElement(A, B, MatElemType);
end;

function TState.PerturbationMatrixElement(Perturbation: TPerturbation; A, B: TBasisState; MatElemType: TMatElemType): Double;
var
  FFp, FF, FF1p, FF1: TQuantumNumber; { Partial F quantum numbers }
  nuc: Integer;
begin
  Result := 0;
  if MatElemType <> meHrot then Exit; { TODO: Perturbation meWidth }
  { Handle multiple nuclear spin case here, by decomposing into simple calls
    for a single nucleus }
  if (Molecule.nNuclei > 0) and (Perturbation.Nucleus > 0) and (Abs(A.J-B.J) <= Perturbation.NuclearRank) then begin
    { Work from last nucleus }
    FFp := A.F;
    FF := B.F;
    for nuc := Molecule.nNuclei-1 downto 0 do begin
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      if (nuc = Perturbation.Nucleus-1) then
        if FF <> FFp then
          Exit
        else
          Result := Minus(FF1+Molecule.Nuclei[nuc].spin+FF)*sixj(FF,Molecule.Nuclei[nuc].spin,FF1p,Perturbation.NuclearRank,FF1,Molecule.Nuclei[nuc].spin)
                    * SymmetrizedHyperfineMatrixElement(A, B, FF, Perturbation.NuclearRank, A.State.Nucleus(nuc), {$IFDEF FPC}@{$ENDIF}RotationalPerturbationMatrixElement, Perturbation)
      else if nuc < Perturbation.Nucleus-1 then
        Result := Result * Minus(FF1p+Molecule.Nuclei[nuc].spin+FF+2)*Sqrt((FFp+1)*(FF+1))
                           * sixj(FF1p, FFp, Molecule.Nuclei[nuc].spin, FF, FF1, Perturbation.NuclearRank);
      FFp := FF1p;
      FF := FF1;
    end;
  end else if (A.HyperfineState = B.HyperfineState) and (A.J = B.J) then
    Result := SymmetrizedRotationalMatrixElement(A, B, Perturbation, Ord(MatElemType), {$IFDEF FPC}@{$ENDIF}RotationalPerturbationMatrixElement);
end;

function TState.TransitionMatrixElement(TransitionMoment: TTransitionMoment; A, B: TBasisState): Double;
{ Similar logic to TState.MatrixElement, but assuming transition moment does not
  act on any nuclear spins }
var
  FFp, FF, FF1p, FF1: TQuantumNumber; { Partial F quantum numbers }
  nuc: Integer;
begin
  Result := 0;
  if Abs(A.J-B.J) > TransitionMoment.Rank then
    Exit;
  Result := 1;
  if Molecule.nNuclei > 0 then begin
    { Work from last nucleus }
    FFp := A.F;
    FF := B.F;
    for nuc := Molecule.nNuclei-1 downto 0 do begin
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      Result := Result * Minus(FF1p+Molecule.Nuclei[nuc].spin+FF+TransitionMoment.Rank)*Sqrt((FFp+1)*(FF+1))
                       * sixj(FF1p, FFp, Molecule.Nuclei[nuc].spin, FF, FF1, TransitionMoment.Rank);
      if Result = 0 then
        Exit;
      FFp := FF1p;
      FF := FF1;
    end;
  end;
  { Phase factor gives consistent sign to matrix element - computes phase factor for smallest M }
  { TODO: Check use of Max - should not be necessary }
  Result := Result * Minus(Max(A.F,B.F)-SmallestProjection(A.F))
            * SymmetrizedRotationalMatrixElement(A, B, TransitionMoment, -1, {$IFDEF FPC}@{$ENDIF}RotationalTransitionMatrixElement, MayCancel);
end;

function TState.Nucleus(Index: Integer): TNucleus;
begin
  Result := Components[Index] as TNucleus;
end;

procedure TState.ReadProperties(Node: TDOMNode);
var
  Child: TDOMNode;
  i, Index: Integer;
  NucName: string;
begin
  if Molecule.nNuclei > 1 then begin
    { Ensure Nuclei read in correct order }
    Index := 0;
    Child := Node.FirstChild;
    while Assigned(Child) do begin
      if Pos('Nucleus', Child.NodeName) = Length(Child.NodeName)-6 then begin
        NucName := ReadValue(Child, 'Name', '');
        if Index < Molecule.nNuclei then
          if (Molecule.Nuclei[Index].Name <> '') and (NucName <> Molecule.Nuclei[Index].Name) then
            for i := 0 to Molecule.nNuclei-1 do
              if Molecule.Nuclei[i].Name = NucName then
                raise Exception.Create('Nucleus ' + NucName + ' out of order for ' + Owner.Name + ' - ' + Node.NodeName);
        Inc(Index);
      end;
      Child := Child.NextSibling;
    end;
  end;
  inherited;
end;

function TState.StateOrigin: Double;
begin
  Result := Origin;
end;

function TState.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TState.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    1: Result := ColourHelpString;
    2: Result := 'Electron Spin';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TState.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'State Origin.'
  else if AIndex = 2 then
    Result := 'Linewidth (rotation independent) for state.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TMixture.SetEUnits(const Index: Integer; const AValue: TEnergyUnits);
begin
  SetUnits(Index, AValue);
end;

function TMixture.GetEUnits(const Index: Integer): TEnergyUnits;
begin
  Result := GetUnits(Index);
end;

procedure TMixture.ChangePlotUnits(NewPlotUnits: TResonanceUnits);
var
  MidPoint: Double; { Centre plot position - needed for scaling widths }
begin
  { Change plot units, scaling affected values also }
  if NewPlotUnits = PlotUnits then
    Exit;
  MidPoint := (FMin + FMax) / 2;
  FMin := Convert(FMin,PlotUnits,NewPlotUnits);
  FMax := Convert(FMax,PlotUnits,NewPlotUnits);
  Gaussian := ConvertWidth(Gaussian, MidPoint, PlotUnits, NewPlotUnits);
  Lorentzian := ConvertWidth(Lorentzian, MidPoint, PlotUnits, NewPlotUnits);
  Foffset := ConvertWidth(Foffset, MidPoint, PlotUnits, NewPlotUnits);
  PlotUnits := NewPlotUnits;
end;

procedure TMixture.CycleUnits;
var
  NewUnits: TEnergyUnits;
begin
  if Units = High(TEnergyUnits) then
    NewUnits := Low(TEnergyUnits)
  else
    NewUnits := Succ(Units);
  ScaleParameters(Units,NewUnits);
  Units := NewUnits;
end;

{ TTransitionMoments }

function TTransitionMoments.GetTransitionMoment(AIndex: Integer): TTransitionMoment;
begin
  Result := Items[AIndex] as TTransitionMoment;
end;

class function TTransitionMoments.OwnerType: TComponentClass;
begin
  Result := TMolecule;
end;

constructor TTransitionMoments.Create(AOwner: TComponent);
begin
  SavedProperties := 'Colour=None,' + SavedProperties;
  inherited Create(AOwner);
end;

constructor TTransitionMoments.Create(AOwner: TComponent; Manifold1Name, Manifold2Name: string);
begin
  Create(AOwner);
  FBraName := Manifold1Name;
  FKetName := Manifold2Name;
end;

function TTransitionMoments.FixupProperties:Exception;
var
  i: Integer;
  Swapped: Boolean;
begin
  Result := inherited FixupProperties;
  if Assigned(Result) then Exit;
  { Sort by Rank }
  repeat
    Swapped := False;
    for i := 0 to ComponentCount-2 do
      if TransitionMoment[i].Rank > TransitionMoment[i+1].Rank then begin
        TransitionMoment[i].ComponentIndex := i+1;
        Swapped := True;
      end;
  until not Swapped;
  if ComponentCount > 0 then
    FnRanks := 1
  else
    FnRanks := 0;
  for i := 0 to ComponentCount-2 do
    if TransitionMoment[i].Rank <> TransitionMoment[i+1].Rank then
      Inc(FnRanks);
end;

function TTransitionMoments.MaxRank: Integer;
begin
  if ComponentCount > 0 then
    Result := TransitionMoment[ComponentCount-1].Rank
  else
    Result := -2;
end;

function TTransitionMoments.Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean;
{ Do any of the transition moments give rise to an allowed transition? }
var
  i: Integer;
begin
  Result := True;
  if IsTriangle(J1, J2, MaxRank) then
    for i := 0 to ComponentCount-1 do begin
      if (Items[i] as TTransitionMoment).Allowed(Sym1, Sym2) then
        Exit;
    end;
  Result := False;
end;

procedure TTransitionMoments.AddToSpectrum(Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; TransitionMoment: TTransitionMoment; PrintLevel: Integer; const Excitation: TExcitation);
var
  D: TMatrix;
begin
  D := TransitionMatrix(Ev1, Ev2, TransitionMoment, PrintLevel);
  try
    AddMatrixToSpectrum(D, Spectrum, Ev1, Ev2, PrintLevel, Excitation);
  finally
    D.Free;
  end;
end;

procedure TTransitionMoments.AddMatrixToSpectrum(D: TMatrix; Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; PrintLevel: Integer; Excitation: TExcitation);
var
  i, f, ColourI: Integer;
  te, wt, str: Double;
  Line, SLine: TLine;
  Swap: Boolean;
  procedure DoLine(var Line: TLine);
  begin
    if PrintLevel >= 0 then begin
      Line.CachedName := '';
      Spectrum.ShowLine(Line, Output, ' ', 0, False);
      Line.CachedName := '';
    end;
    Spectrum.AddLine(Line, Mixture);
  end;
begin
  TStart(AddMatrixToSpectrumTime);
  Line.EigenvaluesL := Ev1;
  Line.EigenvaluesU := Ev2;
  wt := Ev1.StatisticalWeight * (Owner as TMolecule).Abundance * (Owner.Owner as TSpecies).Concentration;
  for i := 1 to Ev1.Size do begin
    { For each lower state }
    Line.IndexL := i;
    ColourI := Colour;
    if ColourI = clNone then
      ColourI := Ev1.DominantBasis[i].State.FindColour;
    for f := 1 to Ev2.Size do begin
      { For each upper state }
      Line.IndexU := f;
      te := Ev2.E[f] - Ev1.E[i];
      Swap := False;
      if Ev2.Manifold.Initial then
        if (Ev1.Manifold = Ev2.Manifold) and (te < 0) then
          Continue { Avoid handling transitions twice if within same manifold }
        else begin
          if te < 0 then begin
            te := -te;
            Swap := True;
          end;
        end
      else if Ev2.Manifold.LowestStateOrigin < Ev1.Manifold.LowestStateOrigin then begin
        te := -te; { Presumably Emission spectrum }
        Swap := True;
      end;
      { TODO : Saturation }
      Str := D[i,f] * wt;
      if Str = 0 then Continue; { Zero intensity transitions are boring }
      Line.Frequency := te;
      Line.Strength := Str*IntensityCorrection(Ev2.J, Ev1.J, Excitation);
      Line.Width := 0;
      if Assigned(Ev1.Widths) then
        Line.Width := Line.Width + Ev1.Widths[i];
      if Assigned(Ev2.Widths) then
        Line.Width := Line.Width + Ev2.Widths[f];
      Line.Colour := ColourI;
      if Line.Colour = clNone then
        Line.Colour := Ev2.DominantBasis[f].State.FindColour;
      if Swap then begin
        SLine := Line;
        SLine.EigenvaluesU := Ev1;
        SLine.EigenvaluesL := Ev2;
        Sline.IndexU := i;
        Sline.IndexL := f;
        DoLine(SLine);
      end else
        DoLine(Line);
    end;
  end;
  TEnd(AddMatrixToSpectrumTime);
end;

function TTransitionMoments.OriginalTransitionMatrix(
            Ev1, Ev2: TEigenvalues; Rank: TQuantumNumber; ATransitionMoment: TTransitionMoment): TMatrix;
{ Transition Moments in original Basis for the given rank only }
var
  tm: TTransitionMoment;
  State1, State2: TState;
  b, bb, t, nTM: Integer;
  Sum: Double;
  TMs: array of TTransitionMoment;
begin
  TStart(MakeTransitionMatrixTime);
  if Ev1.StatisticalWeight <> Ev2.StatisticalWeight then
    raise Exception.CreateFmt('Inconsistent Statistical Weights for %s', [UniqueName]);
  Result := TMatrix.Create(Ev1.Size, Ev2.Size);
  Result.Zero;
  SetLength(TMs, Count);
  nTM := 0;
  { Optimization - pick out relevant transition moments first }
  for t := 0 to Count-1 do
    if (TransitionMoment[t].Rank = Rank) or (TransitionMoment[t] = ATransitionMoment) then begin
      TMs[nTM] := TransitionMoment[t];
      Inc(nTM);
    end;
  if nTM = 0 then
    Exit;
  try
    for b := 1 to Ev1.Size do begin
      State1 := Ev1.Basis[b].State;
      for bb := 1 to Ev2.Size do begin
        Sum := 0;
        State2 := Ev2.Basis[bb].State;
        for t := 0 to nTM-1 do begin { for each transition moment }
          tm := TMs[t];
          if (tm.Bra = State1) and (tm.Ket = State2) then
            Sum := Sum + tm.MatrixElement(Ev1.Basis[b], Ev2.Basis[bb])
          else if (tm.Bra = State2) and (tm.Ket = State1) then
            Sum := Sum + tm.MatrixElement(Ev2.Basis[bb], Ev1.Basis[b]);
        end;
        Result[b,bb] := Sum;
      end;
    end;
  except
    Result.Free;
    raise;
  end;
  TEnd(MakeTransitionMatrixTime);
end;

function TTransitionMoments.TransitionMatrix(Ev1, Ev2: TEigenvalues; TransitionMomentSel: TTransitionMoment; PrintLevel: Integer): TMatrix;
{ Transformed Transition Moments }
var
  D: TMatrix;
  i: Integer;
  Rank: Integer;
begin
  TStart(TransitionMatrixTime);
  if PrintLevel = -10 then
    PrintLevel := Mixture.PrintLevel;
  Rank := -1;
  Result := nil;
  try
    for i := 0 to Count-1 do
      { Each rank must be done separately }
      if (TransitionMoment[i] = TransitionMomentSel)
         or (not Assigned(TransitionMomentSel) and (TransitionMoment[i].Rank <> Rank)) then begin
        Rank := TransitionMoment[i].Rank;
        D := OriginalTransitionMatrix(Ev1, Ev2, Rank, nil);
        if PrintLevel > 2 then
          Ev1.ShowTransitionMatrix(Ev2, D, Format('Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]));
        try
          TransformTransitionMatrix(D, Ev1, Ev2);
          if PrintLevel > 2 then
            Ev1.ShowTransitionMatrix(Ev2, D, Format('Transformed Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]), False);
          D.Sqr;
          if PrintLevel > 3 then
            Ev1.ShowTransitionMatrix(Ev2, D, Format('Squared Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]), False);
        except
          D.Free;
          raise;
        end;
        if Assigned(Result) then begin
          Result.Add(D);
          D.Free;
        end else
          Result := D;
      end;
  except
    Result.Free;
    raise;
  end;
  TEnd(TransitionMatrixTime);
end;

procedure TTransitionMoments.TransformTransitionMatrix(D: TMatrix; Ev1, Ev2: TEigenvalues);
{ Transform Matrix D using eigenvectors - overwrites D }
begin
  TStart(MatrixMultiplyTime);
  Ev1.Multiply(D);
  TEnd(MatrixMultiplyTime);
  TStart(PostMultiplyTime);
  D.PostMultiply(Ev2.C);
  TEnd(PostMultiplyTime);
end;

function TTransitionMoments.Mixture: TMixture;
begin
  Result := Owner.Owner.Owner as TMixture;
end;

class function TTransitionMoments.BraKetClass: TMNodeClass;
begin
  Result := TManifold;
end;

function TTransitionMoments.Molecule: TMolecule;
begin
  Result := Owner as TMolecule;
end;

function TTransitionMoments.PairName: string;
begin
  Result := 'mu';
end;

function TTransitionMoments.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TTransitionMoments.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := ColourHelpString
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TTransitionMoments.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  case AIndex of
    0: Result := TSphericalTransitionMoment;
    1: Result := TCartesianTransitionMoment;
    2: Result := TSpinTransitionMoment;
    3: Result := TOrbitalTransitionMoment;
    4: Result := TRotationalTransitionMoment;
  else
    Result := nil;
  end;
end;

function TTransitionMoments.nChildTypes: Integer;
begin
  Result := 5;
end;

{ TTransitionMoment }

class function TTransitionMoment.OwnerType: TComponentClass;
begin
  Result := TTransitionMoments;
end;

constructor TTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := 'Rank=2,' + SavedProperties;
  if Pos(',Strength=', ','+SavedProperties) = 0 then
    SavedProperties := 'Strength=1,' + SavedProperties;
  DimensionlessParameters := 'Strength,' + DimensionlessParameters;
  inherited Create(AOwner);
end;

constructor TTransitionMoment.Create(AOwner: TComponent; State1Name, State2Name: string);
begin
  Create(AOwner);
  FBraName := State1Name;
  FKetName := State2Name;
end;

function TTransitionMoment.Allowed(Sym1, Sym2: Integer): Boolean;
begin
  Result := (Owner.Owner as TMolecule).Allowed(Self, Sym1, Sym2);
end;

{.$DEFINE CHECKPHASECONSISTENT}
function TTransitionMoment.MatrixElement(ABra, AKet: TBasisState): Double;
{$IFDEF CHECKPHASECONSISTENT}
var
  Temp1: Double;
{$ENDIF}
begin
  { Could also implement this in a descendent of TTransitionMoment }
  Result := ABra.State.TransitionMatrixElement(Self, ABra, AKet);
  {$IFDEF CHECKPHASECONSISTENT}
  Temp1 := AKet.State.TransitionMatrixElement(Self, AKet, ABra);
  if Abs(Temp1-Result) >1e-14 then begin
    WriteLn('Help', Temp1, ' ', Result);
    Result := ABra.State.TransitionMatrixElement(Self, ABra, AKet);
    Temp1 := AKet.State.TransitionMatrixElement(Self, AKet, ABra);
  end;
  {$ENDIF}
  if Magnetic and ((RootNode as TMixture).IntensityUnits > Normalized)  then
    Result := Result * 1e21 * BohrMagneton;
end;

function TTransitionMoment.BraOwner: TMNode;
begin
  Result := (Owner as TTransitionMoments).Bra;
end;

function TTransitionMoment.KetOwner: TMNode;
begin
  Result := (Owner as TTransitionMoments).Ket;
end;

class function TTransitionMoment.BraKetClass: TMNodeClass;
begin
  Result := TState;
end;

function TTransitionMoment.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    if not Assigned(Owner) or (Owner as TTransitionMoments).Mixture.DoubleQuantumNumbers then
      Result := 'Rank of transition - 2 (twice 1) for a normal electric dipole transition.'
    else
      Result := 'Rank of transition - 1 for a normal electric dipole transition.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TTransitionMoment.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative transition (dipole) moment - note that the relative intensity is proportional to the square of this value.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TTransitionMoment.nChildTypes: Integer;
begin
  Result := 0;
end;

{ TCartesianTransitionMoment }

const
  AxisNames: array [0..3] of PChar = ('a', 'b', 'c', nil);

function AxisName(AAxis: TabcAxis): string;
begin
  Result := AxisNames[Ord(AAxis)];
end;

function TCartesianTransitionMoment.GetSettingsNames(AIndex: Integer; var LowestValue:Integer): PPChar;
begin
  if AIndex = 2 then
    Result := @AxisNames
  else
    Result := inherited GetSettingsNames(AIndex,LowestValue);
end;

function TCartesianTransitionMoment.FixupProperties:Exception;
begin
  Result := inherited FixupProperties;
  if not Assigned(Result) and (Rank <> 2) then
    Result := Exception.CreateFmt('%s-%s: Transition Moment Rank Not 1: %d', [Name, AxisName(Axis), Rank]);
end;

function TCartesianTransitionMoment.GetAxis(const Index: Integer): TabcAxis;
begin
  Result := TabcAxis(GetSetting(Index));
end;

procedure TCartesianTransitionMoment.SetAxis(const Index: Integer; const AValue: TabcAxis);
begin
  SetSetting(Index, Ord(AValue));
end;

constructor TCartesianTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := 'Axis=a,' + SavedProperties;
  inherited;
end;

function TCartesianTransitionMoment.PairName: string;
begin
  Result := AxisName(Axis);
end;

function TCartesianTransitionMoment.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 2 then
    Result := 'Axis of transtion moment - a, b or c.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TSphericalTransitionMoment }

function TSphericalTransitionMoment.PairName: string;
begin
  Result := Format('T(%s,%s)', [TMixture(RootNode).QuantumNumberToStr(Rank), TMixture(RootNode).QuantumNumberToStr(Component)]);
end;

function TSphericalTransitionMoment.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 2 then
    if not Assigned(Owner) or (Owner as TTransitionMoments).Mixture.DoubleQuantumNumbers then
      Result := 'Projection quantum number of transition moment - 0 for a parallel transition, 2 (twice 1) for a perpendicular transition.'
    else
      Result := 'Projection quantum number of transition moment - 0 for a parallel transition, 1 for a perpendicular transition.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TEigenvalues }

{
function TEigenvalues.GetBasisState(AIndex: Integer): TBasisState;
begin
  Result := Components[AIndex-1] as TBasisState;
end;
}

procedure TEigenvalues.Fixup;
var
  i: Integer;
begin
  { Make up Basis array }
  SetLength(Basis, Size+1); { Element 0 is unused }
  for i := 1 to Size do
    Basis[i] := Components[i-1] as TBasisState
end;

function TEigenvalues.Size: Integer;
begin
  Result := ComponentCount;
end;

destructor TEigenvalues.Destroy;
begin
  FreeAndNil(FE);
  FreeAndNil(FC);
  FreeAndNil(FWidths);
  FreeAndNil(FDominantBasisIndex);
  inherited Destroy;
end;

function TEigenvalues.DiagonalMatrixElements(MatElemType: TMatElemType; Base: Integer = 0; IncludePerturbations: Boolean = False): TDArray;
var
  i, p: Integer;
begin
  SetLength(Result, Size-Base);
  for i := Base+1 to Size do begin
    Result[i-1-Base] := Basis[i].State.MatrixElement(Basis[i], Basis[i], MatElemType);
    if IncludePerturbations then
      for p := 0 to Manifold.nPerturbations-1 do
        if Manifold.Perturbation[p].Connects(Basis[i].State, Basis[i].State) then
          Result[i-1-Base]  := Result[i-1-Base] + Manifold.Perturbation[p].MatrixElement(Basis[i], Basis[i], MatElemType);
  end;
end;

{
function TEigenvalues.DiagonalMatrixElements(MatElemType: TMatElemType): array of Double;
var
  i: Integer;
begin
  SetLength(Result, Size+1);
  Result := TColumnVector.Create(Size);
  try
    for i := 1 to Size do
      Result[i] := Basis[i].State.MatrixElement(Basis[i], Basis[i], MatElemType);
  except
    Result.Free;
    raise;
  end;
end;
}
function TEigenvalues.MatrixElements(MatElemType: TMatElemType; IncludePerturbations: Boolean): TMatrix;
var
  p, b, bb: Integer;
  Current: TState;
  Sum: Double;
begin
  Result := nil;
  if Size = 0 then Exit;
  Result := TMatrix.Create(Size, Size);
  try
    for b := 1 to Size do begin
      Current := Basis[b].State;
      for bb := 1 to b do begin
        if Current = Basis[bb].State then
          Sum := Current.MatrixElement(Basis[b], Basis[bb], MatElemType)
        else
          Sum := 0;
        if IncludePerturbations then begin
          TStart(PerturbationMatrixElementTime);
          for p := 0 to Manifold.nPerturbations-1 do
            if Manifold.Perturbation[p].Connects(Current, Basis[bb].State) then
              Sum := Sum + Manifold.Perturbation[p].MatrixElement(Basis[b], Basis[bb], MatElemType);
          TEnd(PerturbationMatrixElementTime);
        end;
        Result[b,bb] := Sum;
        Result[bb,b] := Sum;
      end;
    end;
  except
    Result.Free;
    raise;
  end;
end;

procedure TEigenvalues.ShowBasisMatrix(M: TMatrix; const Title: string);
begin
  ShowTransitionMatrix(Self, M, Title);
end;

procedure TEigenvalues.ShowTransitionMatrix(Ev: TEigenValues; M: TMatrix; const Title: string; ShowBasis: Boolean);
var
  b, bb, sw: Integer;
  LWidths: array of Integer;
  s: string;
  procedure ShowState(Ev: TEigenValues);
  begin
    Write(Ev.Basis[1].State.Manifold.Name,', J = ', Ev.Manifold.Mixture.QuantumNumberToStr(Ev.J), ', Symmetry = ', Manifold.Molecule.SymName(Ev.Sym));
  end;
  function SLabel(i: Integer; Ev: TEigenValues): string;
  begin
    if ShowBasis then
      Result := Ev.Basis[i].BasisLabel
    else
      Result := Ev.DominantBasis[i].StateLabel;
  end;
begin
  { Print Hamiltonian or other Matrix, including basis state labels }
  if Size = 0 then Exit;
  Write(Title,' ');
  if Ev <> Self then Write('<');
  ShowState(Self);
  if Ev <> Self then begin
    Write('| |');
    ShowState(Ev);
    Write('>');
  end;
  sw := 4; { Work out width of Row labels }
  for b := 1 to Size do
    sw := Max(sw, Length(SLabel(b,Self))+2);
  WriteLn;
  Write(' ':sw);
  SetLength(LWidths, Ev.Size+1); { Index 0 is not used }
  for b := 1 to Ev.Size do begin
    s := '|'+SLabel(b,Ev)+'>';
    LWidths[b] := Length(s);
    Write(s);
  end;
  WriteLn;
  for b := 1 to Size do begin
    Write('<'+SLabel(b,Self)+'|':sw);
    for bb := 1 to Ev.Size do
      Write(M[b,bb]:LWidths[bb]:2);
    WriteLn;
  end;
  WriteLn;
end;

procedure TEigenvalues.Multiply(M: TMatrix);
{ Multiply matrix M by eigenvalues
  ( Actually Pre multiply by transpose of eigenvector matrix )
    so Requires match of Rows of M with size of Basis }
var
  W: array of Double;
  i, ii, iii: Integer;
  Sum: Double;
  PM, PC: PDouble;
  N: Integer;
begin
  if M.RowCount <> Size then
    raise Exception.Create('TEigenvalues.Multiply: Size mismatch');
  SetLength(W, Size+1); { Index 0 is not used }
  {$DEFINE ALT}
  {$IFDEF ALT}
  { This ordering is significantly quicker, as transition matrices are typically
    banded and thus not very dense }
  N := Size;
  for i := 1 to M.ColCount do begin
    for ii := 1 to M.RowCount do begin
      W[ii] := M[ii,i];
      M[ii,i] := 0;
    end;
    for iii := 1 to M.RowCount do begin
      Sum := W[iii]; { = original M[iii,i] }
      if Sum <> 0 then begin
        PM := M.Item[1,i];
        PC := C.Item[iii,1];
        for ii := 1 to M.RowCount do begin
          {$IFDEF SIMPLE}
          M[ii,i] := M[ii,i] + Sum*C[iii,ii];
          {$ELSE}
          PM^ := PM^ + Sum*PC^;
          Inc(PM);
          Inc(PC,N);
          {$ENDIF}
        end;
      end;
    end;
  end;
  {$ELSE}
  for i := 1 to M.ColCount do begin
    for ii := 1 to Size do begin
      Sum := 0;
      for iii := 1 to Size do
        Sum := Sum + C[iii,ii] * M[iii,i];
      W[ii] := Sum;
    end;
    for ii := 1 to Size do
      M[ii,i] := W[ii];
  end;
  {$ENDIF}
end;

function TEigenvalues.Transform(M: TMatrix): TColumnVector;
{ Return diagnonal elements of M transformed by eigenvectors }
{ M is overwritten }
var
  i, ii: Integer;
  Sum: Double;
begin
  Multiply(M);
  Result := TColumnVector.Create(Size);
  try
    for i := 1 to Size do begin
      Sum := 0;
      for ii := 1 to Size do
        Sum := Sum + M[i, ii] * C[ii, i];
      Result[i] := Sum;
    end;
  except
    M.Free;
    raise;
  end;
end;

procedure TEigenvalues.SortFrom(Start: Integer);
var
  V: TDarray;
begin
  Fixup;
  if Size-Start <= 1 then
    Exit { 1 element must be sorted }
  else begin
    V := DiagonalMatrixElements(meHrot, Start, True);
    SortBasisOn(V, False, Start);
  end;
end;

type
  TBasisShellSorter = class(TShellSorter)
    ByState: Boolean;
    Eigenvalues: TEigenvalues;
    Base: Integer;
    V: TDArray;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

function TBasisShellSorter.InOrder(I,J:Integer):Boolean;
var
  Diff: Integer;
begin
  if ByState then begin
    Diff := Eigenvalues.Basis[I-Base+1].StateIndex - Eigenvalues.Basis[J-Base+1].StateIndex;
    if Diff <> 0 then begin
      Result := Diff <= 0;
      Exit;
    end;
  end;
  Result := V[I] <= V[J];
end;

procedure TBasisShellSorter.Swap(I,J: Integer);
begin
  SwapValues(V[I], V[J], SizeOf(Double));
  SwapValues(Eigenvalues.Basis[I+Base+1], Eigenvalues.Basis[J+Base+1], SizeOf(TBasisState));
end;

procedure TEigenvalues.SortBasisOn(V: TDarray; ByState: Boolean; Base: Integer);
{ Sort the basis states starting from Base;
  Given V[b] for each basis state b, shuffle so that V[b] is in ascending order }
var
  Sorter: TBasisShellSorter;
  i: Integer;
begin
  Sorter := TBasisShellSorter.Create;
  Sorter.ByState := ByState;
  Sorter.Eigenvalues := Self;
  Sorter.Base := Base;
  Sorter.V := V;
  Sorter.Sort(Size-Base);
  for i := Base to Size-1 do
    Basis[i+1].ComponentIndex := i;
  Sorter.Free;
end;

procedure TEigenvalues.DoLimitedEigenSearch;
var
  BasisStateIndex, nBasisStates, nContr: Integer;
  Ind: array of Integer;
  Contr: array of Double;
  IContr: array of Integer;
  Sum: Double;
  i, ii, k: Integer;
  {$IFDEF OLD_ALGORITHM}
  FirstFailed: Integer;
  m: array of Double;
  b: array of Integer; { Index of last basis state used from each state }
  u: array of Boolean; { Which basis states have already been assigned }
  {$ENDIF}
begin
  { Leave Individual basis states in original order,
    but make sure state are correct }
  if (LastStateIndex <= 1) or (Size <= 1) then begin
    FreeAndNil(FDominantBasisIndex); { Leave as is }
    Exit;
  end;
  if Assigned(FDominantBasisIndex) then Exit;
  FDominantBasisIndex := TIntColumnVector.Create(Size);
  FDominantBasisIndex.Clear;
  { For each basis state index:
    - find out ouw many basis states it corresponds to
    - select exactly that many on the basis of largest coefficients }
  SetLength(Ind, Size);   { Index into basis for this BasisStateIndex }
  SetLength(Contr, Size);
  SetLength(IContr, Size);
  for BasisStateIndex := 0 to LastStateIndex-1 do begin
    // WriteLn('BasisStateIndex=', BasisStateIndex);
SetLength(Ind, Size);   { Index into basis for this BasisStateIndex }
    nBasisStates := 0;
    for k := 1 to Size do
      if Basis[k].StateIndex = BasisStateIndex then begin
        Ind[nBasisStates] := k;
        Inc(nBasisStates);
      end;
    if nBasisStates = 0 then
      Continue;
SetLength(Ind, nBasisStates);   { Index into basis for this BasisStateIndex }
SetLength(Contr, nBasisStates);
SetLength(IContr, nBasisStates);
    nContr := 0; { Contr and IContr keep track of the most siginificant contributions }
    for k := 1 to Size do
      if FDominantBasisIndex[k] = 0 then begin
        { Work out how much eigenvalue k has from this particular BasisStateIndex }
        Sum := 0;
        for i := 0 to nBasisStates-1 do
          Sum := Sum + Sqr(C[Ind[i],k]);
        { Insert into sorted list of the nBasisStates largest contributions }
        // WriteLn(k, ' ', Sum);
        { Find insert position in list }
        i := 0;
        while (i < nContr) and (Sum > Contr[i]) do
          Inc(i);
        if nContr < nBasisStates then begin
          { Not enough numbers in list yet, so grow list
            - Some values to shift up }
          for ii := nContr downto i+1 do begin
            Contr[ii] := Contr[ii-1];
            IContr[ii] := IContr[ii-1];
          end;
          Contr[i] := Sum;
          IContr[i] := k;
          Inc(nContr);
        end else if i > 0 then begin
          Dec(i);
          { Replace a current entry, shifting lower values down }
          for ii := 0 to i-1 do begin
            Contr[ii] := Contr[ii+1];
            IContr[ii] := IContr[ii+1];
          end;
          Contr[i] := Sum;
          IContr[i] := k;
        end;
      end;
    if nContr <> nBasisStates then
      raise Exception.CreateFmt('DoLimitedEigenSearch %d %d', [nContr, nBasisStates]);
    Sort(IContr);
    for i := 0 to nBasisStates-1 do
      FDominantBasisIndex[IContr[i]] := Ind[i];
  end;
  {$IFDEF OLD_ALGORITHM}
  SetLength(m, LastStateIndex);
  SetLength(b, LastStateIndex);
  SetLength(u, Size+1);
  for i := 1 to Size do
    u[i] := False;
  for i := 0 to LastStateIndex-1 do
    b[i] := 0;
  FirstFailed := 0;
  for i := 1 to Size do begin
    { Find the largest contributing state for each eigenvector }
    { Start by finding how much each state contributes }
    for ii := 0 to LastStateIndex-1 do begin
      m[ii] := 0;
      for k := 1 to Size do
        if Basis[k].StateIndex = ii then { by picking out each basis state }
          m[ii] := m[ii] + Sqr(C[k,i]);
    end;
    { Now find largest contribution }
    k := 0;
    for ii := 1 to LastStateIndex-1 do
      if m[ii] > m[k] then
        k := ii;
    { Using State number k; find next available basis for this state }
    FDominantBasisIndex[i] := 0;
    while (FDominantBasisIndex[i] = 0) and (b[k] < Size) do begin
      Inc(b[k]);
      if Basis[b[k]].StateIndex = k then begin
        FDominantBasisIndex[i] := b[k];
        u[b[k]] := True;
        Break;
      end;
    end;
    if FDominantBasisIndex[i] = 0 then begin
      WriteLn('Limited EigenSearch fails for ', Manifold.Name, ', J=', Self.J:4, ' Sym=', Sym:3, ', #=', i);
      if FirstFailed = 0 then
        FirstFailed := i;
    end;
  end;
  if FirstFailed = 0 then Exit;
  { Patch Up Failures }
  for i := 1 to Size do
    if not u[i] then begin
      FDominantBasisIndex[FirstFailed] := i;
      repeat
        Inc(FirstFailed);
        if FirstFailed > Size then
          Exit;
      until FDominantBasisIndex[FirstFailed] = 0;
    end;
  raise
    Exception.Create('DoLimitedEigenSearch failure');
  {$ENDIF}
end;

procedure TEigenvalues.DoEigenSearch;
var
  i, ii, iimax, k: Integer;
  cmax: Double;
begin
  if Assigned(FDominantBasisIndex) then Exit;
  FDominantBasisIndex := TIntColumnVector.Create(Size);
  { Initial Permutation }
  for i := 1 to Size do
    FDominantBasisIndex[i] := i;
  { Now scan eigenvalues for largest coefficient }
  for i := 1 to Size do begin
    iimax := 1;
    cmax := Abs(C[1,i]);
    for ii := 2 to Size do
      if Abs(C[ii,i]) > cmax then begin
        iimax := ii;
        cmax := Abs(C[ii,i]);
      end;
    if iimax <> FDominantBasisIndex[i] then begin
      k := FDominantBasisIndex.IndexOf(iimax);
      if k >= i then begin
        { Swap indices - this ensures result is always a permutation;
          This may give slightly odd results where the mixing is large }
        FDominantBasisIndex[k] := FDominantBasisIndex[i];
        FDominantBasisIndex[i] := iimax;
      end;
    end;
  end;
end;

function TEigenvalues.GetDominantBasisState(AIndex: Integer): TBasisState; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  if Assigned(FDominantBasisIndex) then
    AIndex := FDominantBasisIndex[AIndex];
  Result := Basis[AIndex]
end;

procedure TEigenvalues.ShowEigenvalues;
var
  b, bb, sw: Integer;
  s: string;
  BWidths: array of Integer;
begin
  WriteLn('Eigenvalues for ' + Manifold.Name + ', J = ', Manifold.Mixture.QuantumNumberToStr(J), ', Symmetry = ', Manifold.Molecule.SymName(Sym));
  sw := 1; { Maximum width of state label }
  for b := 1 to Size do
    sw := Max(sw, Length(DominantBasis[b].StateLabel));
  Write(' ':18+sw);
  if Assigned(Widths) then Write(' ':10);
  SetLength(BWidths, Size+1);
  for b := 1 to Size do begin
    s := Basis[b].BasisLabel;
    BWidths[b] := Max(10, Length(s)+2);
    Write('|'+s+'>':10);
  end;
  WriteLn;
  for b := 1 to Size do begin
    Write(b:3, ' ', DominantBasis[b].StateLabel:sw+1, E[b]:13:3);
    if Assigned(Widths) then Write(Widths[b]:10:5);
    for bb := 1 to Size do
      Write(C[bb,b]:BWidths[bb]:6);
    WriteLn;
  end;
  WriteLn;
end;

procedure TEigenvalues.MakeHamiltonian(BasisOrder: TBasisOrder; IncludePerturbations: Boolean);
var
  V: TDarray;
begin
  NoPerturbations := not IncludePerturbations;
  if BasisOrder <> boDefault then begin
    V := DiagonalMatrixElements(meHrot, 0, IncludePerturbations);
    SortBasisOn(V, BasisOrder = boStateDiagonal);
  end;
  E := TColumnVector.Create(Size);
  C := MatrixElements(meHrot, IncludePerturbations);
end;

procedure TEigenvalues.Diagonalize(EigenvectorsWanted: Boolean);
begin
  Eigen(E, C, EigenvectorsWanted);
end;

procedure TEigenvalues.SetC(const Value: TMatrix);
begin
  FC.Free;
  FC := Value;
end;

procedure TEigenvalues.SetE(const Value: TColumnVector);
begin
  FE.Free;
  FE := Value;
end;

procedure TEigenvalues.SetWidths(const Value: TColumnVector);
begin
  FWidths.Free;
  FWidths := Value;
end;

{ TBasisState }

function TBasisState.F: TQuantumNumber;
begin
  Result := (Owner as TEigenvalues).J;
end;

function TBasisState.Sym: Integer;
begin
  Result := (Owner as TEigenvalues).Sym;
end;

function TBasisState.BasisLabel: string;
var
  FF: TQuantumNumber;
  i: Integer;
begin
  if State.Manifold.Molecule.nManifolds > 1 then
    Result := State.Manifold.Name + ' '
  else
    Result := '';
  Result := Result + State.Name;
  if State.Manifold.Molecule.nNuclei > 0 then begin
    FF := F;
    for i := State.Manifold.Molecule.nNuclei-2 downto 0 do begin
      Dec(FF,HyperfineState[i+1]);
      Result := Result + Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(FF)]);
    end;
    Result := Result + Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(J)]);
  end;
  if Phase > 0 then
    Result := Result + ' +'
  else if Phase < 0 then
    Result := Result + ' -'
  else
    Result := Result + '  ';
end;

function TBasisState.StateLabel: string;
begin
  if State.Manifold.Molecule.nManifolds > 1 then
    Result := State.Manifold.Name + ' '
  else
    Result := '';
  Result := Result + State.Name;
end;

function TBasisState.StateLabelKey: string;
begin
  Result := 'Name';
end;

function TBasisState.HyperfineStateLabel: string;
var
  FF: TQuantumNumber;
  i: Integer;
begin
  if State.Manifold.Molecule.nNuclei > 0 then begin
    FF := F;
    Result := Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(F)]);
    for i := State.Manifold.Molecule.nNuclei-2 downto 0 do begin
      Dec(FF,HyperfineState[i+1]);
      Result := Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(FF)]) + Result;
    end;
  end else
    Result := '';
end;

function TBasisState.HyperfineStateLabelKey: string;
var
  i: Integer;
begin
  if State.Manifold.Molecule.nNuclei > 0 then begin
    Result := '  F';
    for i := State.Manifold.Molecule.nNuclei-2 downto 0 do
      Result := Format(' F%d', [i+1]) + Result;
  end else
    Result := '';
end;

function TBasisState.TxName(Upper, Lower: TBasisState): string;
begin
  Result := '';
end;

function TBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := '';
end;

function TBasisState.BasisLabelKey: string;
var
  i: Integer;
begin
  Result := 'Name';
  if State.Manifold.Molecule.nNuclei > 0 then begin
    for i := State.Manifold.Molecule.nNuclei-2 downto 0 do
      Result := Result + Format(' F%d', [i+1]);
    Result := Result + '  J';
  end;
  Result := Result + ' +-';
end;

function TBasisState.TrackIndex: Integer;
begin
  Result := 0; { Default is all states in same class }
end;

{ TSpectrum }

procedure TSpectrum.Clear;
begin
  FnLines := 0;
  ITotal := 0;
end;

procedure TSpectrum.AddIn(LineList: TLineList; Mixture: TMixture);
var
  i, c: Integer;
begin
  PushProgress('Accumulating Spectrum', LineList.nLines div 100);
  if Mixture.IntensityUnits >= Normalized then
    Mixture.PreCalculatePartitionFunctions;
  c := 0;
  TStart(AddInLinesTime);
  for i := 0 to LineList.nLines - 1 do begin
    AddLine(LineList.Line[i], Mixture);
    Dec(c);
    if c < 0 then begin
      SetProgress(i div 100);
      c := 100;
    end;
  end;
  TEnd(AddInLinesTime);
  PopProgress;
end;

constructor TSpectrum.Create(AUnits: TResonanceUnits);
begin
  inherited Create;
  Colour := clNone;
  Clear;
  Units := AUnits;
end;

procedure TSpectrum.AddLine(Frequency, Strength, Width: Double; Colour: TColor; AUnits: TResonanceUnits);
var
  Line: TLine;
begin
  Line.Frequency := Frequency;
  Line.Strength := Strength;
  Line.Width := Width;
  Line.EigenvaluesL := nil;
  Line.EigenvaluesU := nil;
  Line.Colour := Colour;
  if AUnits <> Units then
    ConvertLine(Line, AUnits, Units);
  AddLine(Line, nil);
end;


procedure TSpectrum.ShowLine(var Line: TLine; var F: TextFile; Separator: Char; Threshold: Double; LinFormat: Boolean);
var
  Intensity: Double;
  l: Integer;
  Ev1, Ev2: TEigenvalues;
  Mixture: TMixture;
  FirstField: Boolean;
  Lab: string;
  procedure ShowField(const Field: string);
  begin
    if FirstField then
      FirstField := False
    else
      Write(F,Separator);
    if Separator = ',' then
      if Pos('"', Field) <> 0 then
        Write(F,AnsiQuotedStr(Field,'"'))
      else
        Write(F,Field)
    else
      Write(F,Field);
  end;
  procedure ShowInt(V: Integer);
  begin
    ShowField(IntToStr(V));
  end;
  procedure ShowD(V: Double);
  begin
    ShowField(FloatToStr(V));
  end;
begin
  Ev2 := Line.EigenvaluesU;
  Ev1 := Line.EigenvaluesL;
  if not Assigned(Ev1) or not Assigned(Ev2) then
    Exit;
  Mixture := Ev2.Manifold.Mixture;
  Intensity := Mixture.LineIntensity(Line, Temperature, Tvib);
  ITotal := ITotal + Intensity;
  if Intensity < Threshold then
    Exit;
  l := Length(Mixture.LongestManifoldName);
  FirstField := True;
  if not HeadingPrinted then
    if Separator = ' ' then begin
      Write(F, Copy('Molecule',1,Length(Mixture.LongestMoleculeName)):Length(Mixture.LongestMoleculeName), ' ',
              'M':l, ''' J''S'' #''',
              'M':l, '" J#S# #"',
              '    Frequency');
      if LinFormat then
        Write(F, ' Wt I      ')
      else
        Write(F, ' I               Eupper        Elower S      ');
      WriteLn(F, '  Width')
    end else begin
      ShowField('Molecule');
      ShowField('Upper Manifold');
      ShowField('J''');
      ShowField('Sym''');
      ShowField('#''');
      ShowField('Lower Manifold');
      ShowField('J"');
      ShowField('Sym"');
      ShowField('#"');
      ShowField('Frequency');
      if LinFormat then
        ShowField('Wt');
      ShowField('Intensity');
      if not LinFormat then begin
        ShowField('Eupper');
        ShowField('Elower');
        ShowField('Strength');
      end;
      ShowField('Width');
      ShowField('Branch');
      ShowField('Label');
      Writeln(F);
    end;
  HeadingPrinted := True;
  FirstField := True;

  if Separator = ' ' then begin
    Write(F, Ev1.Manifold.Molecule.Name:Length(Mixture.LongestMoleculeName), ' ',
             Ev2.Manifold.Name:l, AtLeast(Ev2.J,3), AtLeast(Ev2.Sym,2), AtLeast(Line.IndexU,3), ' ',
             Ev1.Manifold.Name:l, AtLeast(Ev1.J,3), AtLeast(Ev1.Sym,2), AtLeast(Line.IndexL,3),
             AtLeast(Line.Frequency,14,Mixture.Precision));
    if LinFormat then
      Write(F, ' - ');
    Write(F, DtosMA(Intensity,9,-4));
    if not LinFormat then
      Write(F, AtLeast(Ev2.E[Line.IndexU],14,Mixture.Precision),
               AtLeast(Ev1.E[Line.IndexL],14,Mixture.Precision),
               DtosMA(Line.Strength,9,-4));
    WriteLn(F, DtosMA(Line.Width,9,-4), ' : ', MakeTxName(Line));
  end else begin
    ShowField(Ev1.Manifold.Molecule.Name);
    ShowField(Ev2.Manifold.Name);
    ShowInt(Ev2.J);
    ShowInt(Ev2.Sym);
    ShowInt(Line.IndexU);
    ShowField(Ev1.Manifold.Name);
    ShowInt(Ev1.J);
    ShowInt(Ev1.Sym);
    ShowInt(Line.IndexL);
    ShowD(Line.Frequency);
    if LinFormat then
      ShowField('-');
    ShowD(Intensity);
    if not LinFormat then begin
      ShowD(Ev2.E[Line.IndexU]);
      ShowD(Ev1.E[Line.IndexL]);
      ShowD(Line.Strength);
    end;
    ShowD(Line.Width);
    Lab := MakeTxName(Line);
    l := Pos(':',Lab);
    if l > 0 then begin
      ShowField(Trim(Copy(Lab,1,l-1)));
      ShowField(Trim(Copy(Lab,l+1,Length(Lab))));
    end else begin
      ShowField('');
      ShowField(Trim(Lab));
    end;
    Writeln(F);
  end;
end;

procedure TSpectrum.ShowFooter(var F: TextFile; Separator: Char);
begin
  WriteLn(F);
  if Separator = ' ' then
    WriteLn(F, nLines, ' Lines, Total intensity = ', Itotal:14:4)
  else
    WriteLn(F, 'Lines', Separator, nLines, Separator, 'Total Intensity', Separator, ITotal);
end;

{ TLineList }

function TLineList.GetLine(AIndex: Integer): TLine;
begin
  Result := FLines[AIndex];
end;

procedure TLineList.AddLine(const Line: TLine; Mixture: TMixture);
begin
  if nLines >= High(FLines) then begin
    { Optimization - if small buffer fills up, take larger steps }
    if High(FLines) > 0 then
      SetLength(FLines, High(FLines)+50000)
    else
      SetLength(FLines, High(FLines)+10000);
  end;
  FLines[nLines] := Line;
  if Assigned(Mixture) and (Mixture.Units <> Units) then
    ConvertLine(FLines[nLines], Mixture.Units, Units);
  Inc(FnLines);
  if Line.Frequency < FMin then
    FMin := Line.Frequency;
  if Line.Frequency > FMax then
    FMax := Line.Frequency;
end;

procedure TLineList.Clear;
begin
  inherited;
  FMax := -1.0e99;
  FMin := 1.0e99;
end;

{$IFNDEF CONSOLE}
procedure TLineList.FindLines(Centre, Range: Double; AUnits: TResonanceUnits);
var
  i: Integer;
  p1, p2: Double;
begin
  p1 := Convert(Centre-Range, AUnits, Units);
  p2 := Convert(Centre+Range, AUnits, Units);
  Centre := (p1+p2)/2;
  Range := Abs(p2-Centre);
  for i := 0 to nLines-1 do
    if Abs(FLines[i].Frequency-Centre) <= Range then
      LineForm.AddLine(FLines[i],Units);
end;
{$ENDIF}

procedure TLineList.ShowLine(LineNumber: Integer; var F: TextFile; Separator: Char;  Threshold:Double; LinFormat: Boolean);
begin
  ShowLine(FLines[LineNumber], F, Separator, Threshold, LinFormat);
end;

procedure TSpectrum.SetHeadingPrinted(const Value: Boolean);
begin
  if not Value then
    ITotal := 0;
  FHeadingPrinted := Value;
end;

function TLineList.LargestIntensity: Double;
var
  Mixture: TMixture;
  Intensity: Double;
  i: Integer;
begin
  Result := 0;
  if nLines = 0 then
    Exit;
  Mixture := Line[0].EigenvaluesU.Manifold.Mixture;
  for i := 0 to nLines-1 do begin
    Intensity := Mixture.LineIntensity(Line[i], Temperature, Tvib);
    if Intensity > Result then
      Result := Intensity;
  end;
end;

{ TStreamableMNode }

class function TStreamableMNode.OwnerType: TComponentClass;
begin
  Result := nil;
end;

procedure TStreamableMNode.LoadFromStream(Stream: TStream; AFileName: string);
var
  XMLDocument: TXMLDocument;
begin
  ReadXMLFile(XMLDocument, Stream, AFileName);
  try
    ReadProperties(XMLDocument.FirstChild);
    CheckProperties;
    Name := AFileName;
    Modified := False;
  finally
    XMLDocument.Free;
  end;
end;

procedure TStreamableMNode.LoadFromFile(AFileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(AFilename, fmOpenRead);
  try
    LoadFromStream(Stream, AFileName);
  finally
    Stream.Free;
  end;
end;

procedure TStreamableMNode.MergeFromStream(Stream: TStream; AFileName: string);
var
  XMLDocument: TXMLDocument;
begin
  ReadXMLFile(XMLDocument, Stream, AFileName);
  try
    ReadChildProperties(XMLDocument.FirstChild);
    CheckProperties;
  finally
    XMLDocument.Free;
  end;
end;

procedure TStreamableMNode.SaveToFile(AFileName: string);
var
  XMLDocument: TXMLDocument;
begin
  FixupProperties; { Discard error }
  XMLDocument := TXMLDocument.Create;
  try
    Name := AFileName;
    WriteProperties(XMLDocument);
    WriteXMLFile(XMLDocument, AFilename);
    Modified := False;
  finally
    XMLDocument.Free;
  end;
end;

function TStreamableMNode.GetUnits(const Index: Integer): TResonanceUnits;
begin
  Result := TResonanceUnits(GetSetting(Index));
end;

procedure TStreamableMNode.SetUnits(const Index: Integer; const AValue: TResonanceUnits);
begin
  SetSetting(Index, Ord(AValue));
end;

function TStreamableMNode.UnitsName: string;
begin
  Result := NameOfUnit(GetUnits(1));
end;

function TStreamableMNode.UnitsTypeName: string;
begin
  Result := TypeOfUnit(GetUnits(1));
end;

procedure TStreamableMNode.SetIntensityUnits(const Index: Integer; const AValue: TIntensityUnits);
begin
  SetSetting(Index, Ord(AValue));
end;

function TStreamableMNode.GetIntensityUnits(const Index: Integer): TIntensityUnits;
begin
  Result := TIntensityUnits(GetSetting(Index));
end;

{ TPerturbation }

class function TPerturbation.BraKetClass: TMNodeClass;
begin
  Result := TState;
end;

function TPerturbation.Connects(State1, State2: TState): Boolean;
begin
  Result := ((State1 = Bra)and (State2 = Ket))
            or ((State2 = Bra)and (State1 = Ket));
end;

function TPerturbation.MatrixElement(State1, State2: TBasisState; MatElemType: TMatElemType): Double;
begin
  Result := State1.State.PerturbationMatrixElement(Self, State1, State2, MatElemType)
end;

class function TPerturbation.OwnerType: TComponentClass;
begin
  Result := TManifold;
end;

function TPerturbation.PairName: string;
begin
  Result := 'H';
end;

function TPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Index, starting from 1, of nuclear spin involved in perturbation; 0 (default) for those not involving a nuclear spin.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TPerturbation.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Size of perturbation.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TPerturbation.nChildTypes: Integer;
begin
  Result := 0;
end;

{ TMixtureMNode }

function TMixtureMNode.BoltzmannConstant: Double;
var
  Units: TEnergyUnits;
begin
  if RootNode is TMixture then
    Units := TMixture(RootNode).Units
  else
    Units := cm1;
  Result := BoltzmannConstants[Units];
end;

function TMixtureMNode.DoubleQuantumNumbers: Boolean;
begin
  if RootNode is TMixture then
    Result := TMixture(RootNode).DoubleQnos
  else
    Result := True; { TODO: Raise }
end;

function TMixtureMNode.Population(Eigenvalues: TEigenvalues; i: Integer;
  Temperature, Tvib: Double): Double;
var
  EBottom: Double;
  State: TState;
  function bf(T, E: Double): Double;
  begin
    Result := Exp(-E/T/BoltzmannConstant);
  end;
begin
  EBottom := Eigenvalues.Manifold.Molecule.LowestOrigin;
  if Tvib > 0 then begin
    { Vibrational temperature calc }
    State := Eigenvalues.DominantBasis[i].State;
    Result := bf(Tvib, State.StateOrigin-EBottom);
    if Temperature > 0 then
      Result := Result * bf(Temperature, Eigenvalues.E[i]-State.StateOrigin);
  end else if Temperature > 0 then begin
    { Normal Boltzman calculation }
    Result := bf(Temperature, Eigenvalues.E[i] - EBottom);
  end else
    Result := 1;
//  { TODO: else if Temperature < 0 then use read in populations }
end;

function TMixtureMNode.LineIntensity(const Line: TLine; Temperature, Tvib: Double): Double;
begin
  if Line.EigenvaluesU.Manifold.Initial then
    Result := Population(Line.EigenvaluesU, Line.IndexU, Temperature, Tvib)
  else
    Result := 0;
  if Line.EigenvaluesL.Manifold.Initial then
    { Use population difference - essential for microwave transitions }
    Result := Abs(Population(Line.EigenvaluesL,Line.IndexL,Temperature,Tvib)-Result);
  Result := Result * Line.Strength;
  with RootNode as TMixture do begin
    if IntensityUnits = Squared then
      Result := Sqr(Result);
    if IntensityUnits >= Normalized then
      Result := Result / Line.EigenvaluesL.Manifold.Molecule.PartitionFunction(Temperature, Tvib);
    if IntensityUnits = HzperMolecule then
      { Emission }
      Result := Result * IntPower(Abs(Line.Frequency),3)
                * 3 { Convert to unpolarised light } { TODO: Required for A? }
                * IntensityFactor(Units, IntensityUnits)
    else if IntensityUnits > Normalized then
      Result := Result * Abs(Line.Frequency)
                * 3 { Convert to unpolarised light }
                { TODO: different scaling factor for Quadrupole transitions }
                * IntensityFactor(Units, IntensityUnits);
  end;
  if Result > MaxSingle then
    Result := MaxSingle;
end;

//function TMixureMNode.LineIntensity(const Line: TLine; Temperature, Tvib: Double): Double;
//var
//  EBottom, BpopL, BpopU: Double;
//  Lower, Upper: TState;
//  function bf(T, E: Double): Double;
//  begin
//    Result := Exp(-E/T/BoltzmannConstant);
//  end;
//begin
//  Result := Line.Strength;
//  EBottom := Line.EigenvaluesL.Manifold.Molecule.LowestOrigin;
//  BpopU := 0;
//  { Vibrational temperature calc }
//  if Tvib > 0 then begin
//    Lower := Line.EigenvaluesL.DominantBasis[Line.IndexL].State;
//    BpopL := bf(Tvib, Lower.StateOrigin - EBottom);
//    Upper := Line.EigenvaluesU.DominantBasis[Line.IndexU].State;
//    if Upper.Manifold.Initial then
//      BpopU := bf(Tvib, Upper.StateOrigin - EBottom);
//    if Temperature > 0 then begin
//      BpopL := BpopL * bf(Temperature, Line.EigenvaluesL.E[Line.IndexL] - Lower.StateOrigin);
//      if Upper.Manifold.Initial then
//        { Use population difference - essential for microwave transitions }
//        BpopU :=  BpopU * bf(Temperature, Line.EigenvaluesU.E[Line.IndexU] - Upper.StateOrigin);
//      Result := Result * Abs(BpopL-BPopU);
//    end;
//  end else if Temperature > 0 then begin
//    { Normal Boltzman calculation }
//    BpopL := bf(Temperature, Line.EigenvaluesL.E[Line.IndexL] - EBottom);
//    if Line.EigenvaluesU.Manifold.Initial then
//      { Use population difference - essential for microwave transitions }
//      BpopU := bf(Temperature, Line.EigenvaluesU.E[Line.IndexU] - EBottom);
//    Result := Result * Abs(BpopL-BPopU);
//  end;
//  { TODO: else if Temperature < 0 then use read in populations }
//end;

function TMixtureMNode.Precision: Integer;
begin
  if RootNode is TMixture then
    Result := TMixture(RootNode).Precision
  else
    Result := 4; { TODO: Raise }
end;

function TMixtureMNode.QuantumNumberToStr(QuantumNumber: TQuantumNumber): string;
begin
  if DoubleQuantumNumbers then
    Result := IntToStr(QuantumNumber)
  else if Odd(QuantumNumber) then
    Result := IntToStr(QuantumNumber div 2) + '.5'
  else
    Result := IntToStr(QuantumNumber div 2);
end;

{ TNucleus }

constructor TNucleus.Create(AOwner: TComponent);
begin
  { Default mechanism overwrites name and spin, so defeat during creation }
  FCreating := True;
  inherited Create(AOwner);
  FCreating := False;
end;

function TNucleus.Molecule: TMolecule;
begin
  Result := Owner.Owner.Owner as TMolecule;
end;

class function TNucleus.OwnerType: TComponentClass;
begin
  Result := TState;
end;

procedure TNucleus.SetName(const Value: TComponentName);
begin
  inherited;
  if not FCreating then
    Molecule.Nuclei[ComponentIndex].Name := Value;
end;

function TNucleus.GetName: TComponentName;
begin
  Result := Molecule.Nuclei[ComponentIndex].Name;
  if Result = '' then
    Result := Format('Nucleus%d',[ComponentIndex+1]);
end;

function TNucleus.GetSpin(const Index: Integer): Integer;
begin
  Result := Molecule.Nuclei[ComponentIndex].Spin;
end;

procedure TNucleus.SetSpin(const Index, Value: Integer);
begin
  inherited SetSetting(Index, Value);
  if not FCreating then
    Molecule.Nuclei[ComponentIndex].Spin := Value;
end;

function TNucleus.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Nuclear spin'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TNucleus.nChildTypes: Integer;
begin
  Result := 0;
end;

{ TMagneticTransitionMoment }

function TMagneticTransitionMoment.PairName: string;
begin
  if Rank = 2 then
    Result := Format('(%s)', [TMixture(RootNode).QuantumNumberToStr(Component)])
  else
    Result := Format('(%s,%s)', [TMixture(RootNode).QuantumNumberToStr(Rank), TMixture(RootNode).QuantumNumberToStr(Component)]);
end;

constructor TMagneticTransitionMoment.Create(AOwner: TComponent);
begin
  FMagnetic := True;
  inherited;
end;

{ TSpinTransitionMoment }

function TSpinTransitionMoment.PairName: string;
begin
  Result := 'gS' + inherited PairName;
end;

constructor TSpinTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := 'Strength=' + FloatToStr(ge) + ',' + SavedProperties;
  inherited;
end;

{ TOrbitalTransitionMoment }

function TOrbitalTransitionMoment.PairName: string;
begin
  Result := 'gL' + inherited PairName;
end;

{ TRotationalTransitionMoment }

function TRotationalTransitionMoment.PairName: string;
begin
  Result := 'gr' + inherited PairName;
end;

initialization
  RegisterClasses([TSpecies, TTransitionMoments, TSphericalTransitionMoment, TCartesianTransitionMoment,
                   TSpinTransitionMoment, TOrbitalTransitionMoment, TRotationalTransitionMoment]);
finalization
  FreeAndNil(MoleculeTypes);
end.
