unit QuantumNumber;

interface

type
  TQuantumNumber = Integer;

  TNuclearInfo = record
    Name: String;
    Spin: TQuantumNumber
  end;

function IsTriangle(J1, J2, J3: TQuantumNumber): Boolean;
procedure MatchOdd(A, B: TQuantumNumber);
function SmallestProjection(A: TQuantumNumber): TQuantumNumber;
procedure BadMinus(A: TQuantumNumber);
function Minus(A: TQuantumNumber): Integer; {$IFDEF INLINE} inline; {$ENDIF} { (-1)**A }

function rl(J, K, Kp: Integer): Double;
function rlx(n: Integer; J, K, Kp: TQuantumNumber):Double;
function rly(n: Integer; J, K, Kp: TQuantumNumber):Double;
function ReducedMatrixElement(Rank, J: TQuantumNumber): Double;

function DeltaJName(DeltaJ: TQuantumNumber): Char;

implementation

uses SysUtils, Math, nj;

function IsTriangle(J1, J2, J3: TQuantumNumber): Boolean;
{ Do J1, J2, J3 satisfy the triangle rule for angular momentum addition? }
begin
  if Odd(J1+J2+J3) then
    raise Exception.CreateFmt('Triangle: %d %d %d', [J1, J2, J3]);
  Result := (Abs(J1-J2) <= J3) and (J1+J2 >= J3);
end;

procedure MatchOdd(A, B: TQuantumNumber);
begin
  if Odd((A xor B) and 1) then
    raise Exception.CreateFmt('Mix of integral and half integral quantum numbers: %d %d', [A, B]);
end;

function SmallestProjection(A: TQuantumNumber): TQuantumNumber;
begin
  Result := A and 1;
end;

procedure BadMinus(A: TQuantumNumber);
begin
  { Do not inline this function }
  raise Exception.CreateFmt('Minus(%d)', [A]);
end;

function Minus(A: TQuantumNumber): Integer; {$IFDEF INLINE} inline; {$ENDIF} { (-1)**A }
begin
  if (A and 1) <> 0 then
    BadMinus(A);
  if (A and 2) <> 0 then
    Result := -1
  else
    Result := 1;
end;

function rl(J, K, Kp: Integer): Double;
{ Work out matrix element of arbitrary power of raising/lowering operator }
var
  JSq: Double;
begin
  Result := 1;
  JSq := J*(J+2) / 4;
  while K < Kp do begin
    Result := Result * (JSq - K*(K+2)/4);
    Inc(K, 2);
  end;
  while K > Kp do begin
    Result := Result * (JSq - K*(K-2)/4);
    Dec(K, 2);
  end;
  Result := Sqrt(Result);
end;

function rlxy(n: Integer; J, K, Kp: TQuantumNumber; Sign: Integer):Double;
{ Helper function for matrix elements of Jx^n and Jy^n }
{ TODO: This is likely to be inefficient for large values of n }
begin
  Result := 0;
  if 2*n < Abs(K-Kp) then
    Exit;
  if 2*n = Abs(K-Kp) then
    { Easy case - equivalent to all raising/all lowering operator }
    if n = 0 then
      Result := 1
    else begin
      Result := rl(J,K,Kp) / IntPower(2,n);
      if (K < Kp) and Odd(n) then
        Result := Sign * Result;
    end
  else begin
    { Need to consider combinations;
      e.g. Jx^2 can give DeltaK = 0, +/-2 Matrix elements }
    if Abs(Kp+2) <= J then
      Result := Result + rlxy(n-1,J,K,Kp+2,Sign)*rl(J,Kp+2,Kp)/2;
    if Abs(Kp-2) <=J then
      Result := Result + Sign * rlxy(n-1,J,K,Kp-2,Sign)*rl(J,Kp-2,Kp)/2;
  end;
end;

function rlx(n: Integer; J, K, Kp: TQuantumNumber):Double;
{ Matrix elements of Jx^n }
begin
  Result := rlxy(n,J,K,Kp,1);
end;

function rly(n: Integer; J, K, Kp: TQuantumNumber):Double;
{ Matrix elements of Jy^n; for odd powers the result is multiplied by i }
begin
  Result := rlxy(n,J,K,Kp,-1);
  { Multiply by i^n }
  if Odd(n) then
    Result := Minus(n+1)*Result { Multiply by a bonus i to give a real result }
  else
    Result := Minus(n)*Result;
end;

function ReducedMatrixElement(Rank, J: TQuantumNumber): Double;
{ Return the reduced matrix element <J||T^k(J)||J> formed by
  coupling J with J k times. }
begin
  Result := Sqrt(J*(J+2)*(J+1)/4);
  if Rank > 2 then
    Result := Result * Minus(Rank+2*J) * sixj(2, Rank-2, Rank, J, J, J) * ReducedMatrixElement(Rank-2, J);
end;

function DeltaJName(DeltaJ: TQuantumNumber): Char;
var
  Difference: Integer;
begin
  if Abs(DeltaJ) > (Ord('Z')-Ord('Q'))*2 then
    Result := ' '
  else
    Result := Char(Ord('Q') + DeltaJ div 2);
end;

end.
