unit Units;

interface

type
  TResonanceUnits = (cm1=0, MHz, Kelvin, eV, nmVac, nmAir);
  TEnergyUnits = cm1..eV;

function NameOfUnit(Units: TResonanceUnits): string; overload;
function TypeOfUnit(Units: TResonanceUnits): string;
function StrToUnits(s: string): TResonanceUnits;
function ConversionFactor(UFrom, UTo: TEnergyUnits):Extended;
function Convert(Value:Double; UFrom, UTo: TResonanceUnits):Extended;
function ConvertWidth(Value, Centre:Double; UFrom, UTo: TResonanceUnits):Extended;

type
  TIntensityUnits = (Arbitrary=0,
                     Normalized,
                     Squared,
                     nm2MHzperMolecule {nm^2 MHz / molecule},
                     cm2WavenumberperMolecule { cm^2 cm-1 / molecule },
                     HzperMolecule {s-1, emission} );

function NameOfUnit(Units: TIntensityUnits): string; overload;
function IntensityFactor(FrequencyUnits: TResonanceUnits; IntensityUnits: TIntensityUnits): Double;

const
  BoltzmannConstants: array [TEnergyUnits] of Double =
    ( 0.6950356 { for cm-1 }, 20836.644 { MHz }, 1 { Kelvin }, 8.617343e-5 { eV } );
  ge: Double = 2.0023193043718;
  BohrMagneton: Double = 927.400949e-26; { e hcross / 2 / me }
  SpeedOfLight: Double = 299792458.0;

implementation

uses AirVac, SysUtils, Math;

function NameOfUnit(Units: TResonanceUnits): string;
begin
  case Units of
    cm1: Result := 'cm-1';
    MHz: Result := 'MHz';
    Kelvin: Result := 'Kelvin';
    eV:  Result := 'eV';
    nmVac:  Result := 'nm';
    nmAir:  Result := 'nm';
  end;
end;

function StrToUnits(s: string): TResonanceUnits;
begin
  Result := cm1;
  if CompareText(s, 'cm-1') = 0 then
    Exit;
  repeat
    if CompareText(s, NameOfUnit(Result)) = 0 then
      Exit;
    if Result = High(Result) then
      raise Exception.Create(s + ' is not a valid unit name');
    Inc(Result);
  until False;
end;

function TypeOfUnit(Units: TResonanceUnits): string;
begin
  case Units of
    cm1: Result := 'Wavenumber';
    MHz, eV: Result := 'Energy';
    Kelvin: Result := 'Temperature';
    nmAir:  Result := 'Air Wavelength';
    nmVac:  Result := 'Vacuum Wavelength';
  end;
end;

const
  SpeedOfLight4: Double = 29979.2458;
  eVTocm1: Double = 8065.54445;
  KelvinToMHz: Double = 20836.644;
  Planck: Double = 6.6260693e-34;


function ConversionFactor(UFrom, UTo: TEnergyUnits):Extended;
begin
  { Can't use array of constants, because expressions like 1/SpeedOfLight don't count as constant expressions }
  if UTo < UFrom then
    Result := 1/ConversionFactor(UTo, UFrom)
  else if UFrom = UTo then
    Result := 1
  else
    case UFrom of
      cm1:    case UTo of
                MHz:    Result := SpeedOfLight4;
                Kelvin: Result := SpeedOfLight4/KelvinToMHz;
                ev:     Result := 1/eVTocm1;
              end;
      MHz:    case UTo of
                Kelvin: Result := 1/KelvinToMHz;
                ev:     Result := 1/eVTocm1/SpeedOfLight4;
              end;
      Kelvin: case UTo of
                ev:     Result := KelvinToMHz/SpeedOfLight4/eVTocm1;
              end;
    end;
end;

function Convert(Value:Double; UFrom, UTo: TResonanceUnits):Extended;
begin
  Result := Value;
  if UFrom = UTo then
    Exit;
  case UFrom of
    nmAir: begin UFrom := cm1; Result := vac(Result); end;
    nmVac: begin UFrom := cm1; Result := 1e7/Result; end;
  end;
  if UTo <= High(TEnergyUnits) then
    Result := Result*ConversionFactor(UFrom, UTo)
  else begin
    Result := Result*ConversionFactor(UFrom, cm1);
    case UTo of
      nmAir: Result := air(Result);
      nmVac: Result := 1e7/Result;
    end
  end;
end;

function ConvertWidth(Value, Centre:Double; UFrom, UTo: TResonanceUnits):Extended;
begin
  if UFrom = UTo then
    Result := Value
  else if (UFrom > High(TEnergyUnits)) or (UTo > High(TEnergyUnits)) then
    { Taking inverse or air <-> nm conversion - scale according to Value/Centre }
    Result := Value/Centre * Convert(Centre, UFrom, UTo)
  else
    Result := Convert(Value, UFrom, UTo)
end;

function NameOfUnit(Units: TIntensityUnits): string;
begin
  case Units of
    Arbitrary: Result := '';
    Squared:   Result := 'Intensity^2';
    Normalized: Result := 'Normalized Intensity';
    nm2MHzperMolecule: Result := 'nm^2 MHz / molecule';
    cm2WavenumberperMolecule: Result := 'cm^2 cm-1 / molecule';
    HzperMolecule: Result := 's-1 / molecule';
  end;
end;

const
  FrequencyUnit: array [TIntensityUnits] of TResonanceUnits =
    (cm1 { Actually none },
     cm1 { Actually none },
     cm1 { Actually none },
     MHz, { nm2MHzperMolecule }
     cm1, { cm2WavenumberperMolecule }
     cm1  { HzperMolecule });

var { Actually const, but compiler can't cope with these as constants }
  IntensityFactors: array [TIntensityUnits] of Double;
  { Conversion factors excluding frequency }

function IntensityFactor(FrequencyUnits:TResonanceUnits; IntensityUnits: TIntensityUnits): Double;
var
  co: Double;
begin
  Result := IntensityFactors[IntensityUnits];
  if (IntensityUnits > Normalized)
     and (FrequencyUnits <> FrequencyUnit[IntensityUnits]) then begin
    co := ConversionFactor(FrequencyUnits, FrequencyUnit[IntensityUnits]);
    if IntensityUnits = HzperMolecule then
      Result := Result * IntPower(co,3)
    else
      Result := Result * co;
  end;
end;

initialization
  IntensityFactors[Arbitrary] := 1;
  IntensityFactors[Squared] := 1;
  IntensityFactors[Normalized] := 1;
  IntensityFactors[nm2MHzperMolecule] := 8*Pi*Pi*Pi/(3e31*SpeedOfLight*Planck);
  IntensityFactors[cm2WavenumberperMolecule] := IntensityFactors[nm2MHzperMolecule] / 1e14;
  IntensityFactors[HzperMolecule] := 64*IntPower(pi,4)*1e-43/3/Planck;
end.
