unit MNode;
{ Generic node type - handles loading/saving of parameters }

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Classes, Graphics,
     {$IFNDEF CONSOLE} ComCtrls, {$ENDIF}
     DOM, SysUtils, Units;

type

  TDoubleArray = array of Double;   { Both needed to avoid some type problems }
  TIntegerArray = array of Integer;

  TMNodeClass = class of TMNode;

  { TMNode }

  TMNode = class(TComponent)
  private
    FnParameters: Integer;
    FnSettings: Integer;
    FParameters: array of Double;
    FDefaultParameters: array of Double;
    FIncrements: array of Double;
    FSettings: array of Integer;
    FName: TComponentName;
    FParameterNames: TStringList;
    FSettingNames: TStringList;
    FModified: Boolean;
    FFloating: array of Boolean;
    FComment: string;
    function GetItem(AIndex: Integer): TMNode; {$IFDEF INLINE} inline; {$ENDIF}
    function GetParameterNames(AIndex: Integer): string;
    function GetFloating(AIndex: Integer): Boolean;
    procedure SetFloating(AIndex: Integer; const Value: Boolean);
    function GetIncrement(AIndex: Integer): Double;
    procedure SetIncrement(AIndex: Integer; const Value: Double);
    procedure SetComment(const Value: string);
  protected
    function GetChildTypes(AIndex: Integer): TMNodeClass; virtual; abstract;
    function GetSettingNames(AIndex: Integer): string; virtual;
    function GetName: TComponentName; virtual;
    procedure SetName(const Value: TComponentName); override;
    procedure SetModified(const Value: Boolean); virtual;
  public
    SavedProperties: string;
    DimensionlessParameters: string;
    {$IFNDEF CONSOLE}
    TreeNode: TTreeNode; { These two are helpers for TreeView display }
    ItemIndex: Integer;
    {$ENDIF}
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function OwnerType: TComponentClass; virtual;
    class function DisplayClassName: string; virtual;
    function Count:Integer;

    procedure TransferProperties(DOMNode: TDOMNode; Writing: Boolean);
    procedure ReadChildProperties(DOMNode: TDOMNode);
    function ReadMNode(DOMNode: TDOMNode): TMnode;
    procedure ReadProperties(Node: TDOMNode); virtual;
    function WriteProperties(Parent: TDOMNode): TDomNode; virtual;
    procedure SaveToStream(Stream: TStream; WithLinkedMNodes: Boolean = False);
    function SimpleSaveToStream(Stream: TStream; WithLinkedMNodes: Boolean = False): Boolean; virtual;
    procedure ReadFromStream(Stream: TStream);

    procedure ClearCache; virtual; { Call after significant property value has changed }
    procedure CheckProperties;
    function FixupProperties: Exception; virtual; { Call after change to set up links other than parent/child }
    procedure FixAll(Recurse: Boolean);
    procedure HandleAssignment(NameVal: string);

    function GetParameter(AIndex: Integer): Double; {$IFDEF INLINE} inline; {$ENDIF}
    procedure ModifyParameter(AIndex: Integer; const Value: Double; ChangeModify: Boolean);
    procedure SetParameter(AIndex: Integer; const Value: Double); virtual;
    function HelpOnParameter(AIndex: Integer): string; virtual;
    function IndexOfParameter(const AName: string):Integer;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); virtual;

    function GetSetting(AIndex: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
    procedure ModifySetting(AIndex: Integer; const AValue: Integer; ChangeModify: Boolean);
    procedure SetSetting(AIndex: Integer; const AValue: Integer); virtual;
    function GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar; virtual;
    function HelpOnSetting(AIndex: Integer): string; virtual;
    function SettingToStr(AIndex: Integer): string; virtual;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); virtual;
    function IsSettingEnum(AIndex: Integer):Boolean;
    procedure GetEnumNames(AIndex: Integer; List: TStrings); virtual;
    function GetFlag(AIndex: Integer): Boolean;
    procedure SetFlag(AIndex: Integer; const AValue: Boolean);
    function GetColour(AIndex: Integer): TColor;
    procedure SetColour(AIndex: Integer; const AValue: TColor);
    function IsSettingColour(AIndex: Integer): Boolean; virtual;

    function RootNode: TMNode;
    function LinkedNodes(Delete: Boolean = False; Parent: TDOMNode = nil):Integer;
    function ChildWithClass(MNodeClass: TMNodeClass):TMNode;
    procedure CycleUnits; virtual;

    function UniqueName: string; virtual;
    function UniquePrefix: string; virtual;
    function ComponentWithName(const AName: string; AClass: TClass): TMNode;

    property nParameters: Integer read FnParameters;
    property Parameters[AIndex: Integer]: Double read GetParameter write SetParameter;
    property Increment[AIndex: Integer]: Double read GetIncrement write SetIncrement;
    property nSettings: Integer read FnSettings;
    property Settings[AIndex: Integer]: Integer read GetSetting write SetSetting;
    property Items[AIndex: Integer]: TMNode read GetItem;
    property Floating[AIndex: Integer]: Boolean read GetFloating write SetFloating;
    property ParameterNames[AIndex: Integer]: string read GetParameterNames;
    property SettingNames[AIndex: Integer]: string read GetSettingNames;
    property Modified: Boolean read FModified write SetModified;
    property ChildTypes[AIndex: Integer]: TMNodeClass read GetChildTypes;
    function nChildTypes: Integer; virtual;
  published
    property Name: TComponentName read GetName write SetName;
    property Comment: string read FComment write SetComment;
    {$IFNDEF CONSOLE}
    procedure CopyToClipboard(WithLinkedMNodes: Boolean = False); virtual;
    {$ENDIF}
  end;

  TMNodeChange = (mnCreate, mnDestroy, mnParameterChange, mnSettingChange, mnNameChange, mnNewBaseNode, mnPropNameChange, mnGraphicChanged);
  OnMNodeChange = procedure(Sender: TMNode; Change: TMNodeChange; Index: Integer) of object;

  { TMNodePair }

  TMNodePair = class(TMNode)
  protected
    FBra, FKet: TMNode;
    procedure DoMNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
    function GetName: TComponentName; override;
    procedure SetName(const Value: TComponentName); override;
  public
    FBraName, FKetName: string; { Only used during load/save, hence use of F...Name }
    class function BraKetClass: TMNodeClass; virtual; abstract;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function BraOwner: TMNode; virtual;
    function KetOwner: TMNode; virtual;
    function FixupProperties: Exception; override;
    procedure ReadProperties(Node: TDOMNode); override;
    function WriteProperties(Parent: TDOMNode): TDomNode; override;
    procedure SetBraKetName(const ABraName, AKetName: string);
    function SetBraKetNameE(const ABraName, AKetName: string; ChangingMNode: TMNode = nil): Exception;
    property Bra: TMNode read FBra;
    property Ket: TMNode read FKet;
    function PairName: string; virtual; abstract; { For use in display }
  end;

  EReportException = class(Exception);

  ECantFindState = class(Exception)
  private
    FMissingState: string;
    FMNodePair: TMNodePair;
  public
    property MNodePair: TMNodePair read FMNodePair;
    property MissingState: string read FMissingState;
    constructor Create(AMNodePair: TMNodePair; const AMissingState: string);
  end;

procedure AddMNodeNotification(MNodeChange: OnMnodeChange);
procedure RemoveMNodeNotification(MNodeChange: OnMnodeChange);
procedure PerformMNodeNotification(Sender: TMNode; Change: TMNodeChange; Index: Integer);

function ReadValue(DOMNode: TDOMNode; const PropName, Dflt: string; const AltPropName: string = ''): string;
procedure WriteValue(DOMNode: TDOMNode; PropName, Value, Dflt: string);

function ReadMNodeClass(DOMNode: TDOMNode; ParentClass: TClass): TMNodeClass;

implementation

uses
     {$IFNDEF CONSOLE}
       {$IFDEF MSWINDOWS} Windows, {$ENDIF}
       MNodeClip, Clipbrd, Huge, HugeClip,
     {$ENDIF}
     MNodeColour, TypInfo, Math, MathSup, XMLRead, XMLWrite, PropHelp,
     StrUtils;
 
{ TMNode }

function TMNode.GetParameterNames(AIndex: Integer): string;
begin
  Result := FParameterNames[AIndex];
end;

function TMNode.GetSettingNames(AIndex: Integer): string;
begin
  Result := FSettingNames[AIndex];
end;

function TMNode.GetItem(AIndex: Integer): TMNode; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  {$IFOPT R+}
  Result := Components[AIndex] as TMNode;
  {$ELSE}
  Result := TMNode(Components[AIndex]);
  {$ENDIF}
end;

constructor TMNode.Create(AOwner: TComponent);
var
  XMLDocument: TXMLDocument;
  PropList: PPropList;
  i, nProps, p, p1: Integer;
  Prop: string;
  Propl, spl: string;
begin
  inherited Create(AOwner);
  FnParameters := 0;
  FnSettings := 0;
  { After this, SavedProperties contains a comma delimited list list of all properies to be saved }
  SavedProperties := ',' + SavedProperties + ',';
  {$IFDEF FPC}
  nProps := GetTypeData(PTypeInfo(Self.ClassInfo))^.Propcount;
  GetMem(PropList,nProps*SizeOf(Pointer));
  GetPropInfos(PTypeInfo(Self.ClassInfo), PropList);
  {$ELSE}
  nProps := GetPropList(Self, PropList);
  {$ENDIF}
  for i := nProps-1 downto 0 do begin
    Prop := PropList^[i]^.Name;
    Propl := LowerCase(Prop);
    spl := LowerCase(SavedProperties);
    if (Pos(','+Propl+',',spl) = 0) and (Pos(','+Propl+'=',spl) = 0) then
      SavedProperties := ',' + Prop + SavedProperties;
    if PropList^[i]^.Index > 0 then begin
      case PropList^[i]^.PropType^.Kind of
        tkInteger, tkChar, tkEnumeration {$IFDEF FPC}, tkBool {$ENDIF}:  FnSettings := Max(FnSettings, PropList^[i]^.Index);
        tkFloat: FnParameters := Max(FnParameters, PropList^[i]^.Index);
      else
        raise Exception.CreateFmt('Unknown Type: %d', [Ord(PropList^[i]^.PropType^.Kind)]);
      end;
    end;
  end;
  { Note that index 0 is never used }
  SetLength(FParameters, FnParameters+1);
  SetLength(FDefaultParameters, FnParameters+1);
  SetLength(FFloating, FnParameters+1);
  SetLength(FIncrements, FnParameters+1);
  FParameterNames := TStringList.Create;
  FParameterNames.Capacity := nParameters+1;
  for i := 0 to FnParameters do begin
    FParameters[i] := 0; { Default is 0, unless otherwise specified }
    FParameterNames.Add('');
    FFloating[i] := False;
    FIncrements[i] := 0;
  end;

  SetLength(FSettings, FnSettings+1);
  FSettingNames := TStringList.Create;
  FSettingNames.Capacity := FnSettings+1;
  for i := 0 to FnSettings do begin
    FSettings[i] := 0; { Default is 0, unless otherwise specified }
    FSettingNames.Add('')
  end;

  { ReScan properties, assigning names and recording default values }
  for i := nProps-1 downto 0 do
    if PropList^[i]^.Index >= 0 then
      case PropList^[i]^.PropType^.Kind of
        tkInteger, tkChar, tkEnumeration{$IFDEF FPC}, tkBool {$ENDIF}:
          if FSettingNames[PropList^[i]^.Index] <> '' then
            raise Exception.CreateFmt('Duplicate property index (%d) for %s: %s and %s',
                                      [PropList^[i]^.Index, ClassName, FSettingNames[PropList^[i]^.Index], PropList^[i]^.Name])
          else
            FSettingNames[PropList^[i]^.Index] := PropList^[i]^.Name;
        tkFloat:
          if FParameterNames[PropList^[i]^.Index] <> '' then
            raise Exception.CreateFmt('Duplicate property index (%d) for %s: %s and %s',
                                      [PropList^[i]^.Index, ClassName, FParameterNames[PropList^[i]^.Index], PropList^[i]^.Name])
          else begin
            FParameterNames[PropList^[i]^.Index] := PropList^[i]^.Name;
            Prop := ','+LowerCase(PropList^[i]^.Name)+'=';
            p := Pos(Prop,spl);
            if p > 0 then begin
              Prop := Copy(spl,p+Length(PropList^[i]^.Name)+2,Length(spl));
              p := Pos(',',Prop);
              FDefaultParameters[PropList^[i]^.Index] := StrToFloat(Copy(Prop,1,p-1));
            end else
              FDefaultParameters[PropList^[i]^.Index] := 0;
          end;
      end;

  FreeMem(PropList);
  XMLDocument := TXMLDocument.Create;
  TransferProperties(XMLDocument, False); { Set to defaults by reading from a blank file }
  XMLDocument.Free;
  Modified := False;
  PerformMNodeNotification(Self, mnCreate, -1);
end;

class function TMNode.OwnerType: TComponentClass;
begin
  raise Exception.Create('No OwnerType for ' + Self.ClassName);
end;

function TMNode.Count: Integer;
begin
  Result := ComponentCount;
end;

function EncodePropName(const Name: string):string;
{$IFDEF OFFLCL_}
  function AnsiReplaceStr(s: string; f, t: string):string;
  var
    p: Integer;
  begin
    repeat
      p := Pos(f, s);
      if p > 0 then begin
        Delete(s, p, Length(f));
        Insert(s, t, p);
      end;
    until p <= 0;
    Result := s;
  end;
{$ENDIF}
begin
  Result := AnsiReplaceStr(Name, '''', 'prime');
  Result := AnsiReplaceStr(Result, '"', 'dprime');
  Result := AnsiReplaceStr(Result, '(', '_');
  Result := AnsiReplaceStr(Result, ')', '_');
  Result := AnsiReplaceStr(Result, ',', '_');
end;

procedure WriteValue(DOMNode: TDOMNode; PropName, Value, Dflt: string);
var
  Attr: TDOMAttr;
begin
  if PropName[1] = '(' then
    Value := Dflt; { Do not write bracketed properties }
  PropName := EncodePropName(PropName);
  Attr := (DOMNode as TDOMElement).GetAttributeNode(PropName);
  if Assigned(Attr) then begin
    if Value = Dflt then
      (DOMNode as TDOMElement).RemoveAttributeNode(Attr)
    else if Attr.Value <> Value then
      Attr.Value := Value;
  end else if Value <> Dflt then
    { Create new node }
    (DOMNode as TDOMElement)[PropName] := Value;
end;

function ReadValue(DOMNode: TDOMNode; const PropName, Dflt: string; const AltPropName: string = ''): string;
var
  Attr: TDOMAttr;
begin
  Result := Dflt;
  if DOMNode is TDOMElement then begin
    Attr := TDOMElement(DOMNode).GetAttributeNode(EncodePropName(PropName));
    if Assigned(Attr) then
      Result := Attr.NodeValue
    else if (AltPropName <> '') and (AltPropName[1] <> '(') then begin { ignore properties with brackets in name }
      Attr := TDOMElement(DOMNode).GetAttributeNode(EncodePropName(AltPropName));
      if Assigned(Attr) then
        Result := Attr.NodeValue
    end;
  end;
end;

function TMNode.IsSettingEnum(AIndex: Integer): Boolean;
var
  PropInfo: PPropInfo;
  LowestValue: Integer;
begin
  if Assigned(GetSettingsNames(AIndex,LowestValue)) then
    Result := True
  else begin
    PropInfo := GetPropInfo(Self, FSettingNames[AIndex]);
    Result := PropInfo^.PropType^.Kind = tkEnumeration;
  end;
end;

procedure TMNode.GetEnumNames(AIndex: Integer; List: TStrings);
var
  PropInfo: PPropInfo;
  TypeData: PTypeData;
  i: Longint;
  SettingsNames: PPChar;
begin
  List.Clear;
  SettingsNames := GetSettingsNames(AIndex,i);
  if Assigned(SettingsNames) then
    while SettingsNames^ <> nil do begin
      if SettingsNames^^ <> #0 then
        List.Add(SettingsNames^);
      Inc(SettingsNames);
    end
  else begin
    PropInfo := GetPropInfo(Self, FSettingNames[AIndex]);
    if not Assigned(PropInfo) or (PropInfo^.PropType^.Kind <> tkEnumeration) then
      raise Exception.Create('Not Enum');
    TypeData := GetTypeData(PropInfo^.PropType{$IFNDEF FPC}^{$ENDIF});
    for i := TypeData^.MinValue to TypeData^.MaxValue do
      List.Add(GetEnumName(PropInfo^.PropType{$IFNDEF FPC}^{$ENDIF}, i));
  end;
end;

function TMNode.SettingToStr(AIndex: Integer): string;
var
  SettingsNames: PPChar;
  LowestValue: Integer;
begin
  if IsSettingColour(AIndex) then
    Result := ColourToStr(GetColour(AIndex))
  else begin
    LowestValue := 0;
    SettingsNames := GetSettingsNames(AIndex,LowestValue);
    if Assigned(SettingsNames) then begin
      Inc(SettingsNames, GetSetting(AIndex)-LowestValue);
      Result := SettingsNames^;
    end else
      Result := PropertyToStr(Self, FSettingNames[AIndex]); { Use real name here }
  end;
end;

procedure TMNode.SetSettingFromStr(AIndex: Integer; const s: string);
var
  SettingsNames: PPChar;
  i, LowestValue: Integer;
begin
  if IsSettingColour(AIndex) then
    SetColour(AIndex, StrToColour(s))
  else begin
    LowestValue := 0;
    SettingsNames := GetSettingsNames(AIndex,LowestValue);
    if Assigned(SettingsNames) then begin
      i := 0;
      while SettingsNames^ <> nil do
        if CompareText(SettingsNames^, s) = 0 then begin
          SetSetting(AIndex, i+LowestValue);
          Exit;
        end else begin
          Inc(i);
          Inc(SettingsNames);
        end;
    end;
    SetPropertyFromStr(Self, FSettingNames[AIndex], s); { Use real name here }
  end;
end;

function ReadMNodeClass(DOMNode: TDOMNode; ParentClass: TClass): TMNodeClass;
var
  NewSubClass: TPersistentClass;
begin
  NewSubClass := GetClass('T' + DOMNode.NodeName);
  if (NewSubClass = nil) then
    if Copy(DOMNode.NodeName,1,14) = 'OpenAsymmetric' then
      NewSubClass := GetClass('T' + Copy(DOMNode.NodeName,5,255)) { Compatibility - Delete Open }
    else if Copy(DOMNode.NodeName,1,5) = 'HundA' then
      NewSubClass := GetClass('TLinear' + Copy(DOMNode.NodeName,6,255)); { Compatibility - TLinear was THundA }
  if NewSubClass = nil then
    raise Exception.Create('Can''t find ' + DOMNode.NodeName);
  if not NewSubClass.InheritsFrom(TMNode) then
    raise Exception.Create('Class is not a subclass of TMNode: ' + NewSubClass.ClassName);
  if Assigned(ParentClass) and not ParentClass.InheritsFrom(TMNodeClass(NewSubClass).OwnerType())  then
    raise Exception.Create(DOMNode.NodeName + ' OwnerClass does not match :' + TMNodeClass(ParentClass).OwnerType().ClassName +  ': ' + ParentClass.ClassName);
  Result := TMNodeClass(NewSubClass);
end;

function TMNode.ReadMNode(DOMNode: TDOMNode): TMnode;
var
  NewSubClass: TMNodeClass;
begin
  NewSubClass := ReadMNodeClass(DOMNode, ClassType);
  Result := NewSubClass.Create(Self) as TMNode;
  Result.Name := DOMNode.NodeName;
  Result.ReadProperties(DOMNode);
end;

procedure TMNode.ReadChildProperties(DOMNode: TDOMNode);
var
  Child: TDOMNode;
begin
  Child := DOMNode.FirstChild;
  while Assigned(Child) do begin
    if (Child.NodeName <> 'Parameter') and not (Child is TDOMCDATASection) then begin
      ReadMNode(Child);
    end;
    Child := Child.NextSibling;
  end;
end;


procedure TMNode.TransferProperties(DOMNode: TDOMNode; Writing: Boolean);
var
  i, Index: Integer;
  Props, Prop, Deflt, AltPropName: string;
  PropInfo: PPropInfo;
  TypeData: PTypeData;
  procedure SetValue(Value: string);
  begin
    WriteValue(DOMNode, AltPropName, Value, Deflt);
  end;
  procedure WriteFloatableValue(Index: Integer);
  var
    ParNode: TDOMNode;
  begin
    if not SameValue(GetParameter(Index), StrToFloat(Deflt)) or Floating[Index] or (Increment[Index] <> 0) then begin
      ParNode := DomNode.OwnerDocument.CreateElement('Parameter');
      DomNode.AppendChild(ParNode);
      WriteValue(ParNode, 'Name', Prop, '');
      WriteValue(ParNode, 'Value', FloatToStr(GetParameter(Index)), Deflt);
      if Floating[Index] then
        WriteValue(ParNode, 'Float', 'true', '');
      WriteValue(ParNode, 'Increment', FloatToStr(Increment[Index]), '0');
    end;
  end;
  procedure ReadFloatableValue(Index: Integer);
  var
    Attr: TDOMAttr;
    ParNode: TDOMNode;
    s: string;
  begin
    ParNode := DomNode.FirstChild;
    while Assigned(ParNode) do begin
      if CompareText(ParNode.NodeName, 'Parameter') = 0 then begin
        Attr := (ParNode as TDOMElement).GetAttributeNode('Name');
        if Assigned(Attr) and (CompareText(Attr.NodeValue,Prop) = 0) then
          Break; { Found it }
      end;
      ParNode := ParNode.NextSibling;
    end;
    Floating[Index] := False;
    Increment[Index] := 0;
    s := Deflt;
    if Assigned(ParNode) then begin
      s := ReadValue(ParNode, 'Float', 'false');
      if StrToBoolean(s) then
        Floating[Index] := True;
      s := ReadValue(ParNode, 'Increment', '');
      if s <> ''then
        Increment[Index] := StrToFloat(s);
      s := ReadValue(ParNode, 'Value', Deflt);
    end;
    Parameters[Index] := StrToFloat(s);
  end;
  function GetValue: string;
  begin
    Result := ReadValue(DOMNode, Prop, Deflt, AltPropName);
  end;
begin
  { Note that order in SavedProperties is important for overriden property names;
    property that controls a name must be before the property itself }
  Props := SavedProperties;
  while Props <> '' do begin
    i := Pos(',', Props);
    if i > 0 then begin
      Prop := Copy(Props, 1, i-1);
      System.Delete(Props, 1, i);
    end else begin
      Prop := Props;
      Props := '';
    end;
    if Prop = '' then Continue;
    Deflt := '';
    i := Pos('=', Prop);
    if i > 0 then begin
      Deflt := Copy(Prop, i+1, Length(Prop));
      System.Delete(Prop, i, Length(Prop));
    end;
    PropInfo := GetPropInfo(Self, Prop);
    if (PropInfo = nil) {$IFNDEF FPC} or (PropInfo^.Index = 0) {$ENDIF} then
      raise Exception.Create('No such property: '+ClassName+':'+Prop);
    { Check for overridden name }
    AltPropName := Prop;
    Index := PropInfo^.Index;
    if (Index > 0) and (Index < FSettingNames.Count)
       and (PropInfo^.Name = FSettingNames[Index]) then
      AltPropName := SettingNames[Index];
    case PropInfo^.PropType^.Kind of
      tkInteger:
               begin
                 if Deflt = '' then Deflt := '0';
                 if (PropInfo^.Index >= 0) and IsSettingEnum(PropInfo^.Index) then
                   if Writing then
                     SetValue(SettingToStr(PropInfo^.Index))
                   else
                     SetSettingFromStr(PropInfo^.Index, GetValue())
                 else if Writing then
                   SetValue(IntToStr(GetOrdProp(Self, PropInfo)))
                 else
                   SetOrdProp(Self, PropInfo, StrToInt(GetValue()));
               end;
      tkEnumeration:
               begin
                 if Deflt = '' then begin
                   TypeData := GetTypeData(PropInfo^.PropType{$IFNDEF FPC}^{$ENDIF});
                   Deflt := GetEnumName(PropInfo^.PropType{$IFNDEF FPC}^{$ENDIF}, TypeData^.MinValue);
                 end;
                 if PropInfo^.Index >= 0 then begin
                   if Writing then
                     SetValue(SettingToStr(PropInfo^.Index))
                   else
                     SetSettingFromStr(PropInfo^.Index, GetValue());
                 end else
                   if Writing then
                     SetValue(GetEnumProp(Self, PropInfo))
                   else
                     SetEnumProp(Self, PropInfo, GetValue());
               end;
{$IFDEF FPC}
      tkBool:  begin
                 if Deflt = '' then Deflt := 'False';
                 if Writing then
                   if GetOrdProp(Self, PropInfo) <> 0 then
                     SetValue('True')
                   else
                     SetValue('False')
                 else begin
                   if StrToBoolean(LowerCase(GetValue())) then
                     SetOrdProp(Self, PropInfo, 1)
                   else
                     SetOrdProp(Self, PropInfo, 0);
                  end;
               end;
{$ENDIF}
      tkFloat: begin
                 if Deflt = '' then Deflt := '0';
                 if PropInfo^.Index >= 0 then
                   if Writing then
                     WriteFloatableValue(PropInfo^.Index)
                   else
                     ReadFloatableValue(PropInfo^.Index)
                 else if Writing then
                   SetValue(FloatToStr(GetFloatProp(Self, PropInfo)))
                 else
                   SetFloatProp(Self, PropInfo, StrToFloat(GetValue()));
               end;
      tkString, tkLString {$IFDEF FPC}, tkAString {$ENDIF}:
                   if Writing then
                     SetValue(GetStrProp(Self, PropInfo))
                   else
                     SetStrProp(Self, PropInfo, GetValue());
      tkClass: { Handled as a child MNode or by a TransferProperties descendant }
    else
      raise Exception.Create('Bad property type (' + IntToStr(Ord(PropInfo^.PropType^.Kind)) + ') :' +ClassName+':'+Prop);
    end;
  end;
  if Writing then
    for i := 0 to Count-1 do
      Items[i].WriteProperties(DOMNode)
  else
    ReadChildProperties(DOMNode);
end;

{ Note difference between the two procedures - read takes the current node, write takes the parent node and returns the current node }
procedure TMNode.ReadProperties(Node: TDOMNode);
begin
  TransferProperties(Node, False);
end;

function TMNode.WriteProperties(Parent: TDOMNode): TDomNode;
begin
  Result := Parent.OwnerDocument.CreateElement(Copy(ClassName,2,Length(ClassName)));
  Parent.AppendChild(Result);
  TransferProperties(Result, True);
end;

procedure TMNode.ClearCache;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    Items[i].ClearCache;
end;

function TMNode.FixupProperties: Exception;
var
  i: Integer;
begin
  Result := nil;
  if csDestroying in ComponentState then Exit;
  for i := 0 to Count-1 do begin
    if not (Components[i] is TMNode) then
      raise Exception.CreateFmt('TMNode.FixupProperties:%s', [Components[i].Name]);
    if not (csDestroying in Items[i].ComponentState) then begin
      { Default is just to make sure it is called for all children }
      Result := Items[i].FixupProperties;
      if Assigned(Result) then
        Exit;
    end;
  end;
end;

function TMNode.GetParameter(AIndex: Integer): Double; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
{ if AIndex >= nParameters then
    Result := 0
  else }
    Result := FParameters[AIndex];
end;

procedure TMNode.SetParameter(AIndex: Integer; const Value: Double);
begin
  ModifyParameter(AIndex, Value, True);
end;

function TMNode.IndexOfParameter(const AName: string): Integer;
begin
  for Result := 1 to nParameters do
    if ParameterNames[Result] = AName then
      Exit;
  raise Exception.Create('IndexOfParameter: ' + AName + 'not found');
end;

procedure TMNode.ModifyParameter(AIndex: Integer; const Value: Double;
  ChangeModify: Boolean);
begin
  if IsNanOrInf(Value) then
    ModifyParameter (AIndex, FDefaultParameters[AIndex], ChangeModify)
  else if FParameters[AIndex] <> Value then begin
    FParameters[AIndex] := Value;
    if ChangeModify then
      Modified := True;
    PerformMNodeNotification(Self, mnParameterChange, AIndex);
  end;
end;

function TMNode.GetSetting(AIndex: Integer): Integer; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
{ if AIndex >= nSettings then
    Result := 0
  else }
    Result := FSettings[AIndex];
end;

procedure TMNode.ModifySetting(AIndex: Integer; const AValue: Integer; ChangeModify: Boolean);
begin
  if FSettings[AIndex] <> AValue then begin
    FSettings[AIndex] := AValue;
    if ChangeModify then
      Modified := True;
    PerformMNodeNotification(Self, mnSettingChange, AIndex);
  end;
end;

procedure TMNode.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  ModifySetting(AIndex, AValue, True);
end;

function TMNode.GetFlag(AIndex: Integer): Boolean;
begin
  Result := GetSetting(AIndex) <> 0;
end;

procedure TMNode.SetFlag(AIndex: Integer; const AValue: Boolean);
begin
  SetSetting(AIndex, Ord(AValue));
end;

{ Special case handling for colour, to avoid lots of overrides }

function TMNode.GetColour(AIndex: Integer): TColor;
begin
  Result := GetSetting(AIndex);
end;

procedure TMNode.SetColour(AIndex: Integer; const AValue: TColor);
begin
  SetSetting(AIndex, AValue);
end;

function TMNode.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := False;
end;

function TMNode.GetSettingsNames(AIndex: Integer; var LowestValue:Integer): PPChar;
begin
  if IsSettingColour(AIndex) then
    Result := GetColourNames(LowestValue)
  else
    Result := nil; { Default is to use Delphi Mechanism }
end;

procedure TMNode.SetModified(const Value: Boolean);
var
  i: Integer;
begin
  FModified := Value;
  if Modified then begin
    { Propagate Setting Modified upwards }
    if Owner <> nil then
      (Owner as TMNode).Modified := True;
  end else begin
    { Propagate Clearing Modified downwards }
    for i := 0 to ComponentCount-1 do
      if Items[i] is TMNode then
        TMNode(Items[i]).Modified := False;
  end;
end;

destructor TMNode.Destroy;
begin
  if Owner <> nil then
    (Owner as TMNode).Modified := True;
  PerformMNodeNotification(Self, mnDestroy, -1);
  FreeAndNil(FParameterNames);
  FreeAndNil(FSettingNames);
  inherited Destroy;
end;

function TMNode.GetFloating(AIndex: Integer): Boolean;
begin
  Result := FFloating[Aindex];
end;

procedure TMNode.SetFloating(AIndex: Integer; const Value: Boolean);
begin
  if FFloating[Aindex] <> Value then begin
    FFloating[Aindex] := Value;
    PerformMNodeNotification(Self, mnParameterChange, AIndex);
  end;
end;

function TMNode.GetIncrement(AIndex: Integer): Double;
begin
  Result := FIncrements[AIndex];
end;

procedure TMNode.SetIncrement(AIndex: Integer; const Value: Double);
begin
  if FIncrements[AIndex] <> Value then begin
    FIncrements[AIndex] := Value;
    PerformMNodeNotification(Self, mnParameterChange, AIndex);
  end;
end;

function TMNode.UniqueName: string;
begin
  Result := Name;
  if Owner<> nil then
    Result := (Owner as TMNode).UniquePrefix + Result;
end;

function TMNode.UniquePrefix: string;
var
  i, nSameClass: Integer;
begin
  Result := '';
  if not Assigned(Owner) then Exit;
  nSameClass := 0;
  for i := 0 to Owner.ComponentCount-1 do
    if Self.ClassType = Owner.Components[i].ClassType then
      Inc(nSameClass);
  if nSameClass > 1 then
    Result := UniqueName + ' ' + Result
  else
    Result := (Owner as TMNode).UniquePrefix + Result;
end;

function TMNode.RootNode: TMNode;
begin
  Result := Self;
  while Assigned(Result.Owner) do
    Result := Result.Owner as TMNode;
end;

procedure TMNode.SetName(const Value: TComponentName);
begin
  if FName <> Value then begin
    FName := Value;
    Modified := True;
    PerformMNodeNotification(Self, mnNameChange, -1);
  end;
end;

procedure TMNode.FixAll(Recurse: Boolean);
var
  i: Integer;
begin
  for i := 1 to nParameters do
    Floating[i] := False;
  if Recurse then
    for i := 0 to Count-1 do
      Items[i].FixAll(Recurse);
end;

procedure TMNode.ScaleParameters(UFrom, UTo: TResonanceUnits);
var
  i: Integer;
  s: string;
begin
  s := ','+LowerCase(DimensionlessParameters)+',';
  for i := 1 to nParameters do
    if Pos(','+LowerCase(ParameterNames[i])+',', s) = 0 then
      Parameters[i] := Convert(Parameters[i], UFrom, UTo);
  for i := 0 to Count-1 do
    Items[i].ScaleParameters(UFrom, UTo);
end;

procedure TMNode.ReadFromStream(Stream: TStream);
var
  XMLDocument: TXMLDocument;
begin
  ReadXMLFile(XMLDocument, Stream);
  try
    ReadProperties(XMLDocument.FirstChild);
    CheckProperties;
  finally
    XMLDocument.Free;
  end;
end;

procedure TMNode.SaveToStream(Stream: TStream; WithLinkedMNodes: Boolean = False);
var
  XMLDocument: TXMLDocument;
  ListNode: TDOMNode;
begin
  XMLDocument := TXMLDocument.Create;
  try
    if WithLinkedMNodes then begin
      ListNode := XMLDocument.OwnerDocument.CreateElement('List');
      XMLDocument.AppendChild(ListNode);
      WriteProperties(ListNode);
      LinkedNodes(False, ListNode);
    end else
      WriteProperties(XMLDocument);
    WriteXMLFile(XMLDocument, Stream);
  finally
    XMLDocument.Free;
  end;
end;

{$IFNDEF CONSOLE}
procedure TMNode.CopyToClipboard(WithLinkedMNodes: Boolean);
var
  Stream, Stream1: THugeMemoryStream;
begin
  CheckProperties;
  Stream1 := nil;
  Stream := THugeMemoryStream.Create;
  try
    SaveToStream(Stream, WithLinkedMNodes);
    Stream1 := THugeMemoryStream.Create;
    if not SimpleSaveToStream(Stream1, WithLinkedMNodes) then begin
      Stream.Position := 0;
      Stream1.CopyFrom(Stream, Stream.Size);
    end;
    Clipboard.Open;
    try
      Stream1.CopyToClipboard(CF_TEXT);
      Stream.CopyToClipboard(MNodeClipboardFormat);
    finally
      Clipboard.Close;
    end;
  finally
    Stream.Free;
    Stream1.Free;
  end;
end;
{$ENDIF}

var
  MNodeNotifications: array of OnMnodeChange;

procedure AddMNodeNotification(MNodeChange: OnMnodeChange);
begin
  SetLength(MNodeNotifications, Length(MNodeNotifications)+1);
  MNodeNotifications[High(MNodeNotifications)] := MNodeChange;
end;

procedure RemoveMNodeNotification(MNodeChange: OnMnodeChange);
var
  i: Integer;
begin
  i := 0;
  if Length(MNodeNotifications) = 0 then Exit; { Probably already finalised unit }
//  while (i < Length(MNodeNotifications)) and (@MNodeNotifications[i] <> @MNodeChange) do
  while (i < Length(MNodeNotifications))
        and (TMethod(MNodeNotifications[i]).Data <> TMethod(MNodeChange).Data)
        and (TMethod(MNodeNotifications[i]).Code <> TMethod(MNodeChange).Code) do
    Inc(i);
  if i >= Length(MNodeNotifications) then
    raise Exception.Create('Removal of non-existant MNodeNotification');
  while i < High(MNodeNotifications) do begin
    MNodeNotifications[i] := MNodeNotifications[i+1];
    Inc(i);
  end;
  SetLength(MNodeNotifications, Length(MNodeNotifications)-1);
end;

procedure PerformMNodeNotification(Sender: TMnode; Change: TMNodeChange; Index: Integer);
var
  i: Integer;
begin
  for i := 0 to High(MNodeNotifications) do
    MNodeNotifications[i](Sender, Change, Index);
end;

function TMNode.ComponentWithName(const AName: string; AClass: TClass): TMNode;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to ComponentCount-1 do
    if (Components[i] is AClass) and (TMNode(Components[i]).Name = AName) then begin
      Result := Components[i] as TMNode;
      if csDestroying in Result.ComponentState then
        Result := nil;
      Exit;
    end;
end;

procedure TMNode.CheckProperties;
var
  E: Exception;
begin
  E := FixupProperties;
  if Assigned(E) then
    raise E;
end;

procedure TMNode.SetComment(const Value: string);
begin
  if FComment <> Value then begin
    FComment := Value;
    Modified := True;
  end;
end;

function TMNode.GetName: TComponentName;
begin
  Result := Fname;
end;

function TMNode.LinkedNodes(Delete: Boolean; Parent: TDOMNode): Integer;
  function LinkedCount(MNode: TMNode):Integer;
  var
    i: Integer;
    LCount: Integer;
    function CheckLink(BraOrKet: TMNode): Boolean;
    begin
      Result := True;
      if BraOrKet = Self then begin
        if Assigned(Parent) and (LCount = 0) then
          MNode.WriteProperties(Parent);
        Inc(LCount);
        if Delete then begin
          FreeAndNil(MNode);
          Result := False;
        end;
      end;
    end;
  begin
    LCount := 0;
    if MNode is TMNodePair then begin
      if CheckLink(TMNodePair(MNode).Bra) then
        CheckLink(TMNodePair(MNode).Ket);
    end;
    Result := LCount;
    if not Assigned(MNode) or (MNode = Self) then Exit;
    for i := MNode.ComponentCount-1 downto 0 do
      if MNode.Components[i] is TMNode then
        Inc(Result,LinkedCount(TMNode(MNode.Components[i])));
  end;
begin
  Result := LinkedCount(RootNode);
end;

function TMNode.ChildWithClass(MNodeClass: TMNodeClass): TMNode;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Count-1 do
    if Items[i] is MNodeClass then begin
      if Assigned(Result) then
        raise Exception.CreateFmt('More than one Child of %s with class %s', [Name, MNodeClass.ClassName]);
      Result := Items[i];
    end;
end;

function TMNode.HelpOnSetting(AIndex: Integer): string;
begin
  Result := '';
end;

function TMNode.HelpOnParameter(AIndex: Integer): string;
begin
 Result := '';
end;

function TMNode.nChildTypes: Integer;
begin
  Result := 1; { Sensible default }
end;

class function TMNode.DisplayClassName: string;
var
  i: Integer;
begin
  Result := Copy(ClassName, 2, 255);
  i := 2;
  while i <= Length(Result) do begin
    if Result[i] = UpperCase(Result[i]) then begin
      Insert(' ', Result, i);
      Inc(i);
    end;
    Inc(i);
  end;
end;

procedure TMNode.HandleAssignment(NameVal: string);
{ Handle Child.Child.property=value type strings }
{ TODO: Does not handle MNodePairs }
var
  dot, equals: Integer;
  MNode: TMNode;
  n: string;
  PropInfo: PPropInfo;
begin
  dot := Pos('.',NameVal);
  equals := Pos('=',NameVal);
  if equals <= 0 then
    raise Exception.Create('No value in '+NameVal);
  if (dot > 0) and (dot < equals) then begin
    { Child.property=value - find child }
    n := Copy(NameVal,1,dot-1);
    MNode := ComponentWithName(n, TMNode);
    if MNode = nil then
      raise Exception.CreateFmt('%s not found in %s', [n, NameVal]);
    MNode.HandleAssignment(Copy(NameVal,dot+1,Length(NameVal)));
  end else begin
    n := Copy(NameVal,1,equals-1);
    PropInfo := GetPropInfo(Self, n);
    if PropInfo = nil then
      raise Exception.CreateFmt('Property %s not found in %s', [n, NameVal]);
    SetPropertyFromStr(Self, PropInfo, Copy(NameVal,equals+1,Length(NameVal)));
  end;
end;

procedure TMNode.CycleUnits;
begin
  RootNode.CycleUnits;
end;

function TMNode.SimpleSaveToStream(Stream: TStream; WithLinkedMNodes: Boolean): Boolean;
begin
  Result := False;
end;


{ TMNodePair }

function TMNodePair.BraOwner: TMNode;
begin
  Result := Owner as TMNode;
end;

constructor TMNodePair.Create(AOwner: TComponent);
begin
  inherited;
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}DoMNodeChange);
end;

destructor TMNodePair.Destroy;
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}DoMNodeChange);
  inherited;
end;

procedure TMNodePair.DoMNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
  procedure Process(var State: TMNode; var StateName: string);
  begin
    if Sender = State then
      case Change of
        mnNameChange: StateName := Sender.Name;
        mnDestroy:    State := nil;
      end;
  end;
begin
  Process(FBra, FBraName);
  Process(FKet, FKetName);
end;

function TMNodePair.FixupProperties: Exception;
var
  BO, KO: TMNode;
  ErrMsg: string;
  procedure FixUpStateS(N1, N2: string);
  begin
    if {not Assigned(FBra) and } Assigned(BO) then
      FBra := BO.ComponentWithName(N1, BraKetClass)
    else
      FBra := nil;
    if { not Assigned(FKet) and } Assigned(KO) then
      FKet := KO.ComponentWithName(N2, BraKetClass)
    else
      FKet := nil;
  end;
begin
  BO := BraOwner;
  KO := KetOwner;
  FixUpStates(FBraName, FKetName);
  if not Assigned(FBra) and not Assigned(FKet) and (KO <> BO) then begin
    { Try other way round }
    FixUpStates(FKetName, FBraName);
  end;
  ErrMsg := '';
  if not Assigned(FBra) then
    ErrMsg := FBraName;
  if not Assigned(FKet) then
    ErrMsg := FKetName;
  if not Assigned(FKet) or not Assigned(FBra) then
    Result := ECantFindState.Create(Self, ErrMsg)
  else begin
    { This will swap names over if required }
    FBraName := FBra.Name;
    FKetName := FKet.Name;
    Result := nil;
  end;
  if not Assigned(Result) then
    Result := inherited FixupProperties;
end;

function TMNodePair.KetOwner: TMNode;
begin
  Result := BraOwner;
end;

procedure TMNodePair.ReadProperties(Node: TDOMNode);
begin
  inherited ReadProperties(Node);
  FBraName := ReadValue(Node, 'Bra', '');
  FKetName := ReadValue(Node, 'Ket', '');
  { Below for compatibility only }
  if FBraName = '' then
    FBraName := ReadValue(Node, 'Manifold1', '');
  if FBraName = '' then
    FBraName := ReadValue(Node, 'State1', '');
  if FKetName = '' then
    FKetName := ReadValue(Node, 'Manifold2', '');
  if FKetName = '' then
    FKetName := ReadValue(Node, 'State2', '');
end;

function TMNodePair.SetBraKetNameE(const ABraName, AKetName: string; ChangingMNode: TMNode): Exception;
begin
  Result := nil;
  if FBraName <> ABraName then begin
    FBraName := ABraName;
    FBra := nil;
  end;
  if FKetName <> AKetName then begin
    FKetName := AKetName;
    FKet := nil;
  end;
  if not Assigned(FKet) or not Assigned(FBra) then begin
    if Assigned(ChangingMNode) then
      Result := ChangingMNode.FixupProperties
    else
      Result := FixupProperties;
    if not Assigned(Result) then begin
      Modified := True;
      PerformMNodeNotification(Self, mnNameChange, -1);
    end;
  end;
end;

procedure TMNodePair.SetBraKetName(const ABraName, AKetName: string);
var
  E: Exception;
begin
  E := SetBraKetNameE(ABraName, AKetName);
  if Assigned(E) then
    raise E;
end;

function TMNodePair.GetName: TComponentName;
begin
  Result := '<';
  if Assigned(Bra) then
    Result := Result + Bra.Name
  else
    Result := Result + '?';
  Result := Result + '|' + PairName + '|';
  if Assigned(Ket) then
    Result := Result + Ket.Name
  else
    Result := Result + '?';
  Result := Result + '>';
end;

procedure TMNodePair.SetName(const Value: TComponentName);
begin
  { Ignore }
end;

function TMNodePair.WriteProperties(Parent: TDOMNode): TDomNode;
begin
  Result:=inherited WriteProperties(Parent);
  if Assigned(FBra) then
    FBraName := FBra.Name;
  if Assigned(FKet) then
    FKetName := FKet.Name;
  WriteValue(Result, 'Bra', FBraName, '');
  WriteValue(Result, 'Ket', FKetName, '');
end;

{ ECantFindState }

constructor ECantFindState.Create(AMNodePair: TMNodePair; const AMissingState: string);
begin
  inherited Create('Can''t find State:' + AMissingState);
  FMissingState := AMissingState;
  FMNodePair := AMNodePair;
end;

end.
