unit Obs;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics, Classes, Mixture, QuantumNumber, TextStrm, MNode, Units;

type
  TObservation = record
    Frequency:        Double;
    Weight:           Double;
    ManifoldU, ManifoldL: TManifold;
    JU, JL:           TQuantumNumber;
    SymU, SymL:       Integer;
    IndexU, IndexL:   Integer; { State number - 0 to search }
    AIndexU, AIndexL: Integer; { Actual state number used }
    Name:             string;
    FileName:         string;  { Name of source file }
    LineNumber:       Integer;
    Colour:           TColor;  { Note that this can change following a fit if the colour depends on the State numbers }
  end;

  TObservationsReader = class
  protected
    Obs: TObservation; { Current Observation }
    s: string;         { Current Token }
  public
    nErrors, nObs: Integer;
    Mixture: TMixture;
    DefaultMolecule: TMolecule;
    DefaultManifoldU, DefaultManifoldL: TManifold;
    f: TTextStream;
    FileName: string;
    Take: (tkAll, tkObs, tkCalc);
    OverrideWeight, ScaleWeight: Double;
    IndexOffsetsL, IndexOffsetsU: TIntegerArray;
    Units: TResonanceUnits;
    FrequencyOffset, FrequencyScale: Double;

    constructor Create; overload; virtual;
    constructor Create(AMixture: TMixture); overload;
    constructor Create(AReader: TObservationsReader); overload;
    procedure ResetDefaultMolecule;
    procedure ResetDefaultManifoldU;
    procedure ResetDefaultManifoldL;

    function ReadStream(AStream: TTextStream; AName: string):Boolean; overload;
    function ReadStream(AStream: TStream; AName: string):Boolean; overload;
    function ReadInteger(var a: Integer): Boolean;
    function ReadSymmetry(var Sym: Integer; J: TQuantumNumber; Molecule: TMolecule): Boolean;
    function ReadInt(var a: Integer; const si: string): Boolean;
    function NextItem: string;
    procedure ReportError(es: string);
    function ReadOptionalManifold(Molecule: TMolecule): TManifold;

    function ReadOneLine: Boolean; virtual;
    function DoReadOneLine: Boolean; virtual; abstract;
    function HandleDirective(var Handled: Boolean):Boolean; virtual;
  end;

  TObservationsReaderClass = class of TObservationsReader;

var
  Observations: array of TObservation; { Global Variable for now }

function AddToObservations(Mixture: TMixture; FileName:string; Stream: TStream = nil; ReaderClass: TObservationsReaderClass = nil; CurrentReader: TObservationsReader = nil): Boolean;
procedure ClearObservations;
procedure LoadObservations(Mixture: TMixture; Stream: TStream; FileName:string);
function FindIndex(Eigenvalues: TEigenvalues; Name: string; Lower: Boolean; LineNumber: Integer):Integer;
function FindEigenvaluesAndIndex(var Eigenvalues: TEigenvalues; Manifold: TManifold; J: TQuantumNumber; Sym, Index: Integer; Name: string;
                                 Lower: Boolean; LineNumber: Integer; const FileName: string):Integer;

implementation

uses SysUtils, JPLCat;

procedure ClearObservations;
begin
  SetLength(Observations, 0);
end;

type

  TPgopherReader = class(TObservationsReader)
    function DoReadOneLine: Boolean; override;
  end;

  TOldPgopherReader = class(TObservationsReader)
    function DoReadOneLine: Boolean; override;
  end;

function FindIndex(Eigenvalues: TEigenvalues; Name: string; Lower: Boolean; LineNumber: Integer):Integer;
var
  e, Sep: Integer;
begin
  { Identify section with transition in - use presence of " - " to flag this }
  repeat
    Sep := Pos(' - ', Name);
    if Sep <= 0 then
      Sep := Pos(#9'-'#9, Name);
    e := Pos(':', Name);
    if e <= 0 then
      Break;
    { Note that if no " - " state separator, the level name must be first }
    if Sep < e then
      Delete(Name, e, Length(Name))
    else
      Delete(Name, 1, e);
    Name := Trim(Name);
  until False;

  { Identify upper or lower state name }
  if Sep > 0 then
    if Lower then
      Delete(Name, 1, Sep+2)
    else
      Delete(Name, Sep, Length(Name));
  for Result := 1 to Eigenvalues.Size do
    if Eigenvalues.DominantBasis[Result].StateLabel = Name then
      Exit;
  Writeln;
  Writeln('No match for ' + Name + ' on line number ' + IntToStr(LineNumber) + ' Possible matches:');
  for Result := 1 to Eigenvalues.Size do
    Writeln(Eigenvalues.DominantBasis[Result].StateLabel);
  Result := 0;
end;

function FindEigenvaluesAndIndex(var Eigenvalues: TEigenvalues; Manifold: TManifold; J: TQuantumNumber; Sym, Index: Integer; Name: string; Lower: Boolean; LineNumber: Integer; const FileName: string):Integer;
var
  s: String;
  procedure Error(s: string);
  begin
    WriteLn(s, ' on line number ', LineNumber, ' of ', FileName);
    Result := 0;
  end;
begin
  try
    Eigenvalues := Manifold.Eigenvalues(J, Sym, [], 0);
    Result := Index;
    if Result <= 0 then
      Result := FindIndex(Eigenvalues, Name, Lower, LineNumber);
    if (Result > 0) and (Result > Eigenvalues.Size) then begin
      if Lower then
        s := 'Lower'
      else
        s := 'Upper';
      Error(Format('%s eigenvalue not found (only %d present)', [s, Eigenvalues.Size]));
    end;
  except
    on E:Exception do
      Error(E.Message);
  end;
end;

{ TObservationsReader }

constructor TObservationsReader.Create;
begin
  inherited;
  nErrors := 0;
  Obs.ManifoldU := DefaultManifoldU;
  Obs.ManifoldL := DefaultManifoldL;
end;

constructor TObservationsReader.Create(AMixture: TMixture);
begin
  Mixture := AMixture;
  ResetDefaultMolecule;
  ResetDefaultManifoldU;
  ResetDefaultManifoldL;
  OverrideWeight := -1;
  ScaleWeight := 1;
  Take := tkObs;
  Units := Mixture.PlotUnits;
  FrequencyOffset := 0;
  FrequencyScale := 1;
  Create;
end;

constructor TObservationsReader.Create(AReader: TObservationsReader);
begin
  Mixture := AReader.Mixture;
  DefaultMolecule := AReader.DefaultMolecule;
  DefaultManifoldU := AReader.DefaultManifoldU;
  DefaultManifoldL := AReader.DefaultManifoldL;
  OverrideWeight := AReader.OverrideWeight;
  ScaleWeight := AReader.ScaleWeight;
  IndexOffsetsU := Copy(AReader.IndexOffsetsU);
  IndexOffsetsL := Copy(AReader.IndexOffsetsL);
  Take := AReader.Take;
  Units := AReader.Units;
  FrequencyOffset := AReader.FrequencyOffset;
  FrequencyScale := AReader.FrequencyScale;
  Create;
end;

procedure TObservationsReader.ResetDefaultManifoldU;
begin
  if Assigned(DefaultMolecule) and (DefaultMolecule.nManifolds <=2) then
    if (DefaultMolecule.nManifolds = 1) { Take likely default manifold }
       or (not DefaultMolecule.Manifold[0].Initial and DefaultMolecule.Manifold[1].Initial) then
      DefaultManifoldU := DefaultMolecule.Manifold[0]
    else
      DefaultManifoldU := DefaultMolecule.Manifold[1]
  else
    DefaultManifoldU := nil;
end;

procedure TObservationsReader.ResetDefaultManifoldL;
begin
  if Assigned(DefaultMolecule) and (DefaultMolecule.nManifolds <=2) then
    if (DefaultMolecule.nManifolds = 1) { Take likely default manifold }
       or (DefaultMolecule.Manifold[0].Initial) then
      DefaultManifoldL := DefaultMolecule.Manifold[0]
    else
      DefaultManifoldL := DefaultMolecule.Manifold[1]
  else
    DefaultManifoldL := nil;
end;

procedure TObservationsReader.ResetDefaultMolecule;
begin
  if (Mixture.nMolecules = 1) and (Mixture.Species[0].Count = 1) then
    DefaultMolecule := Mixture.Species[0].Isotopomer[0] { Default to only molecule }
  else
    DefaultMolecule := nil;
end;

procedure TObservationsReader.ReportError(es: string);
begin
  Inc(nErrors);
  WriteLn('Error on line ', f.Line, ' of ', FileName, ': ', es);
end;

function TObservationsReader.NextItem: string;
begin
  if not f.GetFreeItem then
    ReportError('Not enough items on line');
  Result := f.Token;
end;

function TObservationsReader.ReadInt(var a: Integer; const si: string): Boolean;
begin
  Result := not TryStrToInt(si, a);
  if Result then
    ReportError('Bad integer: '+ si);
end;

function TObservationsReader.ReadInteger(var a: Integer): Boolean;
var
  si: string;
begin
  si := NextItem;
  Result := ReadInt(a, si);
end;

function TObservationsReader.ReadSymmetry(var Sym: Integer; J: TQuantumNumber; Molecule: TMolecule): Boolean;
var
  si: string;
  sy: Integer;
begin
  Result := False;
  si := NextItem;
  if TryStrToInt(si, Sym) then begin
    Result := (Sym < 0) or (Sym > Molecule.LastSym);
    if Result then
      ReportError('Symmetry number not in range 0..'+IntToStr(Molecule.LastSym));
    Exit;
  end;
  for sy := 0 to Molecule.LastSym do
    if CompareText(si,Molecule.SymName(sy)) = 0 then begin
      Sym := sy;
      Exit;
    end;
  { TODO: consider (A)Symmetric top case carefully - logic below works only if names distinct }
  for sy := 0 to Molecule.LastSym do
    if CompareText(si,Molecule.JAdjustedSymName(J,sy)) = 0 then begin
      Sym := sy;
      Exit;
    end;
  Result := True;
  ReportError('Bad symmetry: ' + si);
end;

function TObservationsReader.ReadStream(AStream: TTextStream; AName: string): Boolean;
begin
  f := AStream;
  FileName := AName;
  nObs := Length(Observations);
  while not f.AtEndOfFile do begin
    if ReadOneLine then
      Break;
    f.NextLine;
  end;
  SetLength(Observations, nObs);
  Result := nErrors > 0;
end;

function AddToObservations(Mixture: TMixture; FileName:string; Stream: TStream = nil; ReaderClass: TObservationsReaderClass = nil; CurrentReader: TObservationsReader = nil): Boolean;
var
  Reader: TObservationsReader;
  FileStream: TFileStream;
begin
  if ReaderClass = nil then
    if LowerCase(ExtractFileExt(FileName)) = '.cat' then
      ReaderClass := TJPLCatReader
    else
      ReaderClass := TPgopherReader;
  if Assigned(CurrentReader) then
    Reader := ReaderClass.Create(CurrentReader)
  else
    Reader := ReaderClass.Create(Mixture);
  try
    if Assigned(Stream) then
      Result := Reader.ReadStream(Stream, FileName)
    else begin
      Result := True;
      FileStream := TFileStream.Create(FileName,fmOpenRead or fmShareDenyWrite);
      try
        Result := Reader.ReadStream(FileStream, FileName);
      finally
        FileStream.Free;
      end;
    end;
  finally
    Reader.Free;
  end
end;

function TObservationsReader.ReadStream(AStream: TStream; AName: string): Boolean;
var
  TextStream: TTextStream;
begin
  if AStream is TTextStream then
    Result := ReadStream(AStream as TTextStream, AName)
  else begin
    TextStream := TTextStream.Create(AStream);
    try
      Result := ReadStream(TextStream, AName);
    finally
      TextStream.Free;
    end;
  end;
end;

function TObservationsReader.ReadOneLine: Boolean;
var
  Handled: Boolean;
  procedure FixOffset(J: TQuantumNumber; Sym: Integer; var Index: Integer; const IndexOffsets: array of Integer; Manifold: TManifold);
  var
    i: Integer;
  begin
    if (Sym >= 0) and (Index > 0) and (Length(IndexOffsets) > 0) then begin
      i := Manifold.nSym*(J div 2) + Manifold.Molecule.JAdjustedSym(J,Sym);
      while i > High(IndexOffsets) do
        Dec(i, Manifold.nSym);
      if i < 0 then
        i := 0;
      // WriteLn(J:4, Sym:4, Index:4, IndexOffsets[i]:4);
      Inc(Index,IndexOffsets[i]);
      if Index <= 0 then
        ReportError('Offset index = '+IntToStr(Index));
    end;
  end;
begin
  Result := False;
  if not f.GetFreeItem or (f.token[1] in ['/', '#', ';']) then
    Exit; { Blank Line or comment }
  s := f.Token;
  Handled := True;
  Result := HandleDirective(Handled);
  if not Handled and DoReadOneLine then begin
    Obs.Colour := Obs.ManifoldL.Colour;
    if Obs.Colour = clNone then
      Obs.Colour := Obs.ManifoldU.Colour;
    Obs.LineNumber := f.Line;
    Obs.FileName := FileName;
    Obs.Frequency := (Obs.Frequency + FrequencyOffset)*FrequencyScale;
    if Units <> Mixture.PlotUnits then begin
      if Obs.Weight > 0 then begin
        Obs.Weight := ConvertWidth(Obs.Weight*FrequencyScale, Obs.Frequency, Units, Mixture.PlotUnits);
      end;
      Obs.Frequency := Convert(Obs.Frequency,Units,Mixture.PlotUnits);
    end;
    if (OverrideWeight >= 0) and (Obs.Weight > 0) then
      Obs.Weight := OverrideWeight;
    Obs.Weight := Obs.Weight*ScaleWeight;
    if (Take <> tkObs) or (Obs.Weight > 0) then begin
      FixOffset(Obs.JL, Obs.SymL, Obs.IndexL, IndexOffsetsL, Obs.ManifoldL);
      FixOffset(Obs.JU, Obs.SymU, Obs.IndexU, IndexOffsetsU, Obs.ManifoldU);
      if nObs >= Length(Observations) then
        SetLength(Observations, Length(Observations)+1000);
      Observations[nObs] := Obs;
      Inc(nObs);
    end;
  end;
end;

function TObservationsReader.HandleDirective(var Handled: Boolean): Boolean;
var
  Cmd: string;
  procedure ReadFile(IncName: string; ReaderClass: TObservationsReaderClass);
  var
    Stream: TStream;
  begin
    SetLength(Observations, nObs);
    try
      Stream := nil;
      if IncName = '' then
        Stream := f
      else if not (IncName[1] in {$IFDEF FPC}DirSeparators{$ELSE}['\','/']{$ENDIF}) then
        IncName := ExtractFileDir(FileName) + PathDelim + IncName;
      if AddToObservations(Mixture, IncName, Stream, ReaderClass, Self) then
        Inc(nErrors);
    except
      on E: Exception do
        ReportError(E.Message);
    end;
    nObs := Length(Observations);
  end;
  procedure ReadIndexOffsets(var IndexOffsets: TIntegerArray);
  var
    i, Offset: Integer;
  begin
    i := 0;
    while f.GetFreeItem do
      if TryStrToInt(f.Token,Offset) then begin
        if i > High(IndexOffsets) then begin
          SetLength(IndexOffsets, i+1);
          IndexOffsets[i] := Offset;
        end else
          Inc(IndexOffsets[i],Offset);
        Inc(i);
      end;
    SetLength(IndexOffsets, i);
  end;
  function ReadManifold(var Manifold: TManifold):Boolean;
  begin
    Result := True;
    if not f.GetFreeItem then
      Exit;
    s := f.Token;
    if not Assigned(DefaultMolecule) then
      ReportError('Default molecule must be assigned')
    else begin
      Manifold := ReadOptionalManifold(DefaultMolecule);
      if (Manifold = nil) and (s <> 'none') then
        ReportError('Unknown Manifold:'+s)
      else
        Result := False;
    end;
  end;
begin
  Result := False;
  Cmd := LowerCase(s);
  if (s = '-1') or (Cmd = 'stop') then
    Result := True
  else if Cmd = 'include' then
    if not f.GetFreeItem then
      ReportError('No filename after include')
    else
      ReadFile(f.Token, nil)
  else if Cmd = 'includeoldpgopher' then begin
    if f.GetFreeItem then
      ReadFile(f.Token, TOldPgopherReader)
    else
      ReadFile('', TOldPgopherReader);
  end else if Cmd = 'molecule' then begin
    if f.GetFreeItem then
      s := f.Token
    else
      s := '';
    DefaultMolecule := Mixture.FindMolecule(s);
    if (DefaultMolecule = nil) and (s <> 'none') then begin
      ReportError('Unknown Molecule:'+s);
      ResetDefaultMolecule;
    end;
  end else if Cmd = 'uppermanifold' then begin
    if ReadManifold(DefaultManifoldU) then
        ResetDefaultManifoldU;
  end else if Cmd = 'lowermanifold' then begin
    if ReadManifold(DefaultManifoldL) then
        ResetDefaultManifoldL;
  end else if Cmd = 'overrideweight' then
    if f.GetFreeItem then
      OverrideWeight := f.RealToken
    else
      OverrideWeight := -1
  else if Cmd = 'scaleweights' then
    if f.GetFreeItem then
      ScaleWeight := f.RealToken
    else
      ScaleWeight := 1
  else if Cmd = 'frequencyoffset' then
    if f.GetFreeItem then
      FrequencyOffset := f.RealToken
    else
      FrequencyOffset := 0
  else if Cmd = 'frequencyscale' then
    if f.GetFreeItem then
      FrequencyScale := f.RealToken
    else
      FrequencyScale := 1
  else if Cmd = 'takeall' then
    Take := tkAll
  else if Cmd = 'takeobs' then
    Take := tkObs
  else if Cmd = 'takecalc' then
    Take := tkCalc
  else if Cmd = 'upperindexoffsets' then
    ReadIndexOffsets(IndexOffsetsU)
  else if Cmd = 'lowerindexoffsets' then
    ReadIndexOffsets(IndexOffsetsL)
  else if Cmd = 'indexoffsets' then begin
    ReadIndexOffsets(IndexOffsetsU);
    IndexOffsetsL := Copy(IndexOffsetsU);
  end else if Cmd = 'units' then begin
    if f.GetFreeItem then
      Units := StrToUnits(f.Token)
    else
      Units := Mixture.Units;
  end else
    Handled := False;
end;

function TObservationsReader.ReadOptionalManifold(Molecule: TMolecule): TManifold;
begin
  Result := Molecule.ManifoldWithName(s);
  if not Assigned(Result) and not (s[1] in ['.', '+', '-', '0'..'9']) then begin
    { Must be a Manifold name, as it is not a number;
      Allow for Manifold name to have spaces in it }
    while Molecule.ManifoldStartingWithName(s) <> nil do begin
      s := s + ' ' + NextItem;
      Result := Molecule.ManifoldWithName(s);
      if Assigned(Result) then Break;
    end;
    if not Assigned(Result) then
      ReportError('Mainfold not found: ' + s);
  end;
  if Assigned(Result) then
    if f.GetFreeItem then
      s := f.Token
    else
      s := '';
end;

{ TPgopherReader }

function TPgopherReader.DoReadOneLine: Boolean;
var
  Molecule: TMolecule;
begin
  Result := False;

  Molecule := Mixture.FindMolecule(s);
  if Molecule = nil then
    Molecule := DefaultMolecule
  else
    s := NextItem;
  if Molecule = nil then begin
    ReportError('Molecule ' + s + ' not found');
    Exit;
  end;

  Obs.ManifoldU := ReadOptionalManifold(Molecule);
  if Obs.ManifoldU = nil then
    Obs.ManifoldU := DefaultManifoldU;
  if Obs.ManifoldU = nil then begin
    ReportError('Manifold ' + s + ' not found');
    Exit;
  end;
  { Optinally allow manifoldL here }
  Obs.ManifoldL := ReadOptionalManifold(Molecule);
  if Pos('(',s) > 0 then
    // ReadBranchFormat
    raise NotYet
  else begin

    if ReadInt(Obs.JU, s) then Exit;
    if ReadSymmetry(Obs.SymU, Obs.JU, Molecule) then Exit;
    if ReadInteger(Obs.IndexU) then Exit;
    if Obs.IndexU < 0 then ReportError('(Upper) Eigenvalue number must be >=0');

    s := NextItem;
    if Obs.ManifoldL = nil then begin
      Obs.ManifoldL := ReadOptionalManifold(Molecule);
      if Obs.ManifoldL = nil then
        Obs.ManifoldL := DefaultManifoldL;
    end;
    if Obs.ManifoldL = nil then begin
      ReportError('Manifold ' + s + ' not found');
      Exit;
    end;
    if ReadInt(Obs.JL,s) then Exit;
    if ReadSymmetry(Obs.SymL, Obs.JL, Molecule) then Exit;
    if ReadInteger(Obs.IndexL) then Exit;
    if Obs.IndexL < 0 then ReportError('(Lower) Eigenvalue number must be >=0');
  end;

  s := NextItem;
  if not TryStrToFloat(s, Obs.Frequency) then begin
    ReportError('Bad frequency:' + s);
    Exit;
  end;
  s := NextItem;
  if s = '-' then
    Exit; { Not weighted - skip }
  if not TryStrToFloat(s, Obs.Weight) then begin
    ReportError('Bad weight:' + s);
    Exit;
  end;
  Obs.Name := TrimRight(f.RestOfLine);
  if (Obs.Name <> '') and (Obs.Name[1] in [#9, ' ']) then
    Delete(Obs.Name, 1, 1);
  Result := True;
end;

procedure LoadObservations(Mixture: TMixture; Stream: TStream; FileName:string);
begin
  ClearObservations;
  if AddToObservations(Mixture,FileName,Stream) then
    raise EReportException.Create('See log windows for errors detected in loading ' + Filename);
end;

{ TOldPgopherReader }

function TOldPgopherReader.DoReadOneLine: Boolean;
begin
  Result := False;
  if ReadInt(Obs.JU, s) then Exit;
  if ReadInteger(Obs.SymU) then Exit;
  Obs.SymU := (1 - Obs.SymU) div 2;
  if (Obs.SymU < 0) or (Obs.SymU > DefaultMolecule.LastSym) then
    ReportError('Upper Symmetry number not in range 0..'+IntToStr(DefaultMolecule.LastSym));
  if ReadInteger(Obs.IndexU) then Exit;
  if Obs.IndexU <= 0 then ReportError('(Upper) Eigenvalue number must be > 0');

  if ReadInteger(Obs.JL) then Exit;
  if ReadInteger(Obs.SymL) then Exit;
  Obs.SymL := (1 - Obs.SymL) div 2;
  if (Obs.SymL < 0) or (Obs.SymL > DefaultMolecule.LastSym) then
    ReportError('Lower Symmetry number not in range 0..'+IntToStr(DefaultMolecule.LastSym));
  if ReadInteger(Obs.IndexL) then Exit;
  if Obs.IndexL <= 0 then ReportError('(Lower) Eigenvalue number must be > 0');

  s := NextItem;
  if not TryStrToFloat(s, Obs.Frequency) then begin
    ReportError('Bad frequency:' + s);
    Exit;
  end;
  s := NextItem;
  if s = '-' then
    Exit; { Not weighted - skip }
  if not TryStrToFloat(s, Obs.Weight) then begin
    ReportError('Bad weight:' + s);
    Exit;
  end;
  Obs.Name := TrimRight(f.RestOfLine);
  if (Obs.Name <> '') and (Obs.Name[1] in [#9, ' ']) then
    Delete(Obs.Name, 1, 1);
  Result := True;
end;

end.
