unit Huge;
{ Note that the HugeClip unit must also be used for the clipboard routines to work }
interface
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
uses SysUtils, Classes;

type
  TGHandle = class(TPersistent)
  private
    function GetPointer(n:Longint):Pointer; {$IFDEF INLINE} inline; {$ENDIF}
    procedure SetSize(n:Longint);
  protected
    FSize:Longint;
    FBase:Pointer;
    {$IFDEF MSWINDOWS}
    FHandle:THandle;
      {$IFDEF VCL_}
      FromClipboard:Boolean;
      {$ENDIF}
    {$ENDIF}
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure Zero;
    property Size:Longint read FSize write SetSize;
    procedure CopyToClipboard(Format:Cardinal); { Notionally TClipboardFormat }
    procedure SetFromClipboard(Format:Cardinal);
    property Item[n:Longint]:Pointer read GetPointer;
    property Base: Pointer read FBase;
    procedure Assign(Source: TPersistent); override;
    procedure AssignAndNil(var Source: TGHandle);
  end;

  TGVectorN = class(TGHandle)
  private
    FItemSize: Cardinal;
    Mask: Cardinal;
    ShiftCount: Cardinal;
    procedure SetSize(n:Longint);
    function GetPointer(n:Longint):Pointer; {$IFDEF INLINE} inline; {$ENDIF}
    function GetHigh: Integer;
  protected
    function GetSize:Longint;
  public
    Offset: Integer;
    constructor Create(AItemSize:Cardinal);
    procedure Move(Src, Dst:Longint);
    property Size:Longint read GetSize write SetSize;
    property ItemSize:Cardinal read FItemSize;
    property Item[n:Longint]:Pointer read GetPointer;
    property High: Integer read GetHigh;
    procedure Assign(Source: TPersistent); override;
  end;

  { THugeMemoryStream }

  THugeMemoryStream = Class(TStream)
  private
    Buffer:TGHandle;
    FPosition:Longint;
    FSize: Longint;
    function GetMemory: Pointer;
  public
    constructor Create; overload;
    constructor Create(GHandle: TGHandle); overload;
    destructor Destroy; override;
    function GetCapacity:Longint;
    procedure SetCapacity(n:Longint);
    function Read(var Buf; Count: Longint): Longint;  override;
    function Write(const Buf; Count: Longint): Longint;  override;
    function Seek(Offset: Longint; Origin: Word): Longint;  override;
    procedure CopyToClipboard(Format:Cardinal);
    procedure SetFromClipboard(Format:Cardinal);
    property Capacity:Longint read GetCapacity write SetCapacity;
    property Memory: Pointer read GetMemory;
    procedure SetSize(NewSize: Longint); override;
    procedure Disconnect;  { Disconnect from underlying GHandle }
  end;

{$IFDEF VER80}
procedure HugeInc(var HugePtr:Pointer; Ammount:Longint);
{$ENDIF}

var
  GMemoryCount: Longint;
  CopyHandleToClipboard: procedure(GHandle: TGHandle; Format:Cardinal) = nil;
  SetHandleFromClipboard: procedure(GHandle: TGHandle; Format:Cardinal) = nil;

implementation

{$IFDEF MSWINDOWS}
uses Windows;
{$ENDIF}

{ THugeMemoryStream }

constructor THugeMemoryStream.Create(GHandle: TGHandle);
begin
  inherited Create;
  FPosition := 0;
  FSize := 0;
  Buffer := GHandle;
end;

constructor THugeMemoryStream.Create;
begin
  Create(TGHandle.Create);
end;

destructor THugeMemoryStream.Destroy;
begin
  FreeAndNil(Buffer);
  inherited Destroy;
end;

procedure THugeMemoryStream.SetSize(NewSize: Longint);
begin
  FSize := NewSize;
  Buffer.Size := NewSize; { Tuncate buffer contents }
end;

procedure THugeMemoryStream.Disconnect;
begin
  Buffer := nil;
end;

function THugeMemoryStream.GetCapacity:Longint;
begin
  Result := Buffer.Size;
end;

procedure THugeMemoryStream.SetCapacity(n:Longint);
begin
  Buffer.Size := n;
  if FSize > n then
    FSize := n;
end;

function THugeMemoryStream.Read(var Buf; Count: Longint): Longint;
var
  p:Pointer;
  n:Longint;
begin
  Result := 0;
  while (FPosition < FSize) and (Count > 0) do begin
    p := Buffer.Item[FPosition];
    n := Count;
{$IFDEF VER80}
    if Ofs(p^)+n >= 65536 then
      n := 65536-Ofs(p^);
    if n > 32768 then n := 32768; { Move can't cope with 65536 bytes }
{$ENDIF}
    if FPosition+n > FSize then
      n := FSize-FPosition;
    Move(p^, Buf, n);
    Count := Count - n;
    FPosition := FPosition + n;
    Result := Result + n;
  end;
end;

function THugeMemoryStream.Write(const Buf; Count: Longint): Longint;
var
  p:Pointer;
  n:Longint;
begin
  if FPosition+Count > Buffer.Size then
    Buffer.Size := FPosition+Count+65536;
  Result := 0;
  while Count > 0 do begin
    p := Buffer.Item[FPosition];
    n := Count;
{$IFDEF VER80}
    if Ofs(p^)+n >= 65536 then
      n := 65536-Ofs(p^);
    if n > 32768 then n := 32768;
{$ENDIF}
    Move(Buf, p^, n);
    Count := Count - n;
    FPosition := FPosition + n;
    Result := Result + n;
  end;
  if FPosition > FSize then
    FSize := FPosition;
end;

function THugeMemoryStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
    soFromBeginning: FPosition := Offset;
    soFromEnd: FPosition := FSize - Offset;
    soFromCurrent: FPosition := FPosition + Offset;
  else
    raise Exception.CreateFmt('THugeMemoryStream.Seek(%d,%d)', [Offset, Origin]);
  end;
  if FPosition < 0 then FPosition := 0;
  if FPosition > FSize then FPosition := FSize;
  Result := FPosition;
end;

procedure THugeMemoryStream.CopyToClipboard(Format:Cardinal);
begin
  Buffer.Size := FSize;
  Buffer.CopyToClipboard(Format);
end;

procedure THugeMemoryStream.SetFromClipboard(Format:Cardinal);
begin
  Buffer.SetFromClipboard(Format);
  FSize := Buffer.Size;
end;

function THugeMemoryStream.GetMemory: Pointer;
begin
  Result := Buffer.Item[0];
end;

{ TGHandle }

constructor TGHandle.Create;
begin
  FSize := 0;
  {$IFDEF MSWINDOWS}
  {$IFDEF VCL_}
  FromClipboard := False;
  {$ENDIF}
  FHandle := 0;
  {$ENDIF}
  FBase := nil;
end;

destructor TGHandle.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TGHandle.CopyToClipboard(Format:Cardinal);
begin
  if Assigned(CopyHandleToClipboard) then
    CopyHandleToClipboard(Self, Format);
end;

procedure TGHandle.SetFromClipboard(Format:Cardinal);
begin
  if Assigned(SetHandleFromClipboard) then
    SetHandleFromClipboard(Self, Format);
end;

procedure TGHandle.Clear;
begin
  {$IFDEF MSWINDOWS}
    if (FBase <> nil) and (FHandle <> 0) then
      GlobalUnlock(FHandle);
    if (FHandle <> 0) then
      {$IFDEF VCL_}
        if FromClipboard then
          SetHandleFromClipboard(nil,0) { Forces Clipboard.Close }
        else
      {$ENDIF}
        begin
          GlobalFree(FHandle);
          Dec(GMemoryCount, FSize);
        end;
    FHandle := 0;
    {$IFDEF VCL_}
      FromClipboard := False;
    {$ENDIF}
  {$ENDIF}
  {$IFDEF UNIX}
    if (FBase) <> nil then begin
      FreeMem(FBase);
      Dec(GMemoryCount, FSize);
    end;
  {$ENDIF}
  FBase := nil;
  FSize := 0;
end;

procedure TGHandle.Zero;
begin
  {$IFDEF VER80}
  TODO: Implement
  {$ELSE}
  FillChar(Fbase^, FSize, 0);
  {$ENDIF}
end;

procedure TGHandle.SetSize(n:Longint);
begin
  if n = 0 then
    Clear
  else if n <> FSize then begin
    Dec(GMemoryCount, FSize);
    {$IFDEF MSWINDOWS}
    FBase := nil;
    if FSize = 0 then
      FHandle := GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT, n)
    else begin
      GlobalUnlock(FHandle);
      FHandle := GlobalRealloc(FHandle, n, GMEM_MOVEABLE or GMEM_ZEROINIT);
    end;
    if FHandle <> 0 then
      FBase := GlobalLock(FHandle)
    else
      OutOfMemoryError;
    if Fbase = nil then
      raise Exception.Create('GlobalLock Fails');
    {$ENDIF}
    {$IFDEF UNIX}
    if FSize = 0 then
      GetMem(FBase, n)
    else begin
      FSize := 0;  { In case an exception raised }
      ReallocMem(FBase, n);
    end;
    {$ENDIF}
    FSize := n;
    Inc(GMemoryCount, FSize);
  end;
end;

function TGHandle.GetPointer(n:Longint):Pointer; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  {$IFOPT R+}
  if (n < 0) or (n >= FSize) then
    raise Exception.CreateFmt('TGHandle.GetPointer(%d) Size=%d', [n,FSize])
  else
{$ENDIF}
  begin
{$IFNDEF VER80}
    Result := PChar(FBase)+n;
{$ELSE}
    Result := FBase;
    HugeInc(Result, n)
{$ENDIF}
  end;
end;

procedure TGHandle.Assign(Source: TPersistent);
begin
  if Source is TGHandle then begin
    Size := TGHandle(Source).Size;
    {$IFDEF VER80}
    hmemcpy(Base, TGHandle(Source).Base, Size);
    {$ELSE}
    Move(TGHandle(Source).Base^, Base^, Size);
    {$ENDIF}
  end else
    inherited Assign(Source);
end;

procedure TGHandle.AssignAndNil(var Source: TGHandle);
begin
  Clear;
  FSize := Source.Size;
  Source.FSize := 0;
  FBase := Source.Base;
  Source.FBase := nil;
  {$IFDEF MSWINDOWS}
    FHandle := Source.FHandle;
    Source.FHandle := 0;
    {$IFDEF VCL_}
      FromClipboard := Source.FromClipboard;
    {$ENDIF}
  {$ENDIF}
  FreeAndNil(Source);
end;

{$IFDEF VER80}
procedure HugeShift; far; external 'KERNEL' index 113;
procedure HugeInc(var HugePtr:Pointer; Ammount:Longint); assembler;
asm
  mov ax, Ammount.Word[0]
  mov dx, Ammount.Word[2]
  les bx, HugePtr
  add ax, es:[bx]
  adc dx, 0
  mov cx, Offset HugeShift
  shl dx, cl
  add es:[bx+2], dx
  mov es:[bx], ax
end;
{$ENDIF}

constructor TGVectorN.Create(AItemSize:Cardinal);
begin
  inherited Create;
  Offset := 0;
  FItemSize := AItemSize;
  Mask := AItemSize - 1;
  ShiftCount := 0;
  while AItemSize > 1 do begin
    AItemSize := AItemSize div 2;
    Inc(ShiftCount);
  end;
  if Cardinal(1 shl ShiftCount) <> FItemSize then
    raise Exception.CreateFmt('TGVectorN: ItemSize(&d) not a power of 2', [FItemSize]);
end;

function TGVectorN.GetSize:Longint;
begin
  Result := (inherited Size) shr ShiftCount;
end;

procedure TGVectorN.SetSize(n:Longint);
begin
  inherited Size := n * Longint(FItemSize);
end;

function TGVectorN.GetPointer(n:Longint):Pointer; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  n := (n-Offset) shl ShiftCount;
  {$IFOPT R+}
  if (n < 0) or (n >= FSize) or ((n and Mask) <> 0) then
    raise Exception.CreateFmt('TGVectorN.GetPointer(%d) Size=%d', [n,Size])
  else
  {$ENDIF}
  begin
{$IFNDEF VER80}
    Result := PChar(FBase)+n;
{$ELSE}
    Result := FBase;
    HugeInc(Result, n)
{$ENDIF}
  end;
end;

procedure TGVectorN.Move(Src, Dst:Longint);
var
  Bold: array [0..255] of Char;
  P: PChar;
  i: Longint;
begin
  if ItemSize > 256 then
    raise Exception.CreateFmt('TGVectorN.Move: ItemSize=%d', [ItemSize]);
  P := GetPointer(Src);
  System.Move(P^, Bold, ItemSize);
  if Src > Dst then begin
    for i := 1 to Src-Dst do begin
      System.Move((P-ItemSize)^, P^, ItemSize);
      Dec(P,ItemSize);
    end;
  end else begin
    for i := 1 to Dst-Src do begin
      System.Move((P+ItemSize)^, P^, ItemSize);
      Inc(P,ItemSize);
    end;
  end;
  System.Move(Bold, P^, ItemSize);
end;

procedure TGVectorN.Assign(Source: TPersistent);
var
  s: TGVectorN;
begin
  if Source is TGVectorN then begin
    s := TGVectorN(Source);
    if s.ItemSize <> ItemSize then
      raise Exception.CreateFmt('TGVectorN.Assign: ItemSize: %d <> %d', [s.ItemSize, ItemSize]);
    Offset := s.Offset;
    inherited Assign(Source); { Do the copy }
  end else
    inherited Assign(Source);
end;

function TGVectorN.GetHigh: Integer;
begin
  Result := Offset + Size - 1;
end;

end.
