unit TextStrm;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes;

type

  TItemSeparator = (isTab, isComma);

  TTextStream = class(TStream)
  private
    FColumn: Integer;
    FStream:TStream;
    pending:(pNone, pCharSaved, pNewline, pEOF, pRecord);
    SavedChar:Char;
    FLine:Longint;
    FItemRead: Boolean; { Flag if a free format item has been read, but
                          not the field terminator }
  protected
    procedure BadToken(msg:string);
  public
    NULLisEOF:Boolean;
    Token:string;
    ItemSeparators: set of TItemSeparator;
    FlowToNextLine: Boolean;
    FlowToNextRecord: Boolean; { / forces default to be read until nextline called;
                                 NextRecord makes sure a / has been read }
    Comments: Boolean; { If true, '/' '#' and ';' mark end of record }
    AllowDefaults: Boolean;
    constructor Create(Stream:TStream); virtual;
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint;  override;
    function Write(const Buffer; Count: Longint): Longint;  override;
    function Seek(Offset: Longint; Origin: Word): Longint;  override;
  { Write }
    function PutChar(c:Char):TTextStream;
    function PutFloat(v:Extended):TTextStream;
    function PutInteger(i:Longint):TTextStream;
    function PutString(s:string):TTextStream;
    function PutPChar(p:PChar):TTextStream;
    function Space:TTextStream;
    function Tab:TTextStream;
    function Newline:TTextStream;
    function Format(const FormatStr: string; const Args: array of const): TTextStream;
  { Read }
    function Getch:Char;
    function GetFreeItem:Boolean;
    function NextRecord:Boolean;
    function NextLine:Boolean; { Note that token is set to the (trimmed) text read }
    function IsRealToken(var Value:Double):Boolean;
    function RealToken:Double;
    function IntegerToken:Integer;
    function AtEndOfLine:Boolean;
    function AtEndOfFile:Boolean;
    function NextInteger(Default: Integer): Integer;
    function NextDouble(Default: Double): Double;
    function NextString(Default: string): string;
    function NextIntegerVar(var i: Integer):Boolean;
    function NextDoubleVar(out d: Double):Boolean;
    function NextStringVar(var s: string):Boolean;
    function RestOfLine: string;
    procedure SkipItems(n: Integer);
    property Line: Longint read FLine;
    property Column: Integer read FColumn;
  end;

  TFreeFormatStream = class(TTextStream)
  public
    constructor Create(Stream:TStream); override;
  end;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF} StrUtils, SysUtils, MathSup;

function IsEOLChar(ch:Char):Boolean;
begin
  Result := (ch = #10) or (ch = #13);
end;

procedure TTextStream.BadToken(msg:string);
begin
  raise Exception.CreateFmt('Error reading data file on line %d column %d: %s (%s)', [Line, Column, msg, Token]);
end;
                                      
function TTextStream.AtEndOfLine:Boolean;
begin
  Result := pending in [pEOF, pNewline, pRecord];
end;

function TTextStream.AtEndOfFile:Boolean;
begin
  Result := pending = pEOF;
end;

function TTextStream.NextRecord:Boolean;
begin
  if pending <> pEOF then begin
    if (pending <> pRecord) and not FlowToNextRecord then begin
      repeat GetFreeItem until pending in [pEOF, pRecord];
    end;
    pending := pNewline;
    Nextline;
  end;
  FItemRead := False;
  Result := pending = pEOF;
end;

function TTextStream.Nextline:Boolean;
var
  ch:Char;
begin
  if not(pending in [pEOF, pRecord]) then begin
    if pending = pNone then
      SavedChar := Getch;
    pending := pNewline;
    Token := '';
    repeat
      ch := Getch;
      if pending = pEOF then Break;
      if IsEOLChar(SavedChar) then
        { Either LF or CR forces end of line; accept either order,
          but two LF's implies two newlines }
        if IsEOLChar(ch) and (ch <> SavedChar) then
          pending := pNone
        else
          pending := pCharSaved
      else
        Token := Token + ch;
      SavedChar := ch;
    until (pending <> pNewline);
  end;
  Token := Trim(Token);
  FItemRead := False;
  Result := pending = pEOF;
  if not Result then begin
    FLine := Line + 1;
    FColumn := 1;
  end;
end;

function TTextStream.Getch:Char;
begin
  if pending = pEOF then
    Result := #0
  else if pending = pCharSaved then begin
    pending := pNone;
    Result := SavedChar;
  end else if Read(Result, 1) <> 1 then begin
    pending := pEOF;
    Result := #0;
  end else
    Inc(FColumn);
end;

function TTextStream.RestOfLine: string;
begin
  Result := '';
  while not (pending in [pEOF, pRecord, pNewLine]) do begin
    if pending = pNone then
      SavedChar := Getch;
    if IsEOLChar(SavedChar) then
      pending := pNewline
    else if SavedChar <> #0 then begin
      Result := Result + SavedChar;
      pending := pNone;
    end;
  end;
end;

procedure TTextStream.SkipItems(n: Integer);
begin
  while (n > 0) and GetFreeItem do
    Dec(n);
end;

function TTextStream.GetFreeItem:Boolean; { Returns True if got something }
var
  Quote: Char;
  procedure AddChar;
  begin
    if Length(Token) < 256 then
      Token := Token + SavedChar
    else
      BadToken('Token too long');
    if pending = pCharSaved then
      pending := pNone;
    SavedChar := Getch;
  end;
begin
  Token := '';
  Quote := #0;
  Result := True;
  if (pending = pNone) or (pending = pCharSaved)
     or ((Pending = pNewline) and FlowToNextLine)
     or ((pending = pRecord) and FlowToNextRecord)
      then begin
    if pending = pNone then begin
      SavedChar := Getch;
      pending := pCharSaved;
    end else if Pending = pNewline then
      NextLine;
    while (pending <> pEOF) and Result do begin
      while SavedChar = ' ' do SavedChar := Getch;
      if ( (SavedChar = #9) and (isTab in ItemSeparators) )
         or ((SavedChar = ',') and (isComma in ItemSeparators)) then begin
         if FItemRead then
           pending := pNone { Absorb separator of previous item }
         else begin
           Result := False; { Defaulted Item }
           Pending := pCharSaved;
         end;
        FItemRead := False; { Either way no longer in a field }
      end else if IsEOLChar(SavedChar) and FlowToNextLine then begin
        pending := pNewline;
        NextLine;
      end else if SavedChar <> #9 then
        Break;
      SavedChar := Getch;
    end;
    if (pending <> pEOF) and Result then begin
      repeat
        case SavedChar of
          #0: if NULLisEOF then pending := pEOF else Break;
          #9, ' ': if Quote = #0 then begin
                     pending := pCharSaved;
                     Break;
                   end else AddChar;
          ',': if (Quote = #0) and (isComma in ItemSeparators) then begin
                 pending := pCharSaved;
                 Break;
               end else AddChar;
          #10,#13:
            begin
              pending := pNewline;
              if (Quote <> #0) and (Length(Token) > 0) and (Token[Length(Token)] = '\') then
                NextLine
              else
                Break;
            end;
          '/', ';', '#':
            if Comments and (Quote = #0) then
              if FlowToNextRecord then
                pending := pNewline
              else
                pending := pRecord
            else AddChar;
          '''', '"':
            if (Quote = #0) and (Length(Token) = 0) then begin
                Quote := SavedChar;
                SavedChar := Getch;
            end else if Quote = SavedChar then begin
              SavedChar := Getch;
              if SavedChar = Quote then
                AddChar { Doubled quote character - treat as single quote in string }
              else begin
                Pending := pCharSaved;
                Break;
              end;
            end else
              AddChar; { Quote in middle of Token - either unquoted string or other sort of quote }
        else
          AddChar;
        end;
      until AtEndOfLine;
      FItemRead := True;
    end;
  end;
  Result := Length(Token) > 0;
end;

function TTextStream.IsRealToken(var Value: Double): Boolean;
begin
  Result := TryStrToFloat(AnsiReplaceText(Token,'d','e'),Value);
end;

function TTextStream.RealToken:Double;
begin
  try
    if AnsiCompareText(Token, 'NAN') = 0 then
      SetNan(Result)
    else if (Token[1] = '*') and (Token = StringOfChar('*',Length(Token))) then
      SetNan(Result)
    else
      {$IFNDEF FPC}
      Result := StrToFloat(AnsiReplaceText(Token,'d','e'));
      {$ELSE}
      Result := StrToFloat(Token); {TODO: Is 1.0d3 OK?}
      {$ENDIF}
  except
    BadToken('Bad Floating Point Number');
    Result := 0;
  end;
end;

function TTextStream.IntegerToken:Integer;
begin
  try
    Result := StrToInt(Token);
  except
    BadToken('Bad Integer');
    Result := 0;
  end;
end;

function TTextStream.NextInteger(Default: Integer): Integer;
begin
  Result := Default;
  if GetFreeItem then
    Result := IntegerToken;
end;

function TTextStream.NextDouble(Default: Double): Double;
begin
  Result := Default;
  if GetFreeItem then
    Result := RealToken;
end;

function TTextStream.NextString(Default: string): string;
begin
  Result := Default;
  if GetFreeItem then
    Result := Token;
end;

function TTextStream.NextIntegerVar(var i: Integer):Boolean;
begin
  Result := GetFreeItem;
  if Result then
    i := IntegerToken
  else if not AllowDefaults then
    raise Exception.Create('No integer');
end;

function TTextStream.NextDoubleVar(out d: Double):Boolean;
begin
  Result := GetFreeItem;
  if Result then
    d := RealToken
  else if not AllowDefaults then
    raise Exception.Create('No double');
end;

function TTextStream.NextStringVar(var s: string): Boolean;
begin
  Result := GetFreeItem;
  if Result then
    s := Token
  else if not AllowDefaults then
    raise Exception.Create('No string');
end;

constructor TTextStream.Create(Stream:TStream);
begin
  inherited Create;
  FStream := Stream;
  pending := pNone;
  Token := '';
  FLine := 1;
  FColumn := 0;
  NULLisEOF := True;
  ItemSeparators := [isTab];
  FlowToNextRecord := True;
end;

constructor TFreeFormatStream.Create(Stream:TStream);
begin
  inherited Create(Stream);
  ItemSeparators := [isComma];
  Comments := True;
  FlowToNextLine := True;
  FlowToNextRecord := False;
  AllowDefaults := True;
end;

destructor TTextStream.Destroy;
begin
  inherited Destroy;
end;

function TTextStream.Read(var Buffer; Count: Longint): Longint;
begin
  Result := FStream.Read(Buffer, Count);
end;

function TTextStream.Write(const Buffer; Count: Longint): Longint;
begin
  if pending = pCharSaved then
    Seek(0, soFromCurrent);
  pending := pNone;
  Result := FStream.Write(Buffer, Count);
end;

function TTextStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  if (pending = pCharSaved) and (Origin = soFromCurrent) then
    Offset := Offset - 1;
  pending := pNone;
  FItemRead := False;
  Result := FStream.Seek(Offset, Origin);
end;

function TTextStream.PutChar(c:Char):TTextStream;
begin
  WriteBuffer(c, 1);
  Result := Self;
end;

function TTextStream.PutFloat(v:Extended):TTextStream;
begin
  PutString(FloatToStr(v));
  Result := Self;
end;

function TTextStream.PutInteger(i:Longint):TTextStream;
begin
  PutString(IntToStr(i));
  Result := Self;
end;

function TTextStream.PutString(s:string):TTextStream;
begin
  WriteBuffer(s[1], Length(s));
  Result := Self;
end;

function TTextStream.Format(const FormatStr: string; const Args: array of const): TTextStream;
begin
  Result := PutString(SysUtils.Format(FormatStr, Args));
end;

function TTextStream.PutPChar(p:PChar):TTextStream;
begin
  WriteBuffer(p[0], StrLen(p));
  Result := Self;
end;

function TTextStream.Space:TTextStream;
begin
  PutChar(' ');
  Result := Self;
end;

function TTextStream.Tab:TTextStream;
begin
  PutChar(#9);
  Result := Self;
end;

function TTextStream.Newline:TTextStream;
begin
  PutChar(#13);
  PutChar(#10);
  Result := Self;
end;

end.
