unit UCompat;
{ Various routines to help with compatibility between different operating
  systems and compilers }
{$I DFS.inc}
{
Putils - utility routines for pgopger
Copyright (C) 2005 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
interface

uses
  Classes, IniFiles, SysUtils;


{$IFDEF UNIX}
const
  LineEnd: string = #10;
  NewLineLength = 1;
{$ENDIF}

{$IFDEF LINUX}
function CopyFile(Source, Dest:PChar; bFailIfExists: Boolean):Boolean;
{$ENDIF}

{$IFDEF FPC}
{$IFNDEF MSWINDOWS}
  function MulDiv(A, B, C:Integer):Integer;
{$ENDIF}
function RoundTo(const Value: Double; Places: Integer):Double;
{$ENDIF}

{$IFDEF MSWINDOWS}
const
  LineEnd: string = #13#10;
  NewLineLength = 2;

{$IFDEF VER80}
const
  MinusOne:Cardinal = $ffff;
{$ELSE}
const
  MinusOne:LongInt = -1;
{$ENDIF}


{$IFNDEF DFS_COMPILER_6_UP}
{$IFNDEF FPC}
  PathDelim  = '\';
  DriveDelim = ':';
  PathSep    = ';';
procedure RaiseLastOSError;
{$ENDIF}
{$ENDIF}

{$IFNDEF DFS_COMPILER_5_UP}
procedure FreeAndNil(var Obj);
{$ENDIF}
{$ENDIF}

{$ifndef VER80}
procedure hmemcpy(p1, p2: Pointer; L: Longint);
{$endif}

function AddressOfFirstChar(const s:string):PChar;

function GetIniFile(const Name: string; WithBorland:Boolean{ = False}): TCustomIniFile;
procedure AllowCtrlC; { Allow ^C to kill process }

{$IFNDEF FPC}
type
  TLibHandle = THandle;
{$ENDIF}

implementation

uses
  {$IFDEF MSWINDOWS}
    Registry, Windows,
  {$ELSE}
    {$IFNDEF DARWIN} Libc, {$ENDIF}
    {$IFDEF FPC} BaseUnix, {$ENDIF}
  {$ENDIF}
  Math;

{$IFNDEF VER80}
{$IFDEF MSWINDOWS}

procedure hmemcpy(p1, p2: Pointer; L: Longint);
begin
  CopyMemory(p1, p2, L);
end;
{$IFNDEF DFS_COMPILER_6_UP}
{$IFNDEF FPC}
procedure RaiseLastOSError;
begin
  RaiseLastWin32Error;
end;
{$ENDIF}
{$ENDIF}
{$ENDIF}

{$IFDEF UNIX}

{$IFDEF FPC}

function MulDiv(A, B, C:Integer):Integer;
begin
  Result := (A * B) div C;
end;

{$ENDIF}

procedure hmemcpy(p1, p2: Pointer; L: Longint);
begin
  Move(p2^, p1^, L);
end;
{$ENDIF}

{$ELSE}
uses
  SysUtils;
{$ENDIF}

{$IFDEF MSWINDOWS}
{$IFNDEF DFS_COMPILER_5_UP}
procedure FreeAndNil(var Obj);
var
  Temp: TObject;
begin
  Temp := TObject(Obj);
  Pointer(Obj) := nil;
  Temp.Free;
end;
{$ENDIF}
{$ENDIF}

function AddressOfFirstChar(const s:string):PChar;
begin
{$IFNDEF VER80}
  Result := PChar(s);
{$ELSE}
  Result := @s[1];
{$ENDIF}
end;

{$IFDEF LINUX}

{$IFDEF cpVERSION}
var
  CopyDosCommand: TDosCommand;
{$ENDIF}

function CopyFile(Source, Dest:PChar; bFailIfExists: Boolean):Boolean;
var
  // ExitCode: LongWord;
  SourceStream, DestStream: TFileStream;
  Count: Int64;
  st, dt: TStatBuf;
  u: utimbuf;
begin
  Result := False;
  if bFailIfExists and FileExists(Dest) then begin
    {$IFDEF FPC}
    fpseterrno(EEXIST);
    {$ELSE}
    SetLastError(EEXIST);
    {$ENDIF}
    Exit;
  end;
  {$IFDEF cpVERSION}
  if CopyDosCommand = nil then begin
    CopyDosCommand := TDosCommand.Create(Application);
    with CopyDosCommand.Arguments do begin
      Add('cp');
      Add('-p');
      Add(''); { Source }
      Add(''); { Dest }
    end;
  end;
  with CopyDosCommand do begin
    Arguments[2] := Source;
    Arguments[3] := Dest;
    Execute;
    ExitCode := WaitFor;
    if Lines.Count > 0 then
      raise Exception.Create(Lines[0])
    else if ExitCode <> 0 then
      SetLastError(ExitCode) { May not be meaningful }
    else
      Result := True;
  end;
  {$ENDIF}
  SourceStream := TFileStream.Create(Source, fmOpenRead);
  try
    DestStream := TFileStream.Create(Dest, fmCreate);
    Count := DestStream.CopyFrom(SourceStream, 0);
    if (Libc.fstat(SourceStream.Handle, st) = -1)
       or (Libc.fstat(DestStream.Handle, dt) = -1) then
      RaiseLastOSError;
    if Count <> dt.st_size then
      raise Exception.Create('Wrong number of bytes copied to '+Dest);
    if Libc.fchmod(DestStream.Handle, st.st_mode) = -1 then
      RaiseLastOSError;
    { Try and copy owner and group, but ignore errors }
    if st.st_gid <> dt.st_gid then
      Libc.fchown(DestStream.Handle, uid_t(-1), st.st_gid);
    if (st.st_uid <> dt.st_uid) and (Libc.getuid = 0) then
      Libc.fchown(DestStream.Handle, st.st_uid, gid_t(-1));
    FreeAndNil(DestStream);
    u.actime := st.st_atime;
    u.modtime := st.st_mtime;
    if Libc.utime(Dest, @u) = -1 then
      RaiseLastOSError;
    Result := True;
  finally
    DestStream.Free;
  end;
  SourceStream.Free;
end;
{$ENDIF}

{$IFDEF UNIX}
type
  TMyMemIniFile = class(TMemIniFile)
  public
    destructor Destroy; override;
  end;

destructor TMyMemIniFile.Destroy;
begin
  UpdateFile;
  inherited;
end;
{$ENDIF}

function GetIniFile(const Name: string; WithBorland:Boolean{ = False}): TCustomIniFile;
var
  KeyName: string;
begin
{$IFDEF MSWINDOWS}
  {$IFDEF VER80} {$DEFINE USEINIFILE} {$ENDIF}
  {$IFDEF FPC} {$DEFINE USEINIFILE} {$ENDIF} { TODO: Implement TRegistryIniFile for FPFC }
  {$IFDEF USEINIFILE}
    KeyName := Name + '.ini';
    Result := TIniFile.Create(KeyName);
  {$ELSE}
    if WithBorland then
      KeyName := 'SOFTWARE\BORLAND\DELPHI\' + Name
    else
      KeyName := 'SOFTWARE\' + Name;
    Result := TRegistryIniFile.Create(KeyName);
  {$ENDIF}
{$ELSE}
  {$IFDEF UNIX}
    KeyName := GetEnvironmentVariable('HOME');
    if WithBorland then
      KeyName := KeyName + {$IFDEF KYLIX}'/.borland/'{$ELSE}'/.lazarus/'{$ENDIF} + Name
    else
      KeyName := KeyName + '/.' + Name;
    Result := TMyMemIniFile.Create(KeyName);
  {$ENDIF}
{$ENDIF}
end;

procedure AllowCtrlC; { Allow ^C to kill process }
begin
  {$IFDEF LINUX} { TODO: Darwin }
  Libc.signal(SIGINT, TSignalHandler(SIG_DFL));
  {$ENDIF}
end;

{$IFDEF FPC}
function RoundTo(const Value: Double; Places: Integer):Double;
var
  Scale: Double;
begin
  Scale := IntPower(10, Places);
  Result := Value/Scale;
  if Abs(Result) < High(Int64) then
    Result := Round(Result); { Avoid overflow }
  Result := Result * Scale;
end;
{$ENDIF}

end.


