unit ExActn;

{ Replacements for Standard edit actions that understand Log windows }

interface
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
uses {$IFNDEF CLX_} ActnList, Controls, StdCtrls, Messages,
     {$ELSE} QActnList, QControls, QStdCtrls, {$ENDIF}
     Classes;

type

  TActionBase = class(TAction)
  private
    FControl: TWinControl;
    procedure SetControl(Value: TWinControl);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    procedure SetShortCut(Key: Word; Shift: TShiftState);
    property Control: TWinControl read FControl write SetControl;
  end;

  TExAction = class(TActionBase)
  private
    FUpdateMethod: string;
    FExecuteMethod: string;
  public
    function HandlesTarget(Target: TObject): Boolean; override;
    procedure UpdateTarget(Target: TObject); override;
    procedure ExecuteTarget(Target: TObject); override;
  end;

  TExSelAction = class(TExAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TExCut = class(TExSelAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTarget(Target: TObject); override;
  end;

  TExCopy = class(TExSelAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TExClear = class(TExSelAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TExPaste = class(TExAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TExUndo = class(TExAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TExRedo = class(TExAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TExSelectAll = class(TExAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TExPrint = class(TExAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TExPrintPreview = class(TExAction)
  public
    constructor Create(AOwner: TComponent); override;
  end;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

implementation

uses {$IFNDEF CLX_} Menus, Clipbrd,
     {$ELSE} QMenus, QClipbrd, Qt, {$ENDIF}
     {$IFDEF LCL_} LCLType, {$ENDIF}
     {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     SysUtils, ComboFix, WCompat;

type
  TMethodProc = procedure of object;
  TMethodBool = function : Boolean of object;

  TCustomEditHolder = class(TPersistent)
  public
    Control: TCustomEdit;
  published
    procedure CopyToClipboard; virtual;
    function CanCopyToClipboard: Boolean; virtual;
    procedure ClearSelection; virtual;
    procedure PasteFromClipboard; virtual;
    function CanPasteFromClipboard: Boolean; virtual;
    function CanUndo: Boolean; virtual;
    procedure Undo; virtual;
    procedure Redo; virtual;
    function CanRedo: Boolean; virtual;
    procedure SelectAll; virtual;
{   procedure Print; virtual; } { TODO }
{   procedure PrintPreview; virtual; } { TODO }
  end;

  TCustomComboBoxHolder = class(TPersistent)
  public
    Control: TCustomComboBox;
  published
    procedure CopyToClipboard; virtual;
    function CanCopyToClipboard: Boolean; virtual;
    procedure ClearSelection; virtual;
    procedure PasteFromClipboard; virtual;
    function CanPasteFromClipboard: Boolean; virtual;
    function CanUndo: Boolean; virtual;
    procedure Undo; virtual;
    procedure Redo; virtual;
    function CanRedo: Boolean; virtual;
    procedure SelectAll; virtual;
  end;

{$IFDEF CLX_}
  TCustomMemoHolder = class(TPersistent)
  public
    Control: TCustomMemo; {Not a subclass of TCustomEdit}
  published
    procedure CopyToClipboard; virtual;
    function CanCopyToClipboard: Boolean; virtual;
    procedure ClearSelection; virtual;
    procedure PasteFromClipboard; virtual;
    function CanPasteFromClipboard: Boolean; virtual;
    function CanUndo: Boolean; virtual;
    procedure Undo; virtual;
{   procedure Redo; virtual;
    function CanRedo: Boolean; virtual; }
    procedure SelectAll; virtual;
{   procedure Print; virtual; } { TODO }
{   procedure PrintPreview; virtual; } { TODO }
  end;
{$ENDIF}

{ Could do the same again for a rich text edit control }

var
  CustomEditHolder: TCustomEditHolder;
  CustomComboBoxHolder: TCustomComboBoxHolder;
{$IFDEF CLX_}
  CustomMemoHolder: TCustomMemoHolder;
{$ENDIF}

function GetMethod(Target: TObject; var Method: TMethod; const MethodName: string): Boolean;
begin
  Result := True;
  Method.Code := TObject(Target).MethodAddress(MethodName);
  if Method.Code <> nil then
    Method.Data := Target
  else begin
    if Target is TCustomEdit then begin
      CustomEditHolder.Control := Target as TCustomEdit;
      Method.Data := CustomEditHolder;
    end else if Target is TCustomComboBox then begin;
      CustomComboBoxHolder.Control := Target as TCustomComboBox;
      Method.Data := CustomComboBoxHolder;
{$IFDEF CLX_}
    end else if Target is TCustomMemo then begin;
      CustomMemoHolder.Control := Target as TCustomMemo;
      Method.Data := CustomMemoHolder;
{$ENDIF}
    end else
      Result := False;
    if Result then
      Method.Code := TObject(Method.Data).MethodAddress(MethodName);
  end;
  Result := Method.Code <> nil;
end;

procedure DoExecuteMethod(Target: TObject; MethodName: string);
var
  Method: TMethod;
begin
  if GetMethod(Target, Method, MethodName) then
    TMethodProc(Method);
end;

{ TExAction }

procedure TExAction.ExecuteTarget(Target: TObject);
begin
  DoExecuteMethod(Target, FExecuteMethod);
end;

function TExAction.HandlesTarget(Target: TObject): Boolean;
var
  Method: TMethod;
begin
  Result := (Control <> nil) and (Target = Control);
  if not Result and (Control = nil) then
    if GetMethod(Target, Method, FExecuteMethod) then
      Result := TWinControl(Target).Focused;
end;

procedure TExAction.UpdateTarget(Target: TObject);
var
  Method: TMethod;
begin
  if Length(FUpdateMethod) > 0 then begin
    if GetMethod(Target, Method, FUpdateMethod) then
      try
        Enabled := TMethodBool(Method)();
      except
        Enabled := False;
      end;
  end else
    Enabled := True;
end;

{ TExCut }

constructor TExCut.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(Word('X'), [ssCtrl]);
  Hint := 'Cut|Cuts the selection and puts it on the Clipboard';
  ImageIndex := 0;
  Caption := 'Cu&t';
  FExecuteMethod := 'ClearSelection';
end;

procedure TExCut.ExecuteTarget(Target: TObject);
begin
  DoExecuteMethod(Target, 'CopyToClipboard');
  inherited;
end;

{ TExRedo }

constructor TExRedo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(Word('Z'), [ssCtrl, ssShift]);
  Hint := 'Redo|Redo last undone action';
  ImageIndex := 4;
  Caption := '&Redo';
  FUpdateMethod := 'CanRedo';
  FExecuteMethod := 'Redo';
end;

{ TExUndo }

constructor TExUndo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(Word('Z'), [ssCtrl]);
  Hint := 'Undo|Undo last action';
  ImageIndex := 3;
  Caption := '&Undo';
  FUpdateMethod := 'CanUndo';
  FExecuteMethod := 'Undo';
end;

{ TExCopy }

constructor TExCopy.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(Word('C'), [ssCtrl]);
  Hint := 'Copy|Copies the selection and puts it on the Clipboard';
  ImageIndex := 1;
  Caption := '&Copy';
end;

{ TExPaste }

constructor TExPaste.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(Word('V'), [ssCtrl]);
  Hint := 'Paste|Inserts Clipboard contents';
  ImageIndex := 2;
  Caption := '&Paste';
  FUpdateMethod := 'CanPasteFromClipboard';
  FExecuteMethod := 'PasteFromClipboard';
end;

{ TExSelAction }

constructor TExSelAction.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FUpdateMethod := 'CanCopyToClipboard';
  FExecuteMethod := 'CopyToClipboard';
end;

{ TExClear }

constructor TExClear.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(VK_DELETE, [ssCtrl]);
  Hint := 'Clear|Erases the selection';
  ImageIndex := 5;
  Caption := 'Cle&ar';
  FExecuteMethod := 'ClearSelection';
end;

{ TExSelectAll }

constructor TExSelectAll.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(Word('A'), [ssCtrl]);
  Hint := 'Select All|Select all text';
  Caption := 'Select &All';
  FExecuteMethod := 'SelectAll';
end;

{ TActionBase }

procedure TActionBase.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = Control) then Control := nil;
end;

procedure TActionBase.SetControl(Value: TWinControl);
begin
  if Value <> FControl then begin
    FControl := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TActionBase.SetShortCut(Key: Word; Shift: TShiftState);
begin
  ShortCut := {$IFNDEF CLX_}Menus{$ELSE}QMenus{$ENDIF}.ShortCut(Key, Shift);
end;

{ TExPrint }

constructor TExPrint.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Print';
  Caption := '&Print';
  FExecuteMethod := 'Print';
  FUpdateMethod := 'CanPrint';
  ImageIndex := 14;
end;

{ TExPrintPreview }

constructor TExPrintPreview.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Print Preview|Show and adjust how a printed page will look';
  Caption := 'Print Pre&view';
  FExecuteMethod := 'PrintPreview';
  FUpdateMethod := 'CanPrint';
  ImageIndex := 9;
end;

{ TCustomEditHolder }

function TCustomEditHolder.CanCopyToClipboard: Boolean;
begin
  Result := Control.SelLength > 0;
end;

function TCustomEditHolder.CanRedo: Boolean;
begin
  { Neither platform can do this }
  Result := True;
end;

function TCustomEditHolder.CanUndo: Boolean;
begin
  {$IFDEF VCL_}
  Result := Control.CanUndo;
  {$ELSE}
  Result := True;
  {$ENDIF}
end;

procedure TCustomEditHolder.ClearSelection;
begin
  Control.ClearSelection;
end;

procedure TCustomEditHolder.CopyToClipboard;
begin
  Control.CopyToClipboard;
end;

procedure TCustomEditHolder.PasteFromClipboard;
begin
  Control.PasteFromClipboard;
end;

function TCustomEditHolder.CanPasteFromClipboard: Boolean;
begin
  Result := ClipboardHasText;
end;

procedure TCustomEditHolder.Redo;
begin
  {$IFDEF VCL_}
  Control.Undo;
  {$ENDIF}
  {$IFDEF CLX_}
    {$IFNDEF DESIGNTIME}
    Control.Redo;
    {$ENDIF}
  {$ENDIF}
  { TODO: LCL_ }
end;

procedure TCustomEditHolder.SelectAll;
begin
  Control.SelectAll;
end;

procedure TCustomEditHolder.Undo;
begin
  {$IFNDEF DESIGNTIME}
  {$IFNDEF LCL_}
  Control.Undo;
  {$ENDIF}
  {$ENDIF}
  { TODO: LCL_ }
end;

{ TCustomComboBoxHolder }

function TCustomComboBoxHolder.CanCopyToClipboard: Boolean;
begin
  Result := Control.SelLength > 0;
end;

function TCustomComboBoxHolder.CanRedo: Boolean;
begin
  Result := ComboBoxCanRedo(Control);
end;

function TCustomComboBoxHolder.CanUndo: Boolean;
begin
  Result := ComboBoxCanUndo(Control);
end;

procedure TCustomComboBoxHolder.ClearSelection;
begin
  Control.SelText := '';
end;

procedure TCustomComboBoxHolder.CopyToClipboard;
begin
  Clipboard.AsText := Control.SelText;
end;

procedure TCustomComboBoxHolder.PasteFromClipboard;
begin
  Control.SelText := Clipboard.AsText;
end;

function TCustomComboBoxHolder.CanPasteFromClipboard: Boolean;
begin
  Result := ClipboardHasText;
end;

procedure TCustomComboBoxHolder.Redo;
begin
  ComboBoxRedo(Control);
end;

procedure TCustomComboBoxHolder.SelectAll;
begin
  Control.SelectAll;
end;

procedure TCustomComboBoxHolder.Undo;
begin
  ComboBoxUndo(Control);
end;

{$IFDEF SINGLE}
procedure Register;
begin
  RegisterActions('Edit', [TExUndo, TExRedo, TExCut, TExCopy, TExPaste, TExClear, TExSelectAll, TExPrint,
                           TExPrintPreview], nil);
end;
{$ENDIF}

{$IFDEF CLX_}
{ TCustomMemoHolder }
function TCustomMemoHolder.CanCopyToClipboard: Boolean;
begin
  Result := Control.HasSelection;
end;

function TCustomMemoHolder.CanUndo: Boolean;
begin
  Result := QMultiLineEdit_isUndoEnabled(Control.Handle);
end;

procedure TCustomMemoHolder.ClearSelection;
begin
  Control.ClearSelection;
end;

procedure TCustomMemoHolder.CopyToClipboard;
begin
  Control.CopyToClipboard;
end;

procedure TCustomMemoHolder.PasteFromClipboard;
begin
  Control.PasteFromClipboard;
end;

function TCustomMemoHolder.CanPasteFromClipboard: Boolean;
begin
  Result := ClipboardHasText;
end;

procedure TCustomMemoHolder.SelectAll;
begin
  Control.SelectAll;
end;

procedure TCustomMemoHolder.Undo;
begin
  QMultiLineEdit_undo(Control.Handle);
end;
{$ENDIF}

initialization
  CustomEditHolder := TCustomEditHolder.Create;
  CustomComboBoxHolder := TCustomComboBoxHolder.Create;
  {$IFDEF CLX_}
  CustomMemoHolder := TCustomMemoHolder.Create;
  {$ENDIF}

finalization
  FreeAndNil(CustomEditHolder);
  FreeAndNil(CustomComboBoxHolder);
  
end.
