unit FixFnt;

interface

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFNDEF CLX_} Graphics, {$ELSE} QGraphics, Qt, {$ENDIF}
     {$IFDEF LCL_} LCLType, {$ENDIF}
     Classes;

type
  TFixedFont = class (TPersistent)
  private
    FSize: string;
    FName: string;
    FBold: Boolean;
    FHeight: Integer;
    FWidth: Integer;
    FOnChange: TNotifyEvent;
    function NameStored: Boolean;
    function SizeStored: Boolean;
    function GetHandle: {$IFNDEF CLX_} HFont; {$ELSE} QFontH; {$ENDIF}
    procedure SetName(const Value: string);
    procedure SetSize(const Value: string);
    procedure SetBold(const Value: Boolean);
  protected
    procedure Changed;
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    procedure CopyTo(Font: TFont);
    property Width:Integer read FWidth;
    property Height:Integer read FHeight;
    { Note - potentially gives resource leaks, as does not free handle }
    { But it seems that under KYLIX, default is owned handle, contrary to documentation }
    property Handle:{$IFNDEF CLX_} HFont {$ELSE} QFontH {$ENDIF} read GetHandle;
  published
    property Size:string read FSize write SetSize stored SizeStored;
    property Name:string read FName write SetName stored NameStored;
    property Bold: Boolean read FBold write SetBold default False;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

function DefaultFixedName: string;
function DefaultFixedSize: string;

var
  FixedFont: TFixedFont;

implementation

uses {$IFDEF LCL_} LCLIntf, {$ENDIF} SysUtils;

{ TFixedFont }

function DefaultFixedName: string;
begin
  {$IFDEF MSWINDOWS}
  Result := 'Courier New';
  {$ENDIF}
  {$IFDEF UNIX}
  Result := 'misc-fixed';
  {$ENDIF}
  {$IFDEF LCL_}
  Result := 'lucidatypewriter';
  // Result := '-*-fixed-medium-r-*-*-13-*-*-*-*-*-*-1';
  {$ENDIF}
end;

function DefaultFixedSize: string;
begin
  {$IFDEF MSWINDOWS}
  Result := '12';
  {$ENDIF}
  {$IFDEF UNIX}
  Result := '10';
  {$ENDIF}
end;

procedure TFixedFont.Assign(Source: TPersistent);
var
  S: TFixedFont;
  f: TFont;
begin
  if Source is TFixedFont then begin
    s := TFixedFont(Source);
    Size := s.Size;
    Name := s.Name;
    Bold := s.Bold;
  end else if Source is TFont then begin
    f := TFont(Source);
    Size := IntToStr(f.Size);
    Name := f.Name;
    Bold := fsBold in f.Style;
  end else
    inherited Assign(Source);
end;

procedure TFixedFont.CopyTo(Font: TFont);
begin
  { Needs to partly mimic GetHandle Logic }
  {$IFDEF MSWINDOWS}
  if FWidth > 0 then
    Font.Height := FHeight
  else
    Font.Size := FHeight;
  {$ENDIF}
  {$IFDEF UNIX}
  Font.Size := FHeight;
  {$ENDIF}
  Font.Name := Name;
  if Bold then
    Font.Style := [fsBold]
  else
    Font.Style := [];
  Font.Color := clWindowText;
  Font.Pitch := fpFixed;
end;

procedure TFixedFont.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

constructor TFixedFont.Create;
begin
  inherited Create;
  Size := DefaultFixedSize;
  FName := DefaultFixedName;
  Bold := False;
end;

{$IFDEF MSWINDOWS}
function TFixedFont.GetHandle: HFont;
var
  LogFont: TLogFont;
begin
  with LogFont do begin
    lfWidth := FWidth;
    if FWidth > 0 then
      lfHeight := FHeight
    else
      lfHeight := -FHeight;
    lfEscapement := 0;
    lfOrientation := 0; { no rotation }
    if Bold then
      lfWeight := FW_BOLD
    else
      lfWeight := FW_NORMAL;
    lfItalic := 0;
    lfUnderline := 0;
    lfStrikeOut := 0;
    lfCharSet := DEFAULT_CHARSET;
    StrPCopy(lfFaceName, Name);
    lfQuality := DEFAULT_QUALITY;
    lfOutPrecision := OUT_DEFAULT_PRECIS;
    lfClipPrecision := CLIP_DEFAULT_PRECIS;
    lfPitchAndFamily := FIXED_PITCH or FF_DONTCARE;
  end;
  Result := CreateFontIndirect(LogFont);
end;
{$ELSE}
{$IFDEF LCL_}
function TFixedFont.GetHandle: HFont;
var
  LogFont: TLogFont;
begin
  Result := 0;
  if IsFontNameXLogicalFontDesc(Name) then
    LogFont := XLFDNameToLogFont(Name)
  else with LogFont do begin
    lfWidth := FWidth;
    if FWidth > 0 then
      lfHeight := FHeight
    else
      lfHeight := -FHeight;
    lfEscapement := 0;
    lfOrientation := 0; { no rotation }
    if Bold then
      lfWeight := FW_BOLD
    else
      lfWeight := FW_NORMAL;
    lfItalic := 0;
    lfUnderline := 0;
    lfStrikeOut := 0;
    lfCharSet := DEFAULT_CHARSET;
    lfFaceName := Name;
    lfQuality := DEFAULT_QUALITY;
    lfOutPrecision := OUT_DEFAULT_PRECIS;
    lfClipPrecision := CLIP_DEFAULT_PRECIS;
    lfPitchAndFamily := FIXED_PITCH or FF_DONTCARE;
  end;
  Result := CreateFontIndirectEx(LogFont, Name);
end;
{$ENDIF}
{$IFDEF CLX_}
function TFixedFont.GetHandle: QFontH;
var
  WFontName: WideString;
begin
  Result := QFont_create;
  WFontName := Name;
  QFont_setFamily(Result, @WFontName);
  QFont_setPointSize(Result, StrToInt(Size));
  QFont_setFixedPitch(Result, True);
  QFont_setBold(Result, Bold);
end;
{$ENDIF}
{$ENDIF}

procedure TFixedFont.SetBold(const Value: Boolean);
begin
  if FBold <> Value then begin
    FBold := Value;
    Changed;
  end
end;

procedure TFixedFont.SetName(const Value: string);
begin
  if FName <> Value then begin
    FName := Value;
    Changed;
  end
end;

function TFixedFont.NameStored: Boolean;
begin
  Result := Name <> DefaultFixedName;
end;

function TFixedFont.SizeStored: Boolean;
begin
  Result := Size <> DefaultFixedSize;
end;

procedure TFixedFont.SetSize(const Value: string);
var
  i: Integer;
begin
  if FSize <> Value then begin
    FSize := Value;
    i := Pos('x', Size);
    if i > 0 then begin
      FWidth := StrToInt(Copy(Size,1,i-2));
      FHeight := StrToInt(Copy(Size,i+2,10));
    end else begin
      FHeight := StrToInt(Size);
      FWidth := 0;
    end;
    Changed;
  end;
end;

initialization
finalization
  FreeAndNil(FixedFont);
end.
